#pragma once

#include <stdint.h> 
#include <vector>
#include "dpc/impl_type_linegeometry.h"
using namespace dpc;

namespace rm_interface {

struct RMHeader {
uint32_t num_sequence;
uint64_t timestamp;
uint32_t data_version;
uint32_t software_version;
uint32_t status;
uint16_t cyclecounter;
};

struct Index {
uint8_t x;
uint8_t y;
};

struct CarPosition {
uint8_t ego_lane_id;
uint8_t lane_line_crossed;
std::vector<uint8_t> target_lane_id_list;
};

struct NeighborLane
{
float start_range;
float end_range;
uint8_t lane_id;
};

// struct LineGeometry
// {
// float c0;
// float c1;
// float c2;
// float c3;
// float start_range;
// float end_range;
// };

struct SLPoint
{
float s;
float ds;
float dds;
float dds_valid_time;
float l;
float dl;
float dl_valid_time;
float ddl;
float ddl_valid_time;
};


struct ObjectLocation
{
uint32_t object_id;
SLPoint reference_point2corridor_frenet;
float relative_center_x;
float relative_center_y;
float size_width;
float size_length;
uint8_t classification;
};

struct Corridor {
uint32_t corridor_id;
uint8_t feature_tag;
uint32_t ilc_object_target;
uint32_t disenable_reason;
float drivable_width;
LineGeometry reference_line;
std::vector<ObjectLocation> object_in_corridor_list;
std::vector<ObjectLocation> relevant_object_corridor_list;
};

struct SamplingPoint
{
float x;
float y;
float curve;
bool valid;
};

struct LaneControl {
float start_range;
float end_range;
float value;
};

struct Lane {
uint8_t start_id;
uint8_t end_id;
float start_range;
float end_range;
std::vector<NeighborLane> left_neighbor_lane;
std::vector<NeighborLane> right_neighbor_lane;
std::vector<uint32_t> left_lane_marker_id_list;
std::vector<uint32_t> right_lane_marker_id_list;
std::vector<Corridor> corridor_list;
std::vector<Corridor> back_corridor_list;
std::vector<SamplingPoint> sampling_point_list;
uint8_t lane_type;
std::vector<LaneControl> max_speed_limit;
float lane_width;
bool is_start_out_of_range;
bool is_end_out_of_range;
uint8_t lane_start_split_merge_type;
uint8_t lane_end_split_merge_type;
};

struct LaneBoundary
{
uint32_t line_boundary_id;
LineGeometry line_geometry;
float arc_length;
std::vector<SamplingPoint> sampling_point_list;
};

struct LineColor
{
float start_offset;
float end_offset;
uint8_t linecolor;
};

struct LineType
{
float start_offset;
float end_offset;
uint8_t linetype;
};

struct LaneMarker
{
uint32_t id;
std::vector<LaneBoundary> front_lane_boundary_list;
std::vector<LaneBoundary> back_lane_boundary_list;
std::vector<LineColor> line_color_attribute_list;
std::vector<LineType> line_type_attribute_list;
bool is_start_out_of_range;
bool is_end_out_of_range;
};

struct TrajectoryPoint
{
double timestamp;
float x;
float y;
float orientation;
float std_x;
float std_y;
float std_orientation;
};

struct ObjectIntention
{
uint8_t intention_type;
float intention_propability;
float time_to_collision;
float collision_probability;
uint8_t target_lane;
float time_to_lane;
int8_t number_of_trajectory_points;
std::vector<TrajectoryPoint> trajectory_point_list;
};

struct Object
{
uint32_t object_id;
float object_exist_probability;
uint8_t classification;
float relative_center_x;
float relative_center_y;
float relative_position_x;
float relative_position_y;
float absolute_velocity_x;
float absolute_velocity_y;
float absolute_acceleration_x;
float absolute_acceleration_y;
float absolute_yawrate;
float size_width;
float size_length;
bool intentions_valid;
uint8_t intentions_number;
std::vector<ObjectIntention> intention_list;
};

struct RoadEdgeSegment
{
uint32_t id;
uint16_t contributing_sensor;
uint8_t type;
float height_above_ground;
LineGeometry line_geometry;
uint8_t is_start_out_of_view;
uint8_t is_end_out_of_view;
};

struct LongitudinalInfo {
float desired_distance;
float desired_speed;
float set_speed;
float set_distance;
float yellow_border_distance;
float red_border_distance;
uint32_t target_object_id;

// 01 the first nearest speed limit sign info
float sign_spd_limit_val_01;
float sign_spd_limit_dist_01;
uint8_t sign_spd_limit_id_01;
// 02 the second nearest speed limit sign info (reserved)
float sign_spd_limit_val_02;
float sign_spd_limit_dist_02;
uint8_t sign_spd_limit_id_02;
// curve road speed limit info for pre-deceleration
float curv_spd_limit_val;
float curv_spd_limit_dist;
// results in last cycle
bool spd_limit_flg_last_cycle;
uint8_t spd_limit_type_last_cycle;
uint8_t spd_limit_id_last_cycle;    // check when speed limit type is SPD_LIMIT_1st or SPD_LIMIT_2nd
};

struct Passage
{
    Index passage_id;
uint32_t corridor_id;
LongitudinalInfo longt_info;
uint8_t action_type;
uint8_t feature_label;
float duration_time;
std::vector<uint32_t> interest_obj_id_list;
float v_ego;
uint8_t lc_cls;
bool lc_passagegenflg;
bool lc_ego_frt_obj_exist_flg;
uint32_t lc_ego_frt_obj_id;
bool lc_tar_frt_obj_exist_flg;
uint32_t lc_tar_frt_obj_id;
bool lc_tar_bck_obj_exist_flg;
uint32_t lc_tar_bck_obj_id;
uint8_t lcp_cls;
bool lcp_frt_passagegenflg;
bool lcp_bck_passagegenflg;
bool lcp_ego_frt_obj_exist_flg;
uint32_t lcp_ego_frt_obj_id;
bool lcp_ego_bck_obj_exist_flg;
uint32_t lcp_ego_bck_obj_id;
bool lcp_nr_obj_exist_flg;
uint32_t lcp_nr_obj_id;
bool lcp_nr_frt_obj_exist_flg;
uint32_t lcp_nr_frt_obj_id;
bool lcp_nr_bck_obj_exist_flg;
uint32_t lcp_nr_bck_obj_id;
};

struct RoadEdge
{
uint32_t road_edge_id;
std::vector<RoadEdgeSegment> road_edge_segment_list;
};

struct RoadInfo {
uint64_t fusion_timestamp;
RMHeader rm_header;
Index road_id;
CarPosition car_position;
std::vector<Lane> lane_list;
std::vector<LaneMarker> lane_marker_list;
std::vector<Object> object_list;
std::vector<RoadEdge> road_edge_list;
std::vector<Passage> passage_list;
std::vector<Passage> passage_list_ob;
float m_lastnode_s;
float m_lastnode_vx;
float m_lastnode_vy;
float m_lastnode_ax;
float m_lastnode_ay;
float m_lastnode_jkx;
float m_lastnode_spdlimt_val;
float m_lastnode_spdlimt_dist;
bool m_lastnode_spdlimt_flag;
};

}
