#include "global_cost_calculation.h"
#include <algorithm>
#include <functional>
#include <cmath>
#include "glog/logging.h"

void GlobalCostCalculation::Init(const BA::GlobalInfoContainer &globalInfoContainer,
                                 const BA::HV_Info &hostVehInfo,
                                 const RoadInfo &roadModel,
                                 const std::vector<BA::BA_ActionTyp> &actionTypArr)
{
    this->m_GlobalInfoContainer = globalInfoContainer;
    this->m_CurHostVehInfo = hostVehInfo;
    this->m_ActionTypArr.resize(actionTypArr.size());
    copy(actionTypArr.begin(), actionTypArr.end(), this->m_ActionTypArr.begin());
    this->m_CycleTimer = 0;
    this->m_GlobalCost = 0;
    this->m_GlobalPathFllwCost = 0;
    this->m_LgTermFastLnCost = 0;
    this->m_LgTermSlowLnCost = 0;
    this->m_VehFlowInLnCost = 0; // not assigned
    this->m_LaneEndCost = 0;
    this->m_TrffcEffcCost = 0;
    // get road speed limit
    if (roadModel.lane_list.empty())
    {
        DLOG(WARNING) << "lane_list provided by rm is empty!";
    }
    else
    {

        float tmpRdSpdLmt = 80.0f; // TODO: make sure the nearest section's max speed limit is selected
        for (int i = 0; i < roadModel.lane_list.size(); i++)
        {
            // select max max_speed_limit as road speed limit
            if (!roadModel.lane_list[i].max_speed_limit.empty())
            {
                if (roadModel.lane_list[i].max_speed_limit[0].value > tmpRdSpdLmt)
                {
                    tmpRdSpdLmt = roadModel.lane_list[i].max_speed_limit[0].value;
                }
            }
        }
        this->m_RoadSpdLmt = tmpRdSpdLmt;
    }
}

void GlobalCostCalculation::RevCrdSysOrgPnt(float dstToWayPnt)
{
    std::vector<float>::iterator it = this->m_CurHostVehInfo.m_DistanceAlongS.begin();
    for (; it != this->m_CurHostVehInfo.m_DistanceAlongS.end(); it++)
    {
        *it = dstToWayPnt - *it;
    }
}

void GlobalCostCalculation::CalcGlobalPathFllwCost()
{
    if (this->m_GlobalInfoContainer.m_GlobalInfoSrc == BA::GlobalInfoSrc_NoVldWayPoint)
    {
        this->m_GlobalPathFllwCost = 0;
        DLOG(INFO) << "There is no specific global path to follow." << std::endl;
    }
    else
    {
        // revise m_DistanceAlongS' original point
        RevCrdSysOrgPnt(this->m_GlobalInfoContainer.m_DistToWayPoint);
        float k = 1.0 / this->m_CurHostVehInfo.m_LaneLocation.size(); // all sample points share the same weight coefficiency
        for (int i = 0; i < this->m_CurHostVehInfo.m_LaneLocation.size(); i++)
        {
            uint8_t tmpLaneID = this->m_CurHostVehInfo.m_LaneLocation[i];
            // get related lane change point info
            float tmpDesLnChgPntDst = 0;
            float tmpLatestLnChgPntDst = 0;
            float smplPntPathFllwCost = 0;
            if (GetLnChgPnts(tmpLaneID, tmpDesLnChgPntDst, tmpLatestLnChgPntDst))
            {
                float tmpDstAlongS = this->m_CurHostVehInfo.m_DistanceAlongS[i];
                if (tmpDstAlongS > tmpDesLnChgPntDst)
                {
                    smplPntPathFllwCost = 0; // not overpass related desired lane change point
                }
                else if (tmpDstAlongS < tmpLatestLnChgPntDst)
                {
                    smplPntPathFllwCost = 1; // overpass related latest lane change point
                }
                else
                {
                    smplPntPathFllwCost = (tmpDesLnChgPntDst - tmpDstAlongS) / (tmpDesLnChgPntDst - tmpLatestLnChgPntDst); // located btw desired and latest lane change point
                }
            }
            this->m_GlobalPathFllwCost += k * smplPntPathFllwCost;
        }
    }
}

void GlobalCostCalculation::CalcGlobalPathFllwCost_B()
{
    if (this->m_GlobalInfoContainer.m_GlobalInfoSrc == BA::GlobalInfoSrc_NoVldWayPoint)
    {
        this->m_GlobalPathFllwCost = 0;
        DLOG(INFO) << "There is no specific global path to follow." << std::endl;
    }
    else
    {
        // revise m_DistanceAlongS' original point
        RevCrdSysOrgPnt(this->m_GlobalInfoContainer.m_DistToWayPoint);
        // get lane id and distance
        if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
        {
            uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
            uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
            if (stPntLaneID != endPntLaneID)
            {
                // it is a lane change action
                // check whether ego vehicle has overpassed the desired lane change point of current lane
                // 1 if positive, assign global path follow cost
                // 1.1 check whether ego vehicle has overpassed the desired lane change point of evaluated lane
                // 1.1.1 if positive, compare distance btw ego vehicle position and desired lane change points
                // 1.1.1.a if that of ego lane is greater than that of evaluated lane, then set global path follow cost of evaluated lane to 0
                // 1.1.1.b if that of ego lane is less than that of evaluated lane, then set global path follow cost of evaluated lane to 1
                // 1.1.2 if negative, then set global path follow cost of evaluated lane to 0
                // 2 if negative, do not assign global path follow cost

                // get related lane change point info of ego lane
                float egoLnDesLnChgPntDst = 0;
                float egoLnLatestLnChgPntDst = 0;
                #ifdef GLOG
                DLOG(INFO) << "---GlobalPathFllwCost_B---" << std::endl;
                DLOG(INFO) << "   stPntLaneID = " << int(stPntLaneID) << std::endl;
                DLOG(INFO) << "   endPntLaneID = " << int(endPntLaneID) << std::endl;
                DLOG(INFO) << "Global lane info array info" << std::endl;
                for (auto glbLnInfo : this->m_GlobalInfoContainer.m_GlobalLaneInfoArr)
                {
                    DLOG(INFO) << "global lane info id = " << int(glbLnInfo.m_LaneID) << std::endl;
                    if (!glbLnInfo.m_DesiredLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its des lc pnt = " << glbLnInfo.m_DesiredLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                    if (!glbLnInfo.m_LastLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its last lc pnt = " << glbLnInfo.m_LastLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                }
                #endif
                if (GetLnChgPnts(stPntLaneID, egoLnDesLnChgPntDst, egoLnLatestLnChgPntDst))
                {
                    if (egoLnDesLnChgPntDst > stPntDistAlongS)
                    {
                        // ego vehicle has overpassed the desired lane change point of current lane
                        // get related lane change point info of evaluated lane
                        float evlLnDesLnChgPntDst = 0;
                        float evlLnLatestLnChgPntDst = 0;
                        if (GetLnChgPnts(endPntLaneID, evlLnDesLnChgPntDst, evlLnLatestLnChgPntDst))
                        {
                            if (evlLnDesLnChgPntDst > stPntDistAlongS)
                            {
                                // ego vehicle has overpassed the desired lane change point of evaluated lane
                                // get distance btw ego vehicle position and desired lane change point of ego lane
                                float distOfEgoDesLCPnt = std::abs(egoLnDesLnChgPntDst - stPntDistAlongS);
                                // get distance btw ego vehicle position and desired lane change point of evaluated lane
                                float distOfEvlDesLCPnt = std::abs(evlLnDesLnChgPntDst - stPntDistAlongS);
                                if (distOfEgoDesLCPnt > distOfEvlDesLCPnt)
                                {
                                    this->m_GlobalPathFllwCost = 0.0;
                                }
                                else
                                {
                                    this->m_GlobalPathFllwCost = 1.0;
                                }
                            }
                            else
                            {
                                // ego vehicle has not overpassed the desired lane change point of evaluated lane
                                this->m_GlobalPathFllwCost = 0.0;
                            }
                        }
                        else
                        {
                            this->m_GlobalPathFllwCost = 0.0;
                            DLOG(WARNING) << "getting evaluated lane change point failed." << std::endl;
                        }
                    }
                    else
                    {
                        // ego vehicle has not overpassed the desired lane change point of current lane
                        this->m_GlobalPathFllwCost = 0.0;
                    }
                }
                else
                {
                    this->m_GlobalPathFllwCost = 0.0;
                    DLOG(INFO) << "getting ego lane change point failed." << std::endl;
                }
            }
            else
            {
                // it is a lane follow action
                // check whether ego vehicle has overpassed the desired lane change point of current lane
                // 1 if positive, set global path follow cost to 1
                // 2 if negative, set global path follow cost to 0
                // get related lane change point info of ego lane
                float egoLnDesLnChgPntDst = 0;
                float egoLnLatestLnChgPntDst = 0;
                if (GetLnChgPnts(stPntLaneID, egoLnDesLnChgPntDst, egoLnLatestLnChgPntDst))
                {
                    if (egoLnDesLnChgPntDst > stPntDistAlongS)
                    {
                        this->m_GlobalPathFllwCost = 1.0;
                    }
                    else
                    {
                        this->m_GlobalPathFllwCost = 0.0;
                    }
                }
                else
                {
                    this->m_GlobalPathFllwCost = 0.0;
                    DLOG(WARNING) << "getting ego lane change point failed." << std::endl;
                }
            }
        }
        else
        {
            this->m_GlobalPathFllwCost = 0.0;
            DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
        }
    }
}

void GlobalCostCalculation::CalcGlobalPathFllwCost_C()
{
    if (this->m_GlobalInfoContainer.m_GlobalInfoSrc == BA::GlobalInfoSrc_NoVldWayPoint)
    {
        this->m_GlobalPathFllwCost = 0;
        DLOG(WARNING) << "There is no specific global path to follow." << std::endl;
    }
    else
    {
        // revise m_DistanceAlongS' original point
        RevCrdSysOrgPnt(this->m_GlobalInfoContainer.m_DistToWayPoint);
        // get lane id and distance
        if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
        {
            uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
            uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
            // BA::BA_ActionTyp firstActionTyp = this->m_ActionTypArr.front();
            BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
            if (this->m_ActionTypArr.size() > 1)
            {
                firstActionTyp = this->m_ActionTypArr.at(1);
            }
            // if (stPntLaneID != endPntLaneID)
            // as action number increases, modify judgement condition from comparison  btw start ID and end ID TO identification of first action type
            if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange || firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
            {
                // it is a lane change action
                // check whether ego vehicle has overpassed the desired lane change point of current lane
                // 1 if positive, assign global path follow cost
                // 1.1 check whether ego vehicle has overpassed the desired lane change point of evaluated lane
                // 1.1.1 if positive, compare distance btw ego vehicle position and desired lane change points
                // 1.1.1.a if that of ego lane is greater than that of evaluated lane, then set global path follow cost of evaluated lane to 0
                // 1.1.1.b if that of ego lane is less than that of evaluated lane, then set global path follow cost of evaluated lane to 1
                // 1.1.2 if negative, then set global path follow cost of evaluated lane to 0
                // 2 if negative, do not assign global path follow cost

                // get related lane change point info of ego lane
                float egoLnDesLnChgPntDst = 0;
                float egoLnLatestLnChgPntDst = 0;
                if (GetLnChgPnts(stPntLaneID, egoLnDesLnChgPntDst, egoLnLatestLnChgPntDst))
                {
                    if (egoLnDesLnChgPntDst > stPntDistAlongS)
                    {
                        // ego vehicle has overpassed the desired lane change point of current lane
                        // get related lane change point info of evaluated lane
                        float evlLnDesLnChgPntDst = 0;
                        float evlLnLatestLnChgPntDst = 0;
                        if (GetLnChgPnts(endPntLaneID, evlLnDesLnChgPntDst, evlLnLatestLnChgPntDst))
                        {
                            if (evlLnDesLnChgPntDst > stPntDistAlongS)
                            {
                                // ego vehicle has overpassed the desired lane change point of evaluated lane
                                // get distance btw ego vehicle position and desired lane change point of ego lane
                                float distOfEgoDesLCPnt = std::abs(egoLnDesLnChgPntDst - stPntDistAlongS);
                                // get distance btw ego vehicle position and desired lane change point of evaluated lane
                                float distOfEvlDesLCPnt = std::abs(evlLnDesLnChgPntDst - stPntDistAlongS);
                                if (distOfEgoDesLCPnt > distOfEvlDesLCPnt)
                                {
                                    this->m_GlobalPathFllwCost = 0.0;
                                }
                                else
                                {
                                    this->m_GlobalPathFllwCost = 1.0;
                                }
                            }
                            else
                            {
                                // ego vehicle has not overpassed the desired lane change point of evaluated lane
                                this->m_GlobalPathFllwCost = 0.0;
                            }
                        }
                        else
                        {
                            this->m_GlobalPathFllwCost = 0.0;
                            DLOG(WARNING) << "getting evaluated lane change point failed." << std::endl;
                        }
                    }
                    else
                    {
                        // ego vehicle has not overpassed the desired lane change point of current lane
                        this->m_GlobalPathFllwCost = 0.0;
                    }
                }
                else
                {
                    this->m_GlobalPathFllwCost = 0.0;
                    DLOG(WARNING) << "getting ego lane change point failed." << std::endl;
                }
            }
            else
            {
                // it is a lane follow action
                // check whether ego vehicle has overpassed the desired lane change point of current lane
                // 1 if positive, set global path follow cost to 1
                // 2 if negative, set global path follow cost to 0
                // get related lane change point info of ego lane
                float egoLnDesLnChgPntDst = 0;
                float egoLnLatestLnChgPntDst = 0;
                if (GetLnChgPnts(stPntLaneID, egoLnDesLnChgPntDst, egoLnLatestLnChgPntDst))
                {
                    if (egoLnDesLnChgPntDst > stPntDistAlongS)
                    {
                        this->m_GlobalPathFllwCost = 1.0;
                    }
                    else
                    {
                        this->m_GlobalPathFllwCost = 0.0;
                    }
                }
                else
                {
                    this->m_GlobalPathFllwCost = 0.0;
                    DLOG(WARNING) << "getting ego lane change point failed." << std::endl;
                }
            }
        }
        else
        {
            this->m_GlobalPathFllwCost = 0.0;
            DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
        }
    }
}

void GlobalCostCalculation::CalcGlobalPathFllwCost_InclLCP()
{
    if (this->m_GlobalInfoContainer.m_GlobalInfoSrc == BA::GlobalInfoSrc_NoVldWayPoint)
    {
        this->m_GlobalPathFllwCost = 0;
        DLOG(WARNING) << "There is no specific global path to follow." << std::endl;
    }
    else
    {
        // revise m_DistanceAlongS' original point
        RevCrdSysOrgPnt(this->m_GlobalInfoContainer.m_DistToWayPoint);
        // get lane id and distance
        if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
        {
            uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
            // uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            uint8_t endPntLaneID = uint8_t(0);
            float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
            // BA::BA_ActionTyp firstActionTyp = this->m_ActionTypArr.front();
            BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
            if (this->m_ActionTypArr.size() > 1)
            {
                firstActionTyp = this->m_ActionTypArr.at(1);
            }
            // get virtual end point lane id by action
            if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare)
            {
                auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                               [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                               {
                                                   return (stPntLaneID == wpExtLn.lane_id);
                                               });
                if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                {
                    if (!it_wpExtLn->left_extend_lane.empty())
                    {
                        for (auto leftExtLn : it_wpExtLn->left_extend_lane)
                        {
                            uint8_t tmpExtLnID = leftExtLn.adjacent_extendlane_id;
                            auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                              [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                              {
                                                                  return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                              });
                            uint8_t isEmrgLn = uint8_t(0);
                            if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                            {
                                isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                            }
                            // emergency lane should not be included, as emergency lane lack lane change point info                            
                            if (leftExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                            {
                                endPntLaneID = leftExtLn.adjacent_extendlane_id;
                                break;
                            }
                        }
                    }
                }
            }
            else if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange)
            {
                endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
                #ifdef GLOG
                DLOG(INFO) << "---GlobalPathFllwCost_InclLCP---" << std::endl;
                DLOG(INFO) << "   stPntLaneID = " << int(stPntLaneID) << std::endl;
                DLOG(INFO) << "   endPntLaneID = " << int(endPntLaneID) << std::endl;
                DLOG(INFO) << "Global lane info array info" << std::endl;
                for (auto glbLnInfo : this->m_GlobalInfoContainer.m_GlobalLaneInfoArr)
                {
                    DLOG(INFO) << "global lane info id = " << int(glbLnInfo.m_LaneID) << std::endl;
                    if (!glbLnInfo.m_DesiredLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its des lc pnt = " << glbLnInfo.m_DesiredLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                    if (!glbLnInfo.m_LastLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its last lc pnt = " << glbLnInfo.m_LastLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                }
                #endif
            }
            else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare)
            {
                auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                               [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                               {
                                                   return (stPntLaneID == wpExtLn.lane_id);
                                               });
                if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                {
                    if (!it_wpExtLn->right_extend_lane.empty())
                    {
                        for (auto rightExtLn : it_wpExtLn->right_extend_lane)
                        {
                            uint8_t tmpExtLnID = rightExtLn.adjacent_extendlane_id;
                            auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                              [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                              {
                                                                  return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                              });
                            uint8_t isEmrgLn = uint8_t(0);
                            if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                            {
                                isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                            }
                            // emergency lane should not be included, as emergency lane lack lane change point info
                            if (rightExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                            {
                                endPntLaneID = rightExtLn.adjacent_extendlane_id;
                                break;
                            }
                        }
                    }
                }
                #ifdef GLOG
                DLOG(INFO) << "---GlobalPathFllwCost_InclLCP---" << std::endl;
                DLOG(INFO) << "   stPntLaneID = " << int(stPntLaneID) << std::endl;
                DLOG(INFO) << "   endPntLaneID = " << int(endPntLaneID) << std::endl;
                DLOG(INFO) << "Global lane info array info" << std::endl;
                for (auto glbLnInfo : this->m_GlobalInfoContainer.m_GlobalLaneInfoArr)
                {
                    DLOG(INFO) << "global lane info id = " << int(glbLnInfo.m_LaneID) << std::endl;
                    if (!glbLnInfo.m_DesiredLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its des lc pnt = " << glbLnInfo.m_DesiredLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                    if (!glbLnInfo.m_LastLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its last lc pnt = " << glbLnInfo.m_LastLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                }
                #endif
            }
            else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
            {
                endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
                #ifdef GLOG
                DLOG(INFO) << "---GlobalPathFllwCost_InclLCP---" << std::endl;
                DLOG(INFO) << "   stPntLaneID = " << int(stPntLaneID) << std::endl;
                DLOG(INFO) << "   endPntLaneID = " << int(endPntLaneID) << std::endl;
                DLOG(INFO) << "Global lane info array info" << std::endl;
                for (auto glbLnInfo : this->m_GlobalInfoContainer.m_GlobalLaneInfoArr)
                {
                    DLOG(INFO) << "global lane info id = " << int(glbLnInfo.m_LaneID) << std::endl;
                    if (!glbLnInfo.m_DesiredLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its des lc pnt = " << glbLnInfo.m_DesiredLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                    if (!glbLnInfo.m_LastLaneChangePoint.empty())
                    {
                        DLOG(INFO) << "  its last lc pnt = " << glbLnInfo.m_LastLaneChangePoint.at(0) << std::endl;
                    }
                    else
                    {
                        DLOG(INFO) << "Deseired lane change point is empty" << std::endl;
                    }
                }
                #endif
            }
            else if (firstActionTyp == BA::BA_ActionTyp_LaneFollow)
            {
                endPntLaneID = stPntLaneID;
            }
            if (endPntLaneID == uint8_t(0))
            {
                // set endPntLaneID to stPntLaneID
                endPntLaneID = stPntLaneID;
            }
            DLOG(INFO) << "---firstActionTyp = " << int(firstActionTyp) << std::endl;
            DLOG(INFO) << "---endPntLaneID = " << int(endPntLaneID) << std::endl;
            // if (stPntLaneID != endPntLaneID)
            // as action number increases, modify judgement condition from comparison  btw start ID and end ID TO identification of first action type
            if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange || 
                firstActionTyp == BA::BA_ActionTyp_RightLaneChange ||
                firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare ||
                firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare)
            {
                // it is a lane change action
                // check whether ego vehicle has overpassed the desired lane change point of current lane
                // 1 if positive, assign global path follow cost
                // 1.1 check whether ego vehicle has overpassed the desired lane change point of evaluated lane
                // 1.1.1 if positive, compare distance btw ego vehicle position and desired lane change points
                // 1.1.1.a if that of ego lane is greater than that of evaluated lane, then set global path follow cost of evaluated lane to 0
                // 1.1.1.b if that of ego lane is less than that of evaluated lane, then set global path follow cost of evaluated lane to 1
                // 1.1.2 if negative, then set global path follow cost of evaluated lane to 0
                // 2 if negative, do not assign global path follow cost

                // get related lane change point info of ego lane
                float egoLnDesLnChgPntDst = 0;
                float egoLnLatestLnChgPntDst = 0;
                if (GetLnChgPnts(stPntLaneID, egoLnDesLnChgPntDst, egoLnLatestLnChgPntDst))
                {
                    if (egoLnDesLnChgPntDst > stPntDistAlongS)
                    {
                        // ego vehicle has overpassed the desired lane change point of current lane
                        // get related lane change point info of evaluated lane
                        float evlLnDesLnChgPntDst = 0;
                        float evlLnLatestLnChgPntDst = 0;
                        if (GetLnChgPnts(endPntLaneID, evlLnDesLnChgPntDst, evlLnLatestLnChgPntDst))
                        {
                            if (evlLnDesLnChgPntDst > stPntDistAlongS)
                            {
                                // ego vehicle has overpassed the desired lane change point of evaluated lane
                                // get distance btw ego vehicle position and desired lane change point of ego lane
                                float distOfEgoDesLCPnt = std::abs(egoLnDesLnChgPntDst - stPntDistAlongS);
                                // get distance btw ego vehicle position and desired lane change point of evaluated lane
                                float distOfEvlDesLCPnt = std::abs(evlLnDesLnChgPntDst - stPntDistAlongS);
                                if (distOfEgoDesLCPnt > distOfEvlDesLCPnt)
                                {
                                    this->m_GlobalPathFllwCost = 0.0;
                                }
                                else
                                {
                                    this->m_GlobalPathFllwCost = 1.0;
                                }
                            }
                            else
                            {
                                // ego vehicle has not overpassed the desired lane change point of evaluated lane
                                this->m_GlobalPathFllwCost = 0.0;
                            }
                        }
                        else
                        {
                            this->m_GlobalPathFllwCost = 0.0;
                            DLOG(WARNING) << "getting evaluated lane change point failed." << std::endl;
                        }
                    }
                    else
                    {
                        // ego vehicle has not overpassed the desired lane change point of current lane
                        this->m_GlobalPathFllwCost = 0.0;
                    }
                }
                else
                {
                    this->m_GlobalPathFllwCost = 0.0;
                    DLOG(WARNING) << "getting ego lane change point failed." << std::endl;
                }
            }
            else
            {
                // it is a lane follow action
                // check whether ego vehicle has overpassed the desired lane change point of current lane
                // 1 if positive, set global path follow cost to 1
                // 2 if negative, set global path follow cost to 0
                // get related lane change point info of ego lane
                float egoLnDesLnChgPntDst = 0;
                float egoLnLatestLnChgPntDst = 0;
                if (GetLnChgPnts(stPntLaneID, egoLnDesLnChgPntDst, egoLnLatestLnChgPntDst))
                {
                    if (egoLnDesLnChgPntDst > stPntDistAlongS)
                    {
                        this->m_GlobalPathFllwCost = 1.0;
                    }
                    else
                    {
                        this->m_GlobalPathFllwCost = 0.0;
                    }
                }
                else
                {
                    this->m_GlobalPathFllwCost = 0.0;
                    DLOG(WARNING) << "getting ego lane change point failed." << std::endl;
                }
            }
        }
        else
        {
            this->m_GlobalPathFllwCost = 0.0;
            DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
        }
    }
}

bool GlobalCostCalculation::GetLnChgPnts(uint8_t laneID, float &desLnChgPntDst, float &latestLnChgPntDst)
{
    if (!this->m_GlobalInfoContainer.m_GlobalLaneInfoArr.empty())
    {
        std::vector<BA::GlobalLaneInfo>::iterator iter = this->m_GlobalInfoContainer.m_GlobalLaneInfoArr.begin();
        // find the globalLaneInfo with target lane ID, then get desired and latest lane change points
        for (; iter != this->m_GlobalInfoContainer.m_GlobalLaneInfoArr.end(); iter++)
        {
            if (iter->m_LaneID == laneID)
            {
                if (!iter->m_DesiredLaneChangePoint.empty())
                {
                    desLnChgPntDst = iter->m_DesiredLaneChangePoint[0]; // TODO: 0 or 1
                }
                else
                {
                    desLnChgPntDst = 0.0; // if m_DesiredLaneChangePoint is empty, then set distance between waypoint and desired lane change point to 0m
                }
                if (!iter->m_LastLaneChangePoint.empty())
                {
                    latestLnChgPntDst = iter->m_LastLaneChangePoint[0];
                }
                else
                {
                    latestLnChgPntDst = 0; // if m_LatestLaneChangePoint is empty, then set distance between waypoint and latest lane change point to 0
                }
                break;
            }
        }
        if (iter == this->m_GlobalInfoContainer.m_GlobalLaneInfoArr.end())
        {
            DLOG(WARNING) << "target lane ID does not exist in GlobalLaneInfoArr!" << std::endl;
            return false;
        }
    }
    else
    {
        DLOG(WARNING) << "m_GlobalLaneInfoArr is empty!" << std::endl;
        return false;
    }
    return true;
}

/* void GlobalCostCalculation::CalcLgtTermFastLnCost(const uint8_t &fastLnID)
{
    // calc sample points which are located on the most left side and velocity is lower than 25% of road speed limit
    std::list<float> tmpVelLst(this->m_HistHostVehInfo.m_Velocity);
    std::list<uint8_t> tmpLnLocLst(this->m_HistHostVehInfo.m_LaneLocation);
    reverse(tmpVelLst.begin(), tmpVelLst.end());
    reverse(tmpLnLocLst.begin(), tmpLnLocLst.end());
    // from back to front, check each sample points, whether long-term in fast lane
    int histCnt = 0;
    int curCnt = 0;
    int sumCnt = 0;
    std::list<uint8_t>::iterator it_LnLoc_Past = tmpLnLocLst.begin();
    std::list<float>::iterator it_Vel_Past = tmpVelLst.begin();
    for (; it_LnLoc_Past != tmpLnLocLst.end();)
    {
        if (*it_LnLoc_Past == fastLnID && *it_Vel_Past < this->m_RoadSpdLmt * (1 - ADC::common::k_BA_VelLowerPerc))
        {
            histCnt++;
        }
        else
        {
            break;
        }
        it_LnLoc_Past++;
        it_Vel_Past++;
    }
    if (histCnt >= ADC::common::k_BA_LgTermFastLnHistSmplPntThrd)
    {
        // continuous positive sample points number in the past is over preset threshold
        // calc continuous positive sample points number in the future
        std::vector<uint8_t>::iterator it_LnLoc_fut = this->m_CurHostVehInfo.m_LaneLocation.begin();
        std::vector<float>::iterator it_Vel_fut = this->m_CurHostVehInfo.m_Velocity.begin();
        for (; it_LnLoc_fut != this->m_CurHostVehInfo.m_LaneLocation.end();)
        {
            if (*it_LnLoc_fut == fastLnID && *it_Vel_fut < this->m_RoadSpdLmt * (1 - ADC::common::k_BA_VelLowerPerc))
            {
                curCnt++;
            }
            else
            {
                break;
            }
            it_LnLoc_fut++;
            it_Vel_fut++;
        }
        sumCnt = histCnt + curCnt;
        if (sumCnt >= ADC::common::k_BA_LgTermFastLnSumSmplPntNum)
        {
            this->m_LgTermFastLnCost = 1;
        }
        else
        {
            this->m_LgTermFastLnCost = (sumCnt - ADC::common::k_BA_LgTermFastLnHistSmplPntThrd) / (ADC::common::k_BA_LgTermFastLnSumSmplPntNum - ADC::common::k_BA_LgTermFastLnHistSmplPntThrd);
        }
    }
    else
    {
        this->m_LgTermFastLnCost = 0;
    }
} */

void GlobalCostCalculation::CalcLgtTermFastLnCost_B(const uint8_t &fastLnID, const RoadInfo &rm_msg)
{
    // calc sample points which are located on the most left side and velocity is lower than 30% of road speed limit
    // this->m_RoadSpdLmt = 100.0 / ADC::common::c_BA_ms2kmh; // road speed limit set to 100.0 temperarily
    // 20220727: get fast lane max speed limit

    std::list<float> tmpVelLst(this->m_HistHostVehInfo.m_Velocity);
    std::list<uint8_t> tmpLnLocLst(this->m_HistHostVehInfo.m_LaneLocation);
    reverse(tmpVelLst.begin(), tmpVelLst.end());
    reverse(tmpLnLocLst.begin(), tmpLnLocLst.end());
    // from back to front, check each sample points, whether long-term in fast lane
    int histCnt = 0;

    std::list<uint8_t>::iterator it_LnLoc_Past = tmpLnLocLst.begin();
    std::list<float>::iterator it_Vel_Past = tmpVelLst.begin();
    DLOG(INFO) << "m_RoadSpdLmt = " << m_RoadSpdLmt << std::endl;
    for (; it_LnLoc_Past != tmpLnLocLst.end();)
    {
        if (*it_LnLoc_Past == fastLnID && *it_Vel_Past < this->m_RoadSpdLmt / ADC::common::c_BA_ms2kmh * (1 - ADC::common::k_BA_VelLowerPerc))
        {
            histCnt++;
        }
        else
        {
            break;
        }
        it_LnLoc_Past++;
        it_Vel_Past++;
    }
    DLOG(INFO) << "fast lane hist cnt = " << histCnt << std::endl;
    if (histCnt >= ADC::common::k_BA_LgTermFastLnHistSmplPntThrd)
    {
        if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
        {
            uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
            uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            // float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
            if (stPntLaneID != endPntLaneID)
            {
                this->m_LgTermFastLnCost = 0;
            }
            else
            {
                this->m_LgTermFastLnCost = 1;
            }
        }
        else
        {
            this->m_LgTermFastLnCost = 0;
            DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
        }
    }
    else
    {
        this->m_LgTermFastLnCost = 0;
    }
}

bool GlobalCostCalculation::UpdtHistVehInfo(const CarPosition &carPos, const float &vehSpd, const bool &isHistVehInfoUpdt)
{
    static BA::HV_HistInfo histInfo;
    if (isHistVehInfoUpdt)
    {
        bool isTimeToSmplPnt = CalcSmplPntTimeStamp();
        if (isTimeToSmplPnt)
        {
            if (!histInfo.m_LaneLocation.empty())
            {
                if (histInfo.m_LaneLocation.size() == ADC::common::k_BA_HistVehInfoLstSize)
                {
                    // pop oldest vehicle info is needed
                    histInfo.m_DistanceAlongS.pop_front();
                    histInfo.m_LaneLocation.pop_front();
                    histInfo.m_Velocity.pop_front();
                }
                // push latest vehicle info inside
                histInfo.m_LaneLocation.push_back(carPos.ego_lane_id);
                histInfo.m_Velocity.push_back(vehSpd); // unit: m/s
                // renew m_DistanceAlongS, value is negative
                UpdtHistVehInfoDist(vehSpd, histInfo);
                histInfo.m_DistanceAlongS.push_back(0.0); // m
            }
            else
            {
                // push element the first time
                histInfo.m_LaneLocation.push_back(carPos.ego_lane_id);
                histInfo.m_Velocity.push_back(vehSpd);
                histInfo.m_DistanceAlongS.push_back(0.0);
            }
        }
    }
    this->m_HistHostVehInfo = histInfo;
    return true;
}

void GlobalCostCalculation::UpdtHistVehInfoDist(float vehSpd, BA::HV_HistInfo &histInfo)
{
    std::list<float>::iterator it = histInfo.m_DistanceAlongS.begin();
    for (; it != histInfo.m_DistanceAlongS.end(); it++)
    {
        *it -= vehSpd * ADC::common::k_BA_HistVehInfoSmplCycle;
    }
}

bool GlobalCostCalculation::CalcSmplPntTimeStamp()
{
    // this->m_CycleTimer += BEHAVIORANALYZERCYCLE / 1000.0;
    static float cycleTimer = 0.0;
    if (cycleTimer >= ADC::common::k_BA_HistVehInfoSmplCycle)
    {
        // it is time to operate the HistHostVehInfo
        cycleTimer = 0.0; // reset cycle timer
        return true;
    }
    cycleTimer += BEHAVIORANALYZERCYCLE / 1000.0;
    return false;
}

void GlobalCostCalculation::CalcLgtTermSlowLnCost(const uint8_t &slowLnID)
{
    // calc sample points which are located on the most right side
    std::list<uint8_t> tmpLnLocLst(this->m_HistHostVehInfo.m_LaneLocation);
    reverse(tmpLnLocLst.begin(), tmpLnLocLst.end());
    // from back to front, check each sample points, whether long-term in the slowest lane
    int histCnt = 0;
    int curCnt = 0;
    int sumCnt = 0;
    std::list<uint8_t>::iterator it_LnLoc_Past = tmpLnLocLst.begin();
    for (; it_LnLoc_Past != tmpLnLocLst.end();)
    {
        if (*it_LnLoc_Past == slowLnID)
        {
            histCnt++;
        }
        else
        {
            break;
        }
        it_LnLoc_Past++;
    }
    if (histCnt >= ADC::common::k_BA_LgTermSlowLnHistSmplPntThrd)
    {
        // continuous positive sample points number in the past is over preset threshold
        // calc continuous positive sample points number in the future
        std::vector<uint8_t>::iterator it_LnLoc_fut = this->m_CurHostVehInfo.m_LaneLocation.begin();
        for (; it_LnLoc_fut != this->m_CurHostVehInfo.m_LaneLocation.end();)
        {
            if (*it_LnLoc_fut == slowLnID)
            {
                curCnt++;
            }
            else
            {
                break;
            }
            it_LnLoc_fut++;
        }
        sumCnt = histCnt + curCnt;
        if (sumCnt >= ADC::common::k_BA_LgTermSlowLnSumSmplPntNum)
        {
            this->m_LgTermSlowLnCost = 1;
        }
        else
        {
            this->m_LgTermSlowLnCost = (sumCnt - ADC::common::k_BA_LgTermSlowLnHistSmplPntThrd) / (ADC::common::k_BA_LgTermSlowLnSumSmplPntNum - ADC::common::k_BA_LgTermSlowLnHistSmplPntThrd);
        }
    }
    else
    {
        this->m_LgTermSlowLnCost = 0;
    }
}

void GlobalCostCalculation::CalcLgtTermSlowLnCost_B(const uint8_t &slowLnID)
{
    // calc sample points which are located on the most right side
    std::list<uint8_t> tmpLnLocLst(this->m_HistHostVehInfo.m_LaneLocation);
    reverse(tmpLnLocLst.begin(), tmpLnLocLst.end());
    // from back to front, check each sample points, whether long-term in the slowest lane
    int histCnt = 0;
    std::list<uint8_t>::iterator it_LnLoc_Past = tmpLnLocLst.begin();
    for (; it_LnLoc_Past != tmpLnLocLst.end();)
    {
        if (*it_LnLoc_Past == slowLnID)
        {
            histCnt++;
        }
        else
        {
            break;
        }
        it_LnLoc_Past++;
    }
    DLOG(INFO) << "histCnt = " << histCnt << std::endl;
    if (histCnt >= ADC::common::k_BA_LgTermSlowLnHistSmplPntThrd)
    {
        // if ego car has been on the most right side lane for 20 sec, then
        //    1- if lane follow next, the m_LgTermSlowLnCost will be 1
        //    2- if lane change next, the m_LgTermSlowLnCost will be 0
        // namely drive ego car to leave the most right side lane ASAP
        if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
        {
            uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
            uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            // float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
            if (stPntLaneID != endPntLaneID)
            {
                this->m_LgTermSlowLnCost = 0;
            }
            else
            {
                this->m_LgTermSlowLnCost = 1;
            }
        }
        else
        {
            DLOG(WARNING) << "m_CurHostVehInfo is empty" << std::endl;
            this->m_LgTermSlowLnCost = 0;
        }
    }
    else
    {
        this->m_LgTermSlowLnCost = 0;
    }
}

void GlobalCostCalculation::CalcLgtTermSlowLnCost_InclLCP(const uint8_t &slowLnID, const ControlWaypoints &ehr_msg)
{
    // calc sample points which are located on the most right side
    std::list<uint8_t> tmpLnLocLst(this->m_HistHostVehInfo.m_LaneLocation);
    reverse(tmpLnLocLst.begin(), tmpLnLocLst.end());
    // from back to front, check each sample points, whether long-term in the slowest lane
    int histCnt = 0;
    std::list<uint8_t>::iterator it_LnLoc_Past = tmpLnLocLst.begin();
    for (; it_LnLoc_Past != tmpLnLocLst.end();)
    {
        if (*it_LnLoc_Past == slowLnID)
        {
            histCnt++;
        }
        else
        {
            break;
        }
        it_LnLoc_Past++;
    }
    DLOG(INFO) << "histCnt = " << histCnt << std::endl;
    if (histCnt >= ADC::common::k_BA_LgTermSlowLnHistSmplPntThrd)
    {
        // if ego car has been on the most right side lane for 20 sec, then
        //    1- if lane follow next, the m_LgTermSlowLnCost will be 1
        //    2- if lane change next, the m_LgTermSlowLnCost will be 0
        // namely drive ego car to leave the most right side lane ASAP
        if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
        {
            uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
            uint8_t endPntLaneID = uint8_t(0);
            float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
            BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
            if (this->m_ActionTypArr.size() > 1)
            {
                firstActionTyp = this->m_ActionTypArr.at(1);
            }
            // get virtual end point lane id by action
            if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare)
            {
                auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                               [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                               {
                                                   return (stPntLaneID == wpExtLn.lane_id);
                                               });
                if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                {
                    if (!it_wpExtLn->left_extend_lane.empty())
                    {
                        for (auto leftExtLn : it_wpExtLn->left_extend_lane)
                        {
                            uint8_t tmpExtLnID = leftExtLn.adjacent_extendlane_id;
                            auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                              [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                              {
                                                                  return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                              });
                            uint8_t isEmrgLn = uint8_t(0);
                            if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                            {
                                isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                            }
                            // emergency lane should not be included, as emergency lane lack lane change point info                            
                            if (leftExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                            {
                                endPntLaneID = leftExtLn.adjacent_extendlane_id;
                                break;
                            }
                        }
                    }
                }
            }
            else if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange)
            {
                endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            }
            else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare)
            {
                auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                               [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                               {
                                                   return (stPntLaneID == wpExtLn.lane_id);
                                               });
                if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                {
                    if (!it_wpExtLn->right_extend_lane.empty())
                    {
                        for (auto rightExtLn : it_wpExtLn->right_extend_lane)
                        {
                            uint8_t tmpExtLnID = rightExtLn.adjacent_extendlane_id;
                            auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                              [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                              {
                                                                  return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                              });
                            uint8_t isEmrgLn = uint8_t(0);
                            if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                            {
                                isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                            }
                            // emergency lane should not be included, as emergency lane lack lane change point info
                            if (rightExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                            {
                                endPntLaneID = rightExtLn.adjacent_extendlane_id;
                                break;
                            }
                        }
                    }
                }
            }
            else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
            {
                endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
            }
            else if (firstActionTyp == BA::BA_ActionTyp_LaneFollow)
            {
                endPntLaneID = stPntLaneID;
            }
            if (endPntLaneID == uint8_t(0))
            {
                // set endPntLaneID to stPntLaneID
                endPntLaneID = stPntLaneID;
            }
            // 20221031
            // if dist to subMrgRd is less than 500m(TBD)
            // then add leftLaneChangePrepare into consideration
            bool inclLCPAction = false;
            if (ADC::common::k_BA_InclLCP4SlowLnCost == true && !ehr_msg.reminder_waypoint.empty() && !ehr_msg.control_waypoints_list.empty())
            {
                if (ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubMrgRd &&
                    ehr_msg.control_waypoints_list.front().control_waypoint_target_road_type == BA::TargetRoadTyp_VirtFixDstRd)
                {
                    if (ehr_msg.reminder_waypoint.front().reminder_waypoint_offset < 600.0f)
                    {
                        inclLCPAction = true;
                    }
                }
            }
            if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange || 
                (firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare && inclLCPAction == true))
            {
                if (stPntLaneID != endPntLaneID)
                {
                    this->m_LgTermSlowLnCost = 0;
                }
                else
                {
                    this->m_LgTermSlowLnCost = 1;
                }
            }
            else
            {
                this->m_LgTermSlowLnCost = 1;
            }
        }
        else
        {
            DLOG(WARNING) << "m_CurHostVehInfo is empty" << std::endl;
            this->m_LgTermSlowLnCost = 0;
        }
    }
    else
    {
        this->m_LgTermSlowLnCost = 0;
    }    
}

void GlobalCostCalculation::CalcLnEndCost(const ControlWaypoints &ehr_msg,
                                          const CarPosition &carPos,
                                          const float &vehSpd)
{
    if (!this->m_CurHostVehInfo.m_LaneLocation.empty())
    {
        float k = 1.0 / this->m_CurHostVehInfo.m_LaneLocation.size(); // all sample points share the same weight coefficiency;
        float remainDistToLnEndPnt = 0;
        for (int i = 0; i < this->m_CurHostVehInfo.m_DistanceAlongS.size(); i++)
        {
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, this->m_CurHostVehInfo.m_LaneLocation.at(i), this->m_CurHostVehInfo.m_DistanceAlongS.at(i), vehSpd, remainDistToLnEndPnt))
            {
                this->m_LaneEndCost += 1.0 / k;
            }
        }
    }
    else
    {
        DLOG(WARNING) << "m_CurHostVehInfo.m_LaneLocation is empty!" << std::endl;
    }
}

void GlobalCostCalculation::CalcLnEndCost_B(const ControlWaypoints &ehr_msg,
                                            const CarPosition &carPos,
                                            const float &vehSpd)
{
    this->m_LaneEndCost = 0.0;
    if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
    {
        uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
        uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
        // 20220426: transform stPntDistAlongS
        stPntDistAlongS = this->m_GlobalInfoContainer.m_DistToWayPoint - stPntDistAlongS;
        if (stPntLaneID != endPntLaneID)
        {
            // it is a lane change action
            // check whether ego vehicle has been close enough to the lane end point of ego lane
            // 1 if positive, assign lane end cost
            // 1.1 check whether ego vehicle has been close enough to the lane end point of evaluated lane
            // 1.1.1 if positive, compare distance btw ego vehicle position and lane end points
            // 1.1.1.a if that of ego lane is greater than that of evaluated lane, then set lane end cost of evaluated lane to 1
            // 1.1.1.b if that of ego lane is less than that of evaluated lane, then set lane end cost of evaluated lane to 0
            // 1.1.2 if negative, then set lane end cost of evaluated lane to 0
            // 2 if negative, do not assign lane end cost
            float remainDistToEgoLnEndPnt = 0.0f;
            // DLOG(INFO) << "stPntDistAlongS = " << stPntDistAlongS << std::endl;
            // DLOG(INFO) << "stPntLaneID = " << int(stPntLaneID) << std::endl;
            // DLOG(INFO) << "endPntLaneID = " << int(endPntLaneID) << std::endl;
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, stPntLaneID, stPntDistAlongS, vehSpd, remainDistToEgoLnEndPnt))
            {
                // DLOG(INFO) << "remainDistToEgoLnEndPnt = " << remainDistToEgoLnEndPnt << std::endl;
                float remainDistToEvlLnEndPnt = 0.0f;
                if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, endPntLaneID, stPntDistAlongS, vehSpd, remainDistToEvlLnEndPnt))
                {
                    // DLOG(INFO) << "remainDistToEvlLnEndPnt = " << remainDistToEvlLnEndPnt << std::endl;
                    if (remainDistToEgoLnEndPnt > remainDistToEvlLnEndPnt)
                    {
                        this->m_LaneEndCost = 1.0;
                    }
                    else
                    {
                        this->m_LaneEndCost = 0.0;
                    }
                }
            }
        }
        else
        {
            // it is a lane follow action
            // check whether ego vehicle has been close enough to the lane end point of ego lane
            // 1. if positive, then set lane end cost to 1
            // 2. if negative, then set lane end cost to 0
            float remainDistToEgoLnEndPnt = 0.0;
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, stPntLaneID, stPntDistAlongS, vehSpd, remainDistToEgoLnEndPnt))
            {
                this->m_LaneEndCost = 1.0;
            }
        }
    }
    else
    {
        DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
    }
}

void GlobalCostCalculation::CalcLnEndCost_C(const ControlWaypoints &ehr_msg,
                                            const CarPosition &carPos,
                                            const float &vehSpd)
{
    this->m_LaneEndCost = 0.0;
    if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
    {
        uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
        uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
        // 20220426: transform stPntDistAlongS
        stPntDistAlongS = this->m_GlobalInfoContainer.m_DistToWayPoint - stPntDistAlongS;
        // BA::BA_ActionTyp firstActionTyp = this->m_ActionTypArr.front();
        BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
        if (this->m_ActionTypArr.size() > 1)
        {
            firstActionTyp = this->m_ActionTypArr.at(1);
        }
        // if (stPntLaneID != endPntLaneID)
        // as action number increases, modify judgement condition from comparison  btw start ID and end ID TO identification of first action type
        if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange || firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
        {
            // it is a lane change action
            // check whether ego vehicle has been close enough to the lane end point of ego lane
            // 1 if positive, assign lane end cost
            // 1.1 check whether ego vehicle has been close enough to the lane end point of evaluated lane
            // 1.1.1 if positive, compare distance btw ego vehicle position and lane end points
            // 1.1.1.a if that of ego lane is greater than that of evaluated lane, then set lane end cost of evaluated lane to 1
            // 1.1.1.b if that of ego lane is less than that of evaluated lane, then set lane end cost of evaluated lane to 0
            // 1.1.2 if negative, then set lane end cost of evaluated lane to 0
            // 2 if negative, do not assign lane end cost
            float remainDistToEgoLnEndPnt = 0.0;
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, stPntLaneID, stPntDistAlongS, vehSpd, remainDistToEgoLnEndPnt))
            {
                float remainDistToEvlLnEndPnt = 0.0;
                if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, endPntLaneID, stPntDistAlongS, vehSpd, remainDistToEvlLnEndPnt))
                {
                    if (remainDistToEgoLnEndPnt > remainDistToEvlLnEndPnt)
                    {
                        this->m_LaneEndCost = 1.0;
                    }
                    else
                    {
                        this->m_LaneEndCost = 0.0;
                    }
                }
            }
        }
        else
        {
            // it is a lane follow action
            // check whether ego vehicle has been close enough to the lane end point of ego lane
            // 1. if positive, then set lane end cost to 1
            // 2. if negative, then set lane end cost to 0
            float remainDistToEgoLnEndPnt = 0.0;
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, stPntLaneID, stPntDistAlongS, vehSpd, remainDistToEgoLnEndPnt))
            {
                this->m_LaneEndCost = 1.0;
            }
        }
    }
    else
    {
        DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
    }
}

void GlobalCostCalculation::CalcLnEndCost_InclLCP(const ControlWaypoints &ehr_msg,
                                                  const CarPosition &carPos,
                                                  const float &vehSpd)
{
    this->m_LaneEndCost = 0.0;
    if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
    {
        uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
        uint8_t endPntLaneID = uint8_t(0);
        // uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
        // 20220426: transform stPntDistAlongS
        stPntDistAlongS = this->m_GlobalInfoContainer.m_DistToWayPoint - stPntDistAlongS;
        // BA::BA_ActionTyp firstActionTyp = this->m_ActionTypArr.front();
        BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
        if (this->m_ActionTypArr.size() > 1)
        {
            firstActionTyp = this->m_ActionTypArr.at(1);
        }
        // get virtual end point lane id by action
        if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare)
        {
            auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                           [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                           {
                                               return (stPntLaneID == wpExtLn.lane_id);
                                           });
            if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
            {
                if (!it_wpExtLn->left_extend_lane.empty())
                {
                    for (auto leftExtLn : it_wpExtLn->left_extend_lane)
                    {
                        uint8_t tmpExtLnID = leftExtLn.adjacent_extendlane_id;
                        auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                          [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                          {
                                                              return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                          });
                        uint8_t isEmrgLn = uint8_t(0);
                        if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                        {
                            isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                        }
                        // emergency lane should not be included, as emergency lane lack lane change point info
                        if (leftExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                        {
                            endPntLaneID = leftExtLn.adjacent_extendlane_id;
                            break;
                        }
                    }
                }
            }
        }
        else if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange)
        {
            endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        }
        else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare)
        {
            auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                           [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                           {
                                              return (stPntLaneID == wpExtLn.lane_id);
                                           });
            if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
            {
                if (!it_wpExtLn->right_extend_lane.empty())
                {
                    for (auto rightExtLn : it_wpExtLn->right_extend_lane)
                    {
                        uint8_t tmpExtLnID = rightExtLn.adjacent_extendlane_id;
                        auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                          [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                          {
                                                              return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                          });
                        uint8_t isEmrgLn = uint8_t(0);
                        if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                        {
                            isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                        }
                        // emergency lane should not be included, as emergency lane lack lane change point info
                        if (rightExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                        {
                            endPntLaneID = rightExtLn.adjacent_extendlane_id;
                            break;
                        }
                    }
                }
            }
        }
        else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
        {
            endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        }
        else if (firstActionTyp == BA::BA_ActionTyp_LaneFollow)
        {
            endPntLaneID = stPntLaneID;
        }
        if (endPntLaneID == uint8_t(0))
        {
            // if endPntLaneID is not assigned, then set endPntLaneID to stPntLaneID
            endPntLaneID = stPntLaneID;
        }
        // if (stPntLaneID != endPntLaneID)
        // as action number increases, modify judgement condition from comparison  btw start ID and end ID TO identification of first action type
        if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange || 
            firstActionTyp == BA::BA_ActionTyp_RightLaneChange ||
            firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare ||
            firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare)
        {
            // it is a lane change action
            // check whether ego vehicle has been close enough to the lane end point of ego lane
            // 1 if positive, assign lane end cost
            // 1.1 check whether ego vehicle has been close enough to the lane end point of evaluated lane
            // 1.1.1 if positive, compare distance btw ego vehicle position and lane end points
            // 1.1.1.a if that of ego lane is greater than that of evaluated lane, then set lane end cost of evaluated lane to 1
            // 1.1.1.b if that of ego lane is less than that of evaluated lane, then set lane end cost of evaluated lane to 0
            // 1.1.2 if negative, then set lane end cost of evaluated lane to 0
            // 2 if negative, do not assign lane end cost
            float remainDistToEgoLnEndPnt = 0.0;
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, stPntLaneID, stPntDistAlongS, vehSpd, remainDistToEgoLnEndPnt))
            {
                float remainDistToEvlLnEndPnt = 0.0;
                if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, endPntLaneID, stPntDistAlongS, vehSpd, remainDistToEvlLnEndPnt))
                {
                    if (remainDistToEgoLnEndPnt > remainDistToEvlLnEndPnt)
                    {
                        this->m_LaneEndCost = 1.0;
                    }
                    else
                    {
                        this->m_LaneEndCost = 0.0;
                    }
                }
            }
        }
        else
        {
            // it is a lane follow action
            // check whether ego vehicle has been close enough to the lane end point of ego lane
            // 1. if positive, then set lane end cost to 1
            // 2. if negative, then set lane end cost to 0
            float remainDistToEgoLnEndPnt = 0.0;
            if (IsSmplPnt4LnEndPosi(ehr_msg.transit_lanes_list, stPntLaneID, stPntDistAlongS, vehSpd, remainDistToEgoLnEndPnt))
            {
                this->m_LaneEndCost = 1.0;
            }
        }
    }
    else
    {
        DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
    }
}

bool GlobalCostCalculation::IsSmplPnt4LnEndPosi(const std::vector<TransitLaneStruct> &transitLnLst,
                                                const uint8_t &laneLoc,
                                                float &distAlgS,
                                                const float &egoSpd,
                                                float &remainDist)
{
    // if
    // sample point is on a lane with lane end transit type
    // and distance btw ego position and lane end point is less than 300m or time headway btw ego position and lane end point is less than 30 sec
    // then return true
    bool flg = false;
    /* DLOG(INFO) << "transit lane list size = " << transitLnLst.size() << std::endl;
    DLOG(INFO) << "distAlgS inside = " << distAlgS << std::endl; */
    for (auto transitLn : transitLnLst)
    {
        /* DLOG(INFO) << "transit lane id = " << int(transitLn.transit_lane_id) << std::endl;
        DLOG(INFO) << "transit_lane_offset = " << transitLn.transit_lane_offset << std::endl; */
        if (transitLn.transit_type == BA::TransitionTyp_LaneEnd && transitLn.transit_lane_id == laneLoc)
        {
            if (abs(transitLn.transit_lane_offset - distAlgS) > 0.1) // 20220329: in case that two elements are the same
            {
                if (abs(transitLn.transit_lane_offset - distAlgS) < ADC::common::k_BA_LaneEndCalcDistThr || egoSpd * ADC::common::k_BA_LaneEndCalcTHThr + distAlgS > transitLn.transit_lane_offset)
                {
                    flg = true;
                    remainDist = abs(transitLn.transit_lane_offset - distAlgS);
                    // break;
                }
            }
        }
        if (flg == true)
        {
            break;
        }
    }
    return flg;
}

void GlobalCostCalculation::CalculateGlobalCost(const ControlWaypoints &ehr_msg,
                                                const RoadInfo &rm_msg,
                                                const CarPosition &carPos,
                                                float egoVehSpd,
                                                bool vldWayPntDtc,
                                                bool isEffcLCIntGnrt,
                                                BA::DriveMd drvMd,
                                                const bool &spdCondStsfyFlg,
                                                const bool &isEgoOnRghtSndLn)
{
    if (this->m_CurHostVehInfo.m_LaneLocation.empty() || this->m_CurHostVehInfo.m_DistanceAlongS.empty())
    {
        DLOG(WARNING) << "Current Host Vehicle Info is empty!" << std::endl;
        return;
    }
    // calc every cost items
    // CalcGlobalPathFllwCost();
    // CalcGlobalPathFllwCost_B();
    CalcGlobalPathFllwCost_InclLCP();
    // CalcGlobalPathFllwCost_C();

    // get lane number from EHR msg, excluding emergency lane
    // the lane with negative start offset is included
    int totalLnNum = 0;
    for (auto extendLane : this->m_GlobalInfoContainer.m_WpExtendLaneLst)
    {
        if (extendLane.is_emergency_lane == uint8_t(0) && extendLane.current_extendlane_offset < 1.0f)
        {
            totalLnNum++;
        }
    }
    DLOG(INFO) << "totalLnNum = " << totalLnNum << std::endl;

    // if total non-emergency lane number is greater than 2, then to include long-term fast lane cost and long-term slow lane cost
    if (totalLnNum > 2)
    {
        // get fast lane id and slow lane id from RM
        // if ego lane is at the corner, then assign ego lane id to them
        // or assign 255 to them
        uint8_t fastLnID = uint8_t(0);
        uint8_t slowLnID = uint8_t(0);
        auto it_lane = std::find_if(rm_msg.lane_list.begin(), rm_msg.lane_list.end(),
                                    [carPos](const Lane &lane_)
                                    {
                                        return (lane_.start_id == carPos.ego_lane_id && lane_.end_id == carPos.ego_lane_id);
                                    });
        if (it_lane != rm_msg.lane_list.end())
        {
            // ego lane obj has been found
            if ((*it_lane).left_neighbor_lane.empty())
            {
                fastLnID = carPos.ego_lane_id;
            }
            if ((*it_lane).right_neighbor_lane.empty())
            {
                slowLnID = carPos.ego_lane_id;
            }
            else
            {
                bool curRghtNghbLnExist = false;
                for (auto rghtNghbLn : (*it_lane).right_neighbor_lane)
                {
                    if (rghtNghbLn.start_range < 0)
                    {
                        curRghtNghbLnExist = true;
                        uint8_t rghtNghbLnID = rghtNghbLn.lane_id;
                        auto it_lane = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                    [rghtNghbLnID](const WPExtendLaneStruct &wpExtLn)
                                                    {
                                                        return (rghtNghbLnID == wpExtLn.lane_id);
                                                    });
                        if (it_lane != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                        {
                            if (it_lane->is_emergency_lane == 1)
                            {
                                slowLnID = carPos.ego_lane_id;
                            }
                        }
                        break;
                    }
                }
                // 20221101 added below
                if (curRghtNghbLnExist == false)
                {
                    slowLnID = carPos.ego_lane_id;
                }
            }
        }
        DLOG(INFO) << "slow lane id = " << int(slowLnID) << std::endl;
        CalcLgtTermFastLnCost_B(fastLnID, rm_msg);
        // CalcLgtTermSlowLnCost_B(slowLnID);
        CalcLgtTermSlowLnCost_InclLCP(slowLnID, ehr_msg);
    }
    CalcLnEndCost_InclLCP(ehr_msg, carPos, egoVehSpd);
    CalcTrffcEffcyCost(isEffcLCIntGnrt, spdCondStsfyFlg);
    CalcMrgRdLnCost_InclLCP(ehr_msg, carPos, egoVehSpd, drvMd);
    CalcTrgtLnCost(ehr_msg, carPos, egoVehSpd);
    CalcRghtScndLnCost(isEgoOnRghtSndLn);

    // according to valid CWP or RWP is detected or not, calculate m_GlobalCost by different means
    if (vldWayPntDtc == true)
    {
        // add traffic efficiency item
        if (this->m_GlobalInfoContainer.m_TgtRoadTyp_CWP != BA::TargetRoadTyp_MrgRd)
        {
            this->m_GlobalCost = (ADC::common::k_BA_GlobalPathFllwCostCoeff - 0.1f) * this->m_GlobalPathFllwCost +
                                 (ADC::common::k_BA_LaneEndCostCoeff - 0.1f) * this->m_LaneEndCost + 
                                 (ADC::common::k_BA_TrffcEffcCostCoeff - 0.1f) * this->m_TrffcEffcCost +
                                 ADC::common::k_BA_TrgtLnCostCoeff * this->m_TrgtLnCost +
                                 (ADC::common::k_BA_TrffcEffcCostCoeff - 0.1f) * this->m_MrgRdLnCost;
            DLOG(INFO) << "m_GlobalCost = " << this->m_GlobalCost << std::endl;
            DLOG(INFO) << "m_GlobalPathFllwCost = " << this->m_GlobalPathFllwCost << std::endl;
            DLOG(INFO) << "m_LaneEndCost = " << this->m_LaneEndCost << std::endl;
            DLOG(INFO) << "m_TrffcEffcCost = " << this->m_TrffcEffcCost << std::endl;
            DLOG(INFO) << "m_TrgtLnCost = " << this->m_TrgtLnCost << std::endl;
            DLOG(INFO) << "m_MrgRdLnCost = " << this->m_MrgRdLnCost << std::endl;
            DLOG(INFO) << "=============" << std::endl;
        }
        else
        {
            if (totalLnNum > 2)
            {
                this->m_GlobalCost = (ADC::common::k_BA_LaneEndCostCoeff) * this->m_LaneEndCost + 
                                     (ADC::common::k_BA_TrffcEffcCostCoeff - 0.1f) * this->m_TrffcEffcCost +
                                     (ADC::common::k_BA_TrffcEffcCostCoeff - 0.1f) * this->m_MrgRdLnCost;
                DLOG(INFO) << "m_LaneEndCost = " << this->m_LaneEndCost << std::endl;
                DLOG(INFO) << "m_TrffcEffcCost = " << this->m_TrffcEffcCost << std::endl;
                DLOG(INFO) << "m_MrgRdLnCost = " << this->m_MrgRdLnCost << std::endl;
                DLOG(INFO) << "=============" << std::endl;
            }
            else
            {
                this->m_GlobalCost = (ADC::common::k_BA_LaneEndCostCoeff) * this->m_LaneEndCost + 
                                     (ADC::common::k_BA_TrffcEffcCostCoeff - 0.1f) * this->m_TrffcEffcCost;
                DLOG(INFO) << "m_LaneEndCost = " << this->m_LaneEndCost << std::endl;
                DLOG(INFO) << "m_TrffcEffcCost = " << this->m_TrffcEffcCost << std::endl;
                DLOG(INFO) << "=============" << std::endl;
            }
        }
    }
    else
    {
        if (totalLnNum > 2)
        {
            this->m_GlobalCost = /* ADC::common::k_BA_LgtTermFastLnCostCoeff * this->m_LgTermFastLnCost + */
                                 ADC::common::k_BA_LgtTermSlowLnCostCoeff * this->m_LgTermSlowLnCost +
                                 ADC::common::k_BA_LaneEndCostCoeff * this->m_LaneEndCost +
                                 ADC::common::k_BA_TrffcEffcCostCoeff * this->m_TrffcEffcCost +
                                 ADC::common::k_BA_TrffcEffcCostCoeff * this->m_MrgRdLnCost +
                                 (ADC::common::k_BA_GlobalPathFllwCostCoeff - 0.3f) * this->m_GlobalPathFllwCost +
                                 ADC::common::k_BA_TrffcEffcCostCoeff * this->m_RghtScndLnCost;
            DLOG(INFO) << "m_GlobalCost = " << this->m_GlobalCost << std::endl;
            DLOG(INFO) << "m_LgTermFastLnCost = " << this->m_LgTermFastLnCost << std::endl;
            DLOG(INFO) << "m_LgTermSlowLnCost = " << this->m_LgTermSlowLnCost << std::endl;
            DLOG(INFO) << "m_LaneEndCost = " << this->m_LaneEndCost << std::endl;
            DLOG(INFO) << "m_TrffcEffcCost = " << this->m_TrffcEffcCost << std::endl;
            DLOG(INFO) << "m_MrgRdLnCost = " << this->m_MrgRdLnCost << std::endl;
            DLOG(INFO) << "m_RghtScndLnCost = " << this->m_RghtScndLnCost << std::endl;
            DLOG(INFO) << "=============" << std::endl;
        }
        else
        {
            this->m_GlobalCost = ADC::common::k_BA_LaneEndCostCoeff * this->m_LaneEndCost + 
                                 (ADC::common::k_BA_TrffcEffcCostCoeff - 0.1f) * this->m_TrffcEffcCost;;
            DLOG(INFO) << "m_LaneEndCost = " << this->m_LaneEndCost << std::endl;
            DLOG(INFO) << "m_TrffcEffcCost = " << this->m_TrffcEffcCost << std::endl;
            DLOG(INFO) << "=============" << std::endl;
        }
    }
}

void GlobalCostCalculation::CalcTrffcEffcyCost(const bool &isEffcLCIntGnrt, const bool &spdCondStsfy)
{
    // if efficiency relevant intention has been generated, then calc traffic efficiency cost
    // get lane follow avarage vehicle speed and current sequence vehicle speed
    // 20220809: add condition debounce
    if (isEffcLCIntGnrt == true)
    {
        if (spdCondStsfy == true)
        {
            this->m_TrffcEffcCost = 0.0;
        }
        else
        {
            this->m_TrffcEffcCost = 1.0;
        }
    }
    else
    {
        this->m_TrffcEffcCost = 0.0;
    }
}

void GlobalCostCalculation::CalcMrgRdLnCost_InclLCP(const ControlWaypoints &ehr_msg,
                                                    const CarPosition &carPos,
                                                    const float &vehSpd,
                                                    const BA::DriveMd &driveMode)
{
    this->m_MrgRdLnCost = 0.0;
    if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
    {
        uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
        uint8_t endPntLaneID = uint8_t(0);
        float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
        // 20220426: transform stPntDistAlongS
        stPntDistAlongS = this->m_GlobalInfoContainer.m_DistToWayPoint - stPntDistAlongS;
        BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
        if (this->m_ActionTypArr.size() > 1)
        {
            firstActionTyp = this->m_ActionTypArr.at(1);
        }
        // get virtual end point lane id by action
        if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare)
        {
            auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                           [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                           {
                                               return (stPntLaneID == wpExtLn.lane_id);
                                           });
            if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
            {
                if (!it_wpExtLn->left_extend_lane.empty())
                {
                    for (auto leftExtLn : it_wpExtLn->left_extend_lane)
                    {
                        uint8_t tmpExtLnID = leftExtLn.adjacent_extendlane_id;
                        auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                          [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                          {
                                                              return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                          });
                        uint8_t isEmrgLn = uint8_t(0);
                        if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                        {
                            isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                        }
                        // emergency lane should not be included, as emergency lane lack lane change point info
                        if (leftExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                        {
                            endPntLaneID = leftExtLn.adjacent_extendlane_id;
                            break;
                        }
                    }
                }
            }
        }
        else if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange)
        {
            endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        }
        else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare)
        {
            auto it_wpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                           [stPntLaneID](const WPExtendLaneStruct &wpExtLn)
                                           {
                                              return (stPntLaneID == wpExtLn.lane_id);
                                           });
            if (it_wpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
            {
                if (!it_wpExtLn->right_extend_lane.empty())
                {
                    for (auto rightExtLn : it_wpExtLn->right_extend_lane)
                    {
                        uint8_t tmpExtLnID = rightExtLn.adjacent_extendlane_id;
                        auto it_tmpWpExtLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                          [tmpExtLnID](const WPExtendLaneStruct &tmpWpExtLn)
                                                          {
                                                              return (tmpExtLnID == tmpWpExtLn.lane_id);
                                                          });
                        uint8_t isEmrgLn = uint8_t(0);
                        if (it_tmpWpExtLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                        {
                            isEmrgLn = it_tmpWpExtLn->is_emergency_lane;
                        }
                        // emergency lane should not be included, as emergency lane lack lane change point info
                        if (rightExtLn.adjacent_extendlane_offset < 0.0f && isEmrgLn == uint8_t(0))
                        {
                            endPntLaneID = rightExtLn.adjacent_extendlane_id;
                            break;
                        }
                    }
                }
            }
        }
        else if (firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
        {
            endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        }
        else if (firstActionTyp == BA::BA_ActionTyp_LaneFollow)
        {
            endPntLaneID = stPntLaneID;
        }
        if (endPntLaneID == uint8_t(0))
        {
            // if endPntLaneID is not assigned, then set endPntLaneID to stPntLaneID
            endPntLaneID = stPntLaneID;
        }
        // if (stPntLaneID != endPntLaneID)
        // as action number increases, modify judgement condition from comparison  btw start ID and end ID TO identification of first action type
        if (/* firstActionTyp == BA::BA_ActionTyp_LeftLaneChange ||  */
            firstActionTyp == BA::BA_ActionTyp_RightLaneChange ||
            firstActionTyp == BA::BA_ActionTyp_RightLaneChangePrepare /* ||
            firstActionTyp == BA::BA_ActionTyp_LeftLaneChangePrepare */)
        {
            // it is a lane change action
            // check whether end lane is the drivable lane on the most right side
            // 1 if positive, assign merge road lane cost
            // 1.1 if first control waypoint type is VirtFixDistRd or JCTRd/ICRd/SpltRd and distance is greater than 1800m
            // 1.1.1 if right lane of end point lane is emergency lane and first remind waypoint type is not MrgRd or the first remind waypoint type is MrgRd and the dist btw ego vehicle and waypoint is greater than 800m 
            // 1.1.2 or if right lane of end point lane is not included in target lane list   
            // 2 if negative, do not assign merge road lane cost
            if (this->m_GlobalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd || 
                ((this->m_GlobalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_JCTRd || this->m_GlobalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_ICRd || this->m_GlobalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_SpltRd)
                 && this->m_GlobalInfoContainer.m_DistToWayPoint >= 1800.0f))
            {
                auto it_endPntLnWp = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                 [endPntLaneID](const WPExtendLaneStruct &wpLn_)
                                                 {
                                                     return (wpLn_.lane_id == endPntLaneID);
                                                 });
                uint8_t rghtNgbEndPntLnID = uint8_t(0);
                if (it_endPntLnWp != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                {
                    if (!it_endPntLnWp->right_extend_lane.empty())
                    {
                        DLOG(INFO) << "right lane is not empty" << std::endl;
                        for (auto rghtExtLn : it_endPntLnWp->right_extend_lane)
                        {
                            if (rghtExtLn.adjacent_extendlane_offset < 1.0f)
                            {
                                rghtNgbEndPntLnID = rghtExtLn.adjacent_extendlane_id;
                                break;
                            }
                        }
                        DLOG(INFO) << "rghtNgbEndPntLnID: " << int(rghtNgbEndPntLnID) << std::endl;
                        auto it_rghtNgbEndPntLn = std::find_if(this->m_GlobalInfoContainer.m_WpExtendLaneLst.begin(), this->m_GlobalInfoContainer.m_WpExtendLaneLst.end(),
                                                               [rghtNgbEndPntLnID](const WPExtendLaneStruct &rghtWpLn_)
                                                               {
                                                                   return (rghtWpLn_.lane_id == rghtNgbEndPntLnID);
                                                               });
                        if (it_rghtNgbEndPntLn != this->m_GlobalInfoContainer.m_WpExtendLaneLst.end())
                        {
                            if (it_rghtNgbEndPntLn->is_emergency_lane == true)
                            {
                                DLOG(INFO) << "reminder_waypoint_target_road_type: " << int(ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type) << std::endl;
                                DLOG(INFO) << "reminder_waypoint_offset: " << ehr_msg.reminder_waypoint.front().reminder_waypoint_offset << std::endl;
                                if (driveMode == BA::DriveMd_Sport && ADC::common::k_BA_SpprLC2MrgRdLn == false)
                                {
                                    if (ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubMrgRd || ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubSpltRd)
                                    { 
                                        if (ehr_msg.reminder_waypoint.front().reminder_waypoint_offset <= ADC::common::k_BA_MinDstToMrgRdAllwLC)
                                        {
                                            this->m_MrgRdLnCost = 1.0;
                                            DLOG(INFO) << "flg00" << std::endl;
                                        }
                                    }
                                }
                                else
                                {
                                    this->m_MrgRdLnCost = 1.0;
                                    DLOG(INFO) << "flg01" << std::endl;
                                } 
                            }
                            else
                            {
                                auto it_rghtNgbEndPntLnFound = std::find(this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.begin(), this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.end(), rghtNgbEndPntLnID);
                                if (it_rghtNgbEndPntLnFound == this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.end() && this->m_GlobalInfoContainer.m_TgtRoadTyp_CWP == BA::TargetRoadTyp_VirtFixDstRd)
                                {
                                    // right neighbor lane of end point lane is not included in target lane list
                                    this->m_MrgRdLnCost = 1.0;
                                    DLOG(INFO) << "flg02" << std::endl;
                                }
                            }
                        }
                        else
                        {
                            // right right lane does not exist
                            if (driveMode == BA::DriveMd_Sport && ADC::common::k_BA_SpprLC2MrgRdLn == false)
                            {
                                if (ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubMrgRd || ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubSpltRd)
                                { 
                                    if (ehr_msg.reminder_waypoint.front().reminder_waypoint_offset <= ADC::common::k_BA_MinDstToMrgRdAllwLC)
                                    {
                                        this->m_MrgRdLnCost = 1.0;
                                        DLOG(INFO) << "flg03" << std::endl;
                                    }
                                }
                            }
                            else
                            {
                                this->m_MrgRdLnCost = 1.0;
                                DLOG(INFO) << "flg04" << std::endl;
                            } 
                        }
                    }
                    else
                    {
                        DLOG(INFO) << "right lane is empty" << std::endl;
                        if (driveMode == BA::DriveMd_Sport && ADC::common::k_BA_SpprLC2MrgRdLn == false)
                        {
                            if (ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubMrgRd || ehr_msg.reminder_waypoint.front().reminder_waypoint_target_road_type == BA::TargetRoadTyp_SubSpltRd)
                            {
                                if (ehr_msg.reminder_waypoint.front().reminder_waypoint_offset <= ADC::common::k_BA_MinDstToMrgRdAllwLC)
                                {
                                    this->m_MrgRdLnCost = 1.0;
                                    DLOG(INFO) << "flg05" << std::endl;
                                }
                            }
                        }
                        else
                        {
                            this->m_MrgRdLnCost = 1.0;
                            DLOG(INFO) << "flg06" << std::endl;
                        }
                    }
                }
            }
        }
    }
    else
    {
        DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
    }
}

void GlobalCostCalculation::CalcTrgtLnCost(const ControlWaypoints &ehr_msg,
                                           const CarPosition &carPos,
                                           const float &vehSpd)
{
    // 20220720
    // if end lane id is target lane. then target lane cost is 0
    // else the target lane cost is 1
    this->m_TrgtLnCost = 0.0f;
    if (!this->m_CurHostVehInfo.m_LaneLocation.empty() && !this->m_CurHostVehInfo.m_DistanceAlongS.empty())
    {
        uint8_t stPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.front();
        uint8_t endPntLaneID = this->m_CurHostVehInfo.m_LaneLocation.back();
        float stPntDistAlongS = this->m_CurHostVehInfo.m_DistanceAlongS.front();
        // 20220426: transform stPntDistAlongS
        stPntDistAlongS = this->m_GlobalInfoContainer.m_DistToWayPoint - stPntDistAlongS;
        // BA::BA_ActionTyp firstActionTyp = this->m_ActionTypArr.front();
        BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
        if (this->m_ActionTypArr.size() > 1)
        {
            firstActionTyp = this->m_ActionTypArr.at(1);
        }
        // if (stPntLaneID != endPntLaneID)
        // as action number increases, modify judgement condition from comparison  btw start ID and end ID TO identification of first action type
        if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange || firstActionTyp == BA::BA_ActionTyp_RightLaneChange)
        {
            // it is a lane change action
            // check whether end lane ID is target lane 
            // if positive, then set target lane cost to 0
            // if negative, then set target lane cost to 1
            if (this->m_GlobalInfoContainer.m_DistToWayPoint < ADC::common::k_BA_AntiDrctLCDist2LastLCPntThr_Cnsr)
            {
                auto it_isTrgtLn = std::find(this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.begin(), this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.end(), endPntLaneID);
                if (it_isTrgtLn != this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.end())
                {
                    this->m_TrgtLnCost = 0.0f;
                }
                else
                {
                    this->m_TrgtLnCost = 1.0f;
                }
            }
        }
        else
        {
            // it is a lane follow action
            // check whether end lane ID is target lane 
            // if positive, then set target lane cost to 0
            // if negative, then set target lane cost to 1
            if (this->m_GlobalInfoContainer.m_DistToWayPoint < ADC::common::k_BA_AntiDrctLCDist2LastLCPntThr_Cnsr)
            {
                auto it_isTrgtLn = std::find(this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.begin(), this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.end(), endPntLaneID);
                if (it_isTrgtLn != this->m_GlobalInfoContainer.m_TgtLaneLst_CWP.end())
                {
                    this->m_TrgtLnCost = 0.0f;
                }
                else
                {
                    this->m_TrgtLnCost = 1.0f;
                }
            }
        }
    }
    else
    {
        DLOG(WARNING) << "m_CurHostVehInfo is empty." << std::endl;
    }
}

void GlobalCostCalculation::CalcRghtScndLnCost(const bool &isOnRghtScndLn)
{
    BA::BA_ActionTyp firstActionTyp = BA::BA_ActionTyp_None;
    if (this->m_ActionTypArr.size() > 1)
    {
        firstActionTyp = this->m_ActionTypArr.at(1);
    }
    if (isOnRghtScndLn == true)
    {
        if (firstActionTyp == BA::BA_ActionTyp_LeftLaneChange)
        {
            m_RghtScndLnCost = 0.0f;
        }
        else
        {
            m_RghtScndLnCost = 1.0f;
        }
    }
    else
    {
        m_RghtScndLnCost = 0.0f;
    }
}