#include "c0calculator.h"
#include "atsutil.h"

#include <cmath>
#include <iostream>
namespace ACC_Target_Selection {

C0Calculator::C0Calculator(const Param_C0Calculator &param_conf) :
        param_conf_(param_conf),
        curvaturecog_(zero_float32_t),
        radiuscog_(param_conf_.radiuscog),
        slip_(zero_float32_t),
        curvaturerr_(zero_float32_t),
        curvaturerrraw_(zero_float32_t),
        c0_(zero_float32_t),
        c0basedkinematics_(zero_float32_t),
        c0baseddynamics_(zero_float32_t),
        reset_(boolean_t_true) {
}

C0Calculator::C0Calculator(const C0Calculator& r) :
        param_conf_(r.param_conf_),
        curvaturecog_(r.curvaturecog_),
        radiuscog_(r.radiuscog_),
        slip_(r.slip_),
        curvaturerr_(r.curvaturerr_),
        curvaturerrraw_(r.curvaturerrraw_),
        c0_(r.c0_),
        c0basedkinematics_(r.c0basedkinematics_),
        c0baseddynamics_(r.c0baseddynamics_),
        reset_(r.reset_) {
}

C0Calculator::~C0Calculator() {
}

void C0Calculator::blend(const basetype::float32_t &spd) {
    basetype::float32_t factor{Util::lookup(param_conf_.c0blendfactor,
                                            param_conf_.c0blendfactorbreakpoints,
                                            spd,
                                            boolean_t_false)};
    c0_ = factor * c0basedkinematics_ + (one_float32_t - factor) * c0baseddynamics_;
    //std::cout<< "c0basedkinematics_: "<<(C0Calculator::c0basedkinematics_)<<std::endl;
    //std::cout<< "c0baseddynamics_: "<<(C0Calculator::c0baseddynamics_)<<std::endl;
    //std::cout<< "c0_: "<<(C0Calculator::c0_)<<std::endl;
}

void C0Calculator::calculatec0basedkinematics(const basetype::float32_t &spd, const basetype::float32_t &yawrate) {
    basetype::float32_t res{zero_float32_t};
    res =  Util::safetydiv(yawrate, spd,eps4_float32_t) * 0.5f;
    if (reset_ == boolean_t_true) {
        res = zero_float32_t;
    }
    
    c0basedkinematics_ = res;

}

void C0Calculator::calculatec0baseddynamics(const basetype::float32_t &spd,
                                            const basetype::float32_t &steeringangle,
                                            const basetype::float32_t &yawrate) {

    basetype::float32_t pinionangle{Util::safetydiv(steeringangle,
                                                    param_conf_.steeringrateforbicyclemodel,
                                                    one_float32_t)};
    C0Calculator::calulatecurvaturecog(pinionangle, spd);
    
    C0Calculator::calculateslip(spd);
    
    C0Calculator::calculatecurvaturerear(spd, yawrate, steeringangle);

    C0Calculator::lp1r();

    if (reset_ == boolean_t_true) {
        reset_ = boolean_t_false;
    }
}

void C0Calculator::calulatecurvaturecog(const basetype::float32_t &steering,
                                        const basetype::float32_t &spd) {
    basetype::float32_t Kappa{zero_float32_t};
    Kappa = param_conf_.vehiclemassforbicyclemodel *
           (Util::safetydiv((param_conf_.whleelbaseforbicyclemodel - param_conf_.frontdistanceforbicyclemodel),
                             param_conf_.cornrgstfnfrntforbicyclemodel,
                             eps4_float32_t) - 
            Util::safetydiv(param_conf_.frontdistanceforbicyclemodel,param_conf_.cornrgstfnreforbicyclemodel, eps4_float32_t));
    basetype::float32_t KappaTimesSpeedSqr{Kappa * spd * spd};
    basetype::float32_t WheelBaseSqr{param_conf_.whleelbaseforbicyclemodel * param_conf_.whleelbaseforbicyclemodel};
    basetype::float32_t nominator{steering * param_conf_.whleelbaseforbicyclemodel};
    basetype::float32_t denominator{WheelBaseSqr + KappaTimesSpeedSqr};
    curvaturecog_ = Util::safetydiv(nominator, denominator, eps4_float32_t);
    radiuscog_ = Util::safetydiv(denominator, nominator, eps4_float32_t);
}

void C0Calculator::calculateslip(const basetype::float32_t &spd) {
    basetype::float32_t Mulipli{param_conf_.vehiclemassforbicyclemodel * spd * spd * param_conf_.frontdistanceforbicyclemodel};
    basetype::float32_t Multipl{param_conf_.whleelbaseforbicyclemodel * param_conf_.cornrgstfnreforbicyclemodel};

    basetype::float32_t slip = {curvaturecog_ * ((param_conf_.whleelbaseforbicyclemodel - param_conf_.frontdistanceforbicyclemodel) - 
                               Util::safetydiv(Mulipli, Multipl, eps4_float32_t))};
    slip_ = Util::saturate(slip, 0.99f, -0.99f);
}

void C0Calculator::calculatecurvaturerear(const basetype::float32_t &spd, const basetype::float32_t &yawrate, const basetype::float32_t &steeringangle) {
    basetype::float32_t radiusrear{zero_float32_t};
    basetype::float32_t denominator{zero_float32_t};
    basetype::float32_t sin_slip{std::sin(slip_)};
    basetype::float32_t cos_slip{std::cos(slip_)};
    basetype::float32_t sqr1{param_conf_.whleelbaseforbicyclemodel - param_conf_.frontdistanceforbicyclemodel -
                             (sin_slip * radiuscog_)};
    sqr1 = sqr1 * sqr1;
    basetype::float32_t sqr2{cos_slip * radiuscog_};
    sqr2 = sqr2 * sqr2;

    if ((sqr1 + sqr2) < zero_float32_t) {
        radiusrear = zero_float32_t;
    } else {
        denominator = std::sqrt(sqr1 + sqr2) * static_cast<basetype::float32_t>(Util::signfunc(curvaturecog_));
        radiusrear = Util::safetydiv(one_float32_t, denominator, eps4_float32_t);
    }

    if (reset_ == boolean_t_true) {
        radiusrear = zero_float32_t;
    }

    radiusrear = Util::saturate(radiusrear, param_conf_.upperlimitcurvaturerear, param_conf_.lowerlimitcurvaturerear);

    curvaturerr_ = radiusrear * param_conf_.c0biasfactorbasedynamics * 0.5f;
}

void C0Calculator::lp1r() {
    basetype::float32_t tsOvertau{Util::safetydiv(c_SampleTime,
                                                 (c_SampleTime + Util::safetydiv(one_float32_t, param_conf_.c0cutofffrequencybasedynamics, eps4_float32_t)),
                                                  eps4_float32_t)};
    basetype::boolean_t reset{boolean_t_false};
    if (reset == boolean_t_true ||
        (reset_ == boolean_t_true)) {
        c0baseddynamics_ = curvaturerr_;
    } else {
        c0baseddynamics_ = curvaturerr_ * tsOvertau + (c0baseddynamics_ - tsOvertau * c0baseddynamics_);
        //std::cout<< "curvaturerr_: "<<(C0Calculator::curvaturerr_)<<std::endl;
        //std::cout<< "c0baseddynamics_: "<<(C0Calculator::c0baseddynamics_)<<std::endl;
        //std::cout<< "tsOvertau: "<<(tsOvertau)<<std::endl;
    }
}

void C0Calculator::calculatec0(const basetype::float32_t &spd, const basetype::float32_t &steeringangle, const basetype::float32_t &yawrate) {
    C0Calculator::calculatedeadzone(spd, yawrate, steeringangle);
    C0Calculator::calculatec0basedkinematics(spd, yawrate);
    C0Calculator::calculatec0baseddynamics(spd, steeringangle, yawrate);
    C0Calculator::blend(spd);
    basetype::float32_t steeringangleabs{Util::abs(steeringangle)};
    basetype::float32_t maxc0 {Util::lookup(param_conf_.c0kinematicsyawratemax,
                                            param_conf_.c0c0kinematicsyawratemaxbreakpoints,
                                            steeringangleabs,
                                            boolean_t_false)};
    c0_ = Util::saturate(c0_, maxc0, -maxc0);
}

void C0Calculator::calculatedeadzone(const basetype::float32_t &spd, const basetype::float32_t &yawrate, const basetype::float32_t &steeringangle) {
    if ((spd < 1.0f) &&
        (spd > -1.0) &&
        (yawrate < 0.0055f) &&
        (yawrate > -0.0055f)) {
        //curvature = zero_float32_t;
        C0Calculator::reset_ = boolean_t_true;
    } else if ((steeringangle < eps3_float32_t) &&
               (steeringangle > -eps3_float32_t) &&
               (yawrate < 0.0055f) &&
               (yawrate > -0.0055f)) {
        //curvature = zero_float32_t;
        C0Calculator::reset_ = boolean_t_true;
    } else {
        //*curvature
        C0Calculator::reset_ = boolean_t_false;
    }
}

basetype::float32_t C0Calculator::get_C0() {
    return c0_;
}

void C0Calculator::enablereset() {
    reset_ = boolean_t_true;
}

} // namespace C0Calculator