#include "selector.h"
#include "atsutil.h"
#include <cmath>

namespace ACC_Target_Selection{
    Selector::Selector(const Param_Selection &param_conf, const Logger &logger): 
        param_conf_(param_conf.selector),
        logger_(logger),
        c0calculator_(param_conf.c0calculator),
        validator_(param_conf.validator, logger),
        offsetcalculator_(param_conf.offsetcalculator, logger),
        funnel_(param_conf.funnel, logger),
        road_(param_conf.road, logger),
        pedestrain_(zero_uint32_t),
        iv_(zero_uint32_t),
        iiv_(zero_uint32_t),
        cutin_(zero_uint32_t),
        fliv_(zero_uint32_t),
        sliv_(zero_uint32_t),
        fliiv_(zero_uint32_t),
        sliiv_(zero_uint32_t),
        riv_(zero_uint32_t),
        riiv_(zero_uint32_t),
        rcutin_(zero_uint32_t),
        rfliv_(zero_uint32_t),
        rsliv_(zero_uint32_t),
        rfliiv_(zero_uint32_t),
        rsliiv_(zero_uint32_t),
        ivquality_(zero_uint8_t),
        ivcutingout_(boolean_t_false),
        ivcutoutstatus_(ObjectLaneStatus::ObjectKeepinglane),
        ivoffset_(zero_float32_t),
        iivoffset_(zero_float32_t),
        cutinoffset_(zero_float32_t),
        flivoffset_(zero_float32_t),
        slivoffset_(zero_float32_t),
        fliivoffset_(zero_float32_t),
        sliivoffset_(zero_float32_t),
        rivoffset_(zero_float32_t),
        riivoffset_(zero_float32_t),
        rcutinoffset_(zero_float32_t),
        rflivoffset_(zero_float32_t),
        rslivoffset_(zero_float32_t),
        rfliivoffset_(zero_float32_t),
        rsliivoffset_(zero_float32_t),  
        predictionmaxdistance_(zero_float32_t),
        yawplusafterindicatortimer_(boolean_t_false,c_SampleTime),
        yawplusafterlanechangetimer_(boolean_t_false,c_SampleTime) {
        for (basetype::objectid_t i = zero_objectid_t; i < c_ObjectSize; i++) {
            inpathcounter_.at(i)     = zero_uint8_t;
            inwidepathcounter_.at(i) = zero_uint8_t;
            cutinpathcounters_.at(i) = zero_uint8_t;
            cutinholdcounters_.at(i) = zero_uint8_t;
        }
    }

    Selector::Selector(const Selector& r) :
            param_conf_(r.param_conf_),
            logger_(r.logger_),
            c0calculator_(r.c0calculator_),
            validator_ (r.validator_),
            offsetcalculator_(r.offsetcalculator_),
            funnel_(r.funnel_),
            road_(r.road_),
            pedestrain_(r.pedestrain_),
            iv_(r.iv_),
            iiv_(r.iiv_),
            cutin_(r.cutin_),
            fliv_(r.fliv_),
            sliv_(r.sliv_),
            fliiv_(r.fliiv_),
            sliiv_(r.sliiv_),
            riv_(r.riv_),
            riiv_(r.riiv_),
            rcutin_(r.rcutin_),
            rfliv_(r.rfliv_),
            rsliv_(r.rsliv_),
            rfliiv_(r.rfliiv_),
            rsliiv_(r.rsliiv_),
            ivquality_(r.ivquality_),
            ivcutingout_(r.ivcutingout_),
            ivcutoutstatus_(r.ivcutoutstatus_),
            ivoffset_(r.ivoffset_),
            iivoffset_(r.iivoffset_),
            cutinoffset_(r.cutinoffset_),
            flivoffset_(r.flivoffset_),
            slivoffset_(r.slivoffset_),
            fliivoffset_(r.fliivoffset_),
            sliivoffset_(r.sliivoffset_),
            rivoffset_(r.rivoffset_),
            riivoffset_(r.riivoffset_),
            rcutinoffset_(r.rcutinoffset_),
            rflivoffset_(r.rflivoffset_),
            rslivoffset_(r.rslivoffset_),
            rfliivoffset_(r.rfliivoffset_),
            rsliivoffset_(r.rsliivoffset_),
            predictionmaxdistance_(r.predictionmaxdistance_),
            yawplusafterindicatortimer_(r.yawplusafterindicatortimer_),
            yawplusafterlanechangetimer_(r.yawplusafterlanechangetimer_) {
    }

    Selector::~Selector() {
    }

    basetype::boolean_t Selector::checkpredicition(const Object &obj) {
        basetype::boolean_t res{boolean_t_false};
        if ((obj.Est.DisLgt <= predictionmaxdistance_) &&
            (obj.Est.DisLgt >= param_conf_.mindistanceforprediction) &&
            (obj.Est.Spd >= param_conf_.minspeedforprediction) &&
            (obj.Info.Type != ObjectType::UnknownVehicleClass) &&
            (obj.Info.Type != ObjectType::Unknown) &&
            (obj.Ppty.FusionSource == ObjectFusionSource::Fused)) {
            res = boolean_t_true;
        }
        else{
            res = boolean_t_false;
        }

        return res;
    }

    basetype::boolean_t Selector::checkcutout(const basetype::objectid_t &idn0,
                                              const basetype::boolean_t &isrearobject,
                                              const basetype::float32_t &dislgt,
                                              const basetype::float32_t &spd,
                                              const basetype::float32_t &offset,
                                              const basetype::float32_t &offsetderiv) {
        basetype::boolean_t res{boolean_t_false};
        FunnelWidthParam funnelwidthparam{};
        funnelwidthparam = funnel_.getouter(isrearobject, idn0);

        basetype::float32_t timeleft{Util::safetydiv((funnelwidthparam.leftwidth - offset), offsetderiv, eps3_float32_t)};

        basetype::float32_t timeright{Util::safetydiv((funnelwidthparam.rightwidth - offset), offsetderiv, eps3_float32_t)};

        basetype::float32_t timegap{Util::safetydiv(dislgt, spd, eps3_float32_t)};

        basetype::float32_t timegapfactor{Util::lookup(param_conf_.cutouttimegapfactor,
                                                       param_conf_.cutouttimegapfactor,
                                                       timegap,
                                                       boolean_t_false)};
        basetype::float32_t timethreshold{timegapfactor * param_conf_.cutouttimethreshold};

        ivcutoutstatus_ = ObjectLaneStatus::ObjectKeepinglane;

        if (std::signbit(offset) == std::signbit(offsetderiv)) {
            if ((timeleft > zero_float32_t) &&
                (timeleft <= timethreshold)) {
                res = boolean_t_true;
                ivcutoutstatus_ = ObjectLaneStatus::ObjectChangeingtoLeft;
            }
            else if ((timeright > zero_float32_t) &&
                (timeright <= timethreshold)) {
                res = boolean_t_true;
                ivcutoutstatus_ = ObjectLaneStatus::ObjectChangeingtoRight;
            }
        }

        return res;
    }

    basetype::boolean_t Selector::checkinpath(const basetype::objectid_t &idn0,
                                              const basetype::boolean_t &isrearobject,
                                              const basetype::float32_t &offset,
                                              FunnelWidthParam &funnelwidthparam) {
        basetype::boolean_t res{boolean_t_false};
        basetype::uint8_t maxinpathcounter{param_conf_.maxinpathcounter};
    
        if (inpathcounter_.at(idn0)> maxinpathcounter) {
            funnelwidthparam = funnel_.getouter(isrearobject, idn0);
        } else {
            funnelwidthparam = funnel_.getinner(isrearobject,idn0);
        }

        if ((offset >= funnelwidthparam.rightwidth) &&
            (offset <= funnelwidthparam.leftwidth)) {
            res = boolean_t_true;
            if (inpathcounter_.at(idn0) <= maxinpathcounter) {
                inpathcounter_.at(idn0) = inpathcounter_.at(idn0) + one_uint8_t;
            }
        } else {
            res = boolean_t_false;
            inpathcounter_.at(idn0) = zero_uint8_t;
        }
        
        return res;                            
    }

    basetype::boolean_t Selector::checkinwidepath(const basetype::objectid_t &idn0,
                                                  const basetype::boolean_t &isrearobject,
                                                  const basetype::float32_t &offset,
                                                  FunnelWidthParam &funnelwidthparam) {
        basetype::boolean_t res{boolean_t_false};
        basetype::uint8_t maxinpathcounter{static_cast<basetype::uint8_t>(Util::min(Util::safetydiv(param_conf_.funnelhysteresis,
                                                                                                    c_SampleTime, eps3_float32_t),
                                                                                    254.0f))};
        if (inwidepathcounter_.at(idn0) > maxinpathcounter) {
            funnelwidthparam = funnel_.getwideouter(isrearobject, idn0);
        } else {
            funnelwidthparam = funnel_.getwideinner(isrearobject, idn0);
        }

        if ((offset >= funnelwidthparam.rightwidth) &&
            (offset <= funnelwidthparam.leftwidth)) {
            res = boolean_t_true;
            if (inwidepathcounter_.at(idn0) <= maxinpathcounter) {
                inwidepathcounter_.at(idn0) = inwidepathcounter_.at(idn0) + one_uint8_t;
            }
        } else {
            res = boolean_t_false;
            inwidepathcounter_.at(idn0) = zero_uint8_t;
        }
        
        return res;
    }

    basetype::boolean_t Selector::checkcutin(const basetype::uint8_t &idn0,
                                             const basetype::boolean_t &isrearobject,
                                             const basetype::boolean_t &useprediction,
                                             const basetype::float32_t &predictedoffset) {
        basetype::boolean_t res{boolean_t_false};
        FunnelWidthParam funnelwidthparam{};
        if (useprediction == boolean_t_true) {
            funnelwidthparam = funnel_.getinner(isrearobject, idn0);
            if ((predictedoffset >= funnelwidthparam.rightwidth) &&
                (predictedoffset <= funnelwidthparam.leftwidth)) {
                res = boolean_t_true;
            }
        }

        if (res == boolean_t_true) {
            cutinpathcounters_.at(idn0) = static_cast<basetype::uint8_t>(4u);
        } else {
            if (cutinpathcounters_.at(idn0) >= one_uint8_t) {
                cutinpathcounters_.at(idn0) = cutinpathcounters_.at(idn0) - one_uint8_t;
            }
        }
        
        if (cutinpathcounters_.at(idn0) > zero_uint8_t) {
            res = boolean_t_true;
        } else {
            res = boolean_t_false;
        }
        
        return res;
    }

    void Selector::resetassignment() {
        pedestrain_       = zero_objectid_t;
        iv_               = zero_objectid_t;
        iiv_              = zero_objectid_t;
        cutin_            = zero_objectid_t;
        fliv_             = zero_objectid_t;
        sliv_             = zero_objectid_t;
        fliiv_            = zero_objectid_t;
        sliiv_            = zero_objectid_t;
        riv_              = zero_objectid_t;
        riiv_             = zero_objectid_t;
        rcutin_           = zero_objectid_t;
        rfliv_            = zero_objectid_t;
        rsliv_            = zero_objectid_t;
        ivquality_        = zero_uint8_t;
        ivcutingout_      = boolean_t_false;
        ivcutoutstatus_   = ObjectLaneStatus::ObjectKeepinglane;
        ivoffset_         = zero_float32_t;
        iivoffset_        = zero_float32_t;
        cutinoffset_      = zero_float32_t;
        flivoffset_       = zero_float32_t;
        slivoffset_       = zero_float32_t;
        fliivoffset_      = zero_float32_t;
        sliivoffset_      = zero_float32_t;
        rivoffset_        = zero_float32_t;
        riivoffset_       = zero_float32_t;
        rcutinoffset_     = zero_float32_t;
        rflivoffset_      = zero_float32_t;
        rslivoffset_      = zero_float32_t;
    }

    basetype::boolean_t Selector::isrearobject(const Object &obj) {
        basetype::boolean_t res = boolean_t_false;
        if (obj.Est.DisLgt <= param_conf_.mintargetdistance) {
            if (!((Util::abs(obj.Est.DisLat) <= param_conf_.mintargetlatdistance) &&
               (obj.Est.DisLgt > param_conf_.minreartargetdistance))) {
                res = boolean_t_true;
            }
        }
        return res;
    }

    void Selector::update(const std::array<Object, c_ObjectSize> &objects,
                          const EgoVehicleInformation &egovehinfo,
                          const RoadProperty &roadproperty,
                          const basetype::boolean_t &leftindicator,
                          const basetype::boolean_t &rightindicator,
                          const LaneChangeType &lanechange) {
        c0calculator_.calculatec0((egovehinfo.SpdLgt),
                                  (egovehinfo.SteeringAngle),
                                  (egovehinfo.YawRate));
        
        validator_.setprevtgta(Selector::iv_, Selector::riv_);
        Selector::resetassignment();

        basetype::boolean_t indicatoractive;
        if ((leftindicator == boolean_t_true) ||
            (rightindicator == boolean_t_true)) {
            indicatoractive = boolean_t_true;
        } else {
            indicatoractive = boolean_t_false;
        }
        
        yawplusafterindicatortimer_.update(indicatoractive);

        basetype::boolean_t indicatortimerhigh{yawplusafterindicatortimer_.hold(boolean_t_true, param_conf_.maxholdforyawplus)};

        basetype::boolean_t lanechangeactive{boolean_t_false};
        if ((lanechange > LaneChangeType::None)) {
            if (indicatortimerhigh == boolean_t_true) {
                lanechangeactive = boolean_t_true;
            } else {
                lanechangeactive = boolean_t_false;
            }
        } else {
            lanechangeactive = boolean_t_false;
        }
        yawplusafterlanechangetimer_.update(lanechangeactive);

        basetype::boolean_t lanechangetimerhigh{yawplusafterlanechangetimer_.hold(boolean_t_true, param_conf_.holdyawpluslanechange)};
        if (lanechangetimerhigh == boolean_t_true) {
           yawplusafterindicatortimer_.reset(boolean_t_false);
        }

        basetype::boolean_t useyawplusinlanechange{boolean_t_false};
        if ((indicatortimerhigh == boolean_t_true) ||
            (lanechangetimerhigh == boolean_t_true)) {
            useyawplusinlanechange = boolean_t_true;
        } else {
            useyawplusinlanechange = boolean_t_false;
        }

        road_.set(roadproperty, egovehinfo);

        //log c0
        logger_.logc0(c0calculator_.get_C0());

        basetype::float32_t laneoffset{zero_float32_t};
        if (useyawplusinlanechange == boolean_t_true) {
            laneoffset = zero_float32_t;
        }
        else{
            laneoffset = road_.getlaneoffset();
        }

        funnel_.prepare(egovehinfo.SpdLgt, leftindicator, rightindicator,laneoffset);
        basetype::boolean_t isrearobject{boolean_t_false};
        for (basetype::objectid_t index = zero_objectid_t; index < c_ObjectSize; index++) {
            isrearobject = Selector::isrearobject(objects.at(index));
            logger_.logisrearobject(isrearobject, index);

            if (validator_.checkvalid(objects.at(index), egovehinfo, isrearobject)) {
                // log valid
                logger_.logvalid(boolean_t_true, index);
            } else {
                offsetcalculator_.reset(index);
                Selector::reset(index);

                // log valid highqualityoffsetfromlanecenter withinfunnel leftfunnel rightfunnel withinwidefunnel
                // useprediction
                logger_.logvalid(boolean_t_false, index);
                logger_.loghighqualityoffsetfromlanecenter(boolean_t_false, index);
                logger_.logwithinfunnel(boolean_t_false, index);
                logger_.loginnerleftfunnel(zero_float32_t, index);
                logger_.loginnerrightfunnel(zero_float32_t, index);
                logger_.logouterleftfunnel(zero_float32_t, index);
                logger_.logouterrightfunnel(zero_float32_t, index);
                logger_.logwithinwidefunnel(boolean_t_false, index);

                continue;
            }
            
            offsetcalculator_.calculate(objects.at(index),
                                         isrearobject,
                                         road_.getisoffsetfromlanecenterusable(),
                                         road_.getisoffsetfromlanecentervalid(),
                                         road_.getlaneoffset(),
                                         egovehinfo,
                                         c0calculator_.get_C0(),
                                         useyawplusinlanechange);
            basetype::float32_t offset{offsetcalculator_.getoffset()};

            //log highqualityoffsetfromlanecenter
            logger_.loghighqualityoffsetfromlanecenter(offsetcalculator_.gethighqualityoffsetfromlanecenter(), index);

            funnel_.prepareobject(isrearobject,
                                  objects.at(index).Est.DisLgt,
                                  index,
                                  road_.getlanewidth(objects.at(index).Est.DisLgt));
            
            FunnelWidthParam funnelwidthparam{};
            basetype::boolean_t withinfunnel{Selector::checkinpath(index,
                                                                   isrearobject,
                                                                   offset,
                                                                   funnelwidthparam)};
            
            // log withinfunnel leffunnel rightfunnel
            logger_.logwithinfunnel(withinfunnel, index);
            logger_.loginnerleftfunnel(funnelwidthparam.leftwidth, index);
            logger_.loginnerrightfunnel(funnelwidthparam.rightwidth, index);

            FunnelWidthParam widerfunnelwidthparam{};
            basetype::boolean_t withinwidefunnel{Selector::checkinwidepath(index,
                                                                           isrearobject,
                                                                           offset,
                                                                           widerfunnelwidthparam)};
            
            // log withwidefunnel
            logger_.logwithinwidefunnel(withinwidefunnel, index);
            logger_.logouterleftfunnel(widerfunnelwidthparam.leftwidth, index);
            logger_.logouterrightfunnel(widerfunnelwidthparam.rightwidth, index);

            basetype::boolean_t useprediction{Selector::checkpredicition(objects.at(index))};

            // log useprediction
            logger_.loguseprediction(useprediction, index);

            if (isrearobject == boolean_t_false) {
                if ((iv_ > zero_objectid_t) &&
                    (objects.at(index).Ppty.FusionSource == ObjectFusionSource::CameraOnly)) {
                    basetype::float32_t distobj2obj{Util::abs(objects.at(index).Est.DisLgt - 
                                                              objects.at(iv_ - one_objectid_t).Est.DisLgt)};
                    basetype::float32_t perobj2obj{Util::abs(one_float32_t -
                                                   (Util::safetydiv(objects.at(index).Est.DisLgt, 
                                                                    objects.at(iv_ - one_objectid_t).Est.DisLgt, eps5_float32_t)))};
                    if ((distobj2obj < param_conf_.maxoverlapforcameraonlycar) ||
                        (perobj2obj < param_conf_.maxrelativeoverlapforcameraonlycar)) {
                        continue;
                    }
                }
                if (withinfunnel == boolean_t_true && objects.at(index).Info.Type != ObjectType::Pedestrain) {
                    if (iv_ == zero_objectid_t) {
                        iv_ = index + one_objectid_t;
                        ivoffset_ = offset;
                    } else if ((objects.at(index).Est.DisLgt > objects.at(iv_ - one_objectid_t).Est.DisLgt) &&
                                (Util::abs(objects.at(index).Est.DisLgt - objects.at(iv_ - one_objectid_t).Est.DisLgt) < param_conf_.maxoverlapforcameraonlycar) &&
                                (objects.at(index).Ppty.FusionSource == ObjectFusionSource::Fused) &&
                                (objects.at(iv_ - one_objectid_t).Ppty.FusionSource == ObjectFusionSource::CameraOnly) &&
                                (Util::abs(objects.at(index).Ppty.DisLgtStdDev - static_cast<basetype::float32_t>(299.0)) < static_cast<basetype::float32_t>(0.25))) {
                        iv_ = index + one_objectid_t;
                        ivoffset_ = offset;
                    } else if ((objects.at(index).Est.DisLgt > objects.at(iv_ - one_objectid_t).Est.DisLgt) &&
                               (Util::abs(objects.at(index).Est.DisLgt - objects.at(iv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                        if ((iiv_ == zero_objectid_t) &&
                            (objects.at(index).Info.Type != ObjectType::Motorcycle) &&
                            (objects.at(index).Info.Type != ObjectType::Bicycle)) {
                            iiv_ = index + one_objectid_t;
                            iivoffset_ = offset;
                        } else if ((iiv_ > zero_objectid_t) &&
                                   (objects.at(index).Est.DisLgt < objects.at(iiv_ - one_objectid_t).Est.DisLgt)) {
                            iiv_ = index + one_objectid_t;
                            iivoffset_ = offset;
                        }
                    } else if ((objects.at(index).Est.DisLgt < objects.at(iv_ - one_objectid_t).Est.DisLgt) &&
                               (Util::abs(objects.at(index).Est.DisLgt - objects.at(iv_ - one_objectid_t).Est.DisLgt) < param_conf_.maxoverlapforcameraonlycar) &&
                               (objects.at(index).Ppty.FusionSource == ObjectFusionSource::CameraOnly) &&
                               (objects.at(iv_ - one_objectid_t).Ppty.FusionSource == ObjectFusionSource::Fused) &&
                               (Util::abs(objects.at(index).Ppty.DisLgtStdDev - static_cast<basetype::float32_t>(299.0)) < static_cast<basetype::float32_t>(0.25))) {
                        continue;
                    } else if ((objects.at(index).Est.DisLgt < objects.at(iv_ - one_objectid_t).Est.DisLgt)) {
                        if ((objects.at(iv_ - one_objectid_t).Info.Type != ObjectType::Motorcycle) &&
                            (objects.at(iv_ - one_objectid_t).Info.Type != ObjectType::Bicycle) &&
                            (Util::abs(objects.at(index).Est.DisLgt - objects.at(iv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                            iiv_ = iv_;
                            iivoffset_ = ivoffset_;

                        } else {
                            iiv_ = zero_objectid_t;
                            iivoffset_ = zero_float32_t;
                        }
                        iv_ = index + one_objectid_t;
                        ivoffset_ = offset;
                    }  
                } else if (withinwidefunnel == boolean_t_true && objects.at(index).Info.Type != ObjectType::Pedestrain) {
                    if (offset > zero_float32_t) {
                        if (fliv_ == zero_objectid_t) {
                            fliv_ = index + one_objectid_t;
                            flivoffset_ = offset;
                        } else {
                            if (fliiv_ == zero_objectid_t) {
                                if ((objects.at(index).Est.DisLgt < objects.at(fliv_ - one_objectid_t).Est.DisLgt) &&
                                    (Util::abs(objects.at(index).Est.DisLgt - objects.at(fliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    fliiv_ = fliv_;
                                    fliivoffset_ = flivoffset_ ;
                                    fliv_ = index + one_objectid_t;
                                    flivoffset_ = offset;
                                } else if ((objects.at(index).Est.DisLgt > objects.at(fliv_ - one_objectid_t).Est.DisLgt) &&
                                           (Util::abs(objects.at(index).Est.DisLgt - objects.at(fliv_ - one_objectid_t).Est.DisLgt) >param_conf_.maxoverlapforcameraonlycar)) {
                                    fliiv_ = index + one_objectid_t;
                                    fliivoffset_ = offset;
                                }
                            } else if ((objects.at(index).Est.DisLgt < objects.at(fliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(fliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                fliiv_ = fliv_;
                                fliivoffset_ = flivoffset_ ;
                                fliv_ = index + one_objectid_t;
                                flivoffset_ = offset;
                            } else if ((objects.at(index).Est.DisLgt > objects.at(fliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(Selector::fliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar) &&
                                       (objects.at(index).Est.DisLgt < objects.at(fliiv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(fliiv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                fliiv_ = index + one_objectid_t;
                                fliivoffset_ = offset;
                            }
                        }

                    } else {
                        if (sliv_ == zero_objectid_t) {
                            sliv_ = index + one_objectid_t;
                            slivoffset_ = offset;
                        } else {
                            if (sliiv_ == zero_objectid_t) {
                                if ((objects.at(index).Est.DisLgt < objects.at(sliv_ - one_objectid_t).Est.DisLgt) &&
                                    (Util::abs(objects.at(index).Est.DisLgt - objects.at(sliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    sliiv_ = sliv_;
                                    sliivoffset_ = slivoffset_ ;
                                    sliv_ = index + one_objectid_t;
                                    slivoffset_ = offset;
                                }
                                else if ((objects.at(index).Est.DisLgt > objects.at(sliv_ - one_objectid_t).Est.DisLgt) &&
                                         (Util::abs(objects.at(index).Est.DisLgt - objects.at(sliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    sliiv_ = index + one_objectid_t;
                                    sliivoffset_ = offset;
                                }
                            } else if ((objects.at(index).Est.DisLgt < objects.at(sliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(sliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                sliiv_ = sliv_;
                                sliivoffset_ = slivoffset_ ;
                                sliv_ = index + one_objectid_t;
                                slivoffset_ = offset;
                            } else if ((objects.at(index).Est.DisLgt > objects.at(sliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(sliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar) &&
                                       (objects.at(index).Est.DisLgt < objects.at(sliiv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(sliiv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                sliiv_ = index + one_objectid_t;
                                sliivoffset_ = offset;
                            }
                        }
                    }

                    if ((cutin_ == zero_objectid_t) ||
                        (cutin_ > zero_objectid_t && 
                         objects.at(index).Est.DisLgt < objects.at(cutin_ - one_objectid_t).Est.DisLgt)) {
                        basetype::float32_t offsetderiv{offsetcalculator_.getoffsetderiv()};

                        basetype::float32_t predictedoffset{offset + offsetderiv * param_conf_.cutinpredictiontime};

                        if (std::signbit(offset) != std::signbit(offsetderiv)) {
                            if (Selector::checkcutin(index, isrearobject, useprediction, predictedoffset) == boolean_t_true) {
                                cutin_ = index + one_objectid_t;
                                cutinoffset_ = offset;
                            }
                        }
                    }
                } else if (withinfunnel == boolean_t_true && objects.at(index).Info.Type == ObjectType::Pedestrain) {
                    if (pedestrain_ > zero_objectid_t) {
                        if ((objects.at(index).Est.DisLgt < objects.at(Selector::pedestrain_ - one_objectid_t).Est.DisLgt)) {
                            pedestrain_ = index + one_objectid_t;
                        }
                    } else {
                        pedestrain_ = index + one_objectid_t;
                    }
                }
            } else {
                if ((riv_ > zero_objectid_t) &&
                    (objects.at(index).Ppty.FusionSource == ObjectFusionSource::CameraOnly)) {
                    basetype::float32_t distobj2obj{Util::abs(objects.at(index).Est.DisLgt - 
                                                              objects.at(riv_ - one_objectid_t).Est.DisLgt)};
                    basetype::float32_t perobj2obj{Util::abs(one_float32_t -
                                                   (Util::safetydiv(objects.at(index).Est.DisLgt, 
                                                                    objects.at(riv_ - one_objectid_t).Est.DisLgt, eps5_float32_t)))};
                    if ((distobj2obj < param_conf_.maxoverlapforcameraonlycar) ||
                        (perobj2obj < param_conf_.maxrelativeoverlapforcameraonlycar)) {
                        continue;
                    }
                }
                if (withinfunnel == boolean_t_true && objects.at(index).Info.Type != ObjectType::Pedestrain) {
                    if (riv_ == zero_objectid_t) {
                        riv_ = index + one_objectid_t;
                        rivoffset_ = offset;
                    } else if ((objects.at(index).Est.DisLgt < objects.at(riv_ - one_objectid_t).Est.DisLgt) &&
                             (Util::abs(objects.at(index).Est.DisLgt - objects.at(riv_ - one_objectid_t).Est.DisLgt) < param_conf_.maxoverlapforcameraonlycar) &&
                             (objects.at(index).Ppty.FusionSource == ObjectFusionSource::Fused) &&
                             (objects.at(riv_ - one_objectid_t).Ppty.FusionSource == ObjectFusionSource::CameraOnly) &&
                             (Util::abs(objects.at(index).Ppty.DisLgtStdDev - static_cast<basetype::float32_t>(299.0)) < static_cast<basetype::float32_t>(0.25))) {
                        riv_ = index + one_objectid_t;
                        rivoffset_ = offset;
                    } else if ((objects.at(index).Est.DisLgt < objects.at(riv_ - one_objectid_t).Est.DisLgt) &&
                               (Util::abs(objects.at(index).Est.DisLgt - objects.at(riv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                        if ((riiv_ == zero_objectid_t) &&
                            (objects.at(index).Info.Type != ObjectType::Motorcycle) &&
                            (objects.at(index).Info.Type != ObjectType::Bicycle)) {
                            riiv_ = index + one_objectid_t;
                            riivoffset_ = offset;
                        } else if ((riiv_ > zero_objectid_t) &&
                                   (objects.at(index).Est.DisLgt > objects.at(riiv_ - one_objectid_t).Est.DisLgt)) {
                            riiv_ = index + one_objectid_t;
                            riivoffset_ = offset;
                        }
                    } else if ((objects.at(index).Est.DisLgt > objects.at(riv_ - one_objectid_t).Est.DisLgt) &&
                             (Util::abs(objects.at(index).Est.DisLgt - objects.at(riv_ - one_objectid_t).Est.DisLgt) < param_conf_.maxoverlapforcameraonlycar) &&
                             (objects.at(index).Ppty.FusionSource == ObjectFusionSource::CameraOnly) &&
                             (objects.at(riv_ - one_objectid_t).Ppty.FusionSource == ObjectFusionSource::Fused) &&
                             (Util::abs(objects.at(index).Ppty.DisLgtStdDev - static_cast<basetype::float32_t>(299.0)) < static_cast<basetype::float32_t>(0.25))) {
                        continue;
                    } else if ((objects.at(index).Est.DisLgt > objects.at(riv_ - one_objectid_t).Est.DisLgt)) {
                        if ((objects.at(riv_ - one_objectid_t).Info.Type != ObjectType::Motorcycle) &&
                            (objects.at(riv_ - one_objectid_t).Info.Type != ObjectType::Bicycle) &&
                            (Util::abs(objects.at(index).Est.DisLgt - objects.at(riv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                            riiv_ = riv_;
                            riivoffset_ = rivoffset_;

                        } else {
                            riiv_ = zero_objectid_t;
                            riivoffset_ = zero_float32_t;
                        }
                        riv_ = index + one_objectid_t;
                        rivoffset_ = offset;
                    }  
                } else if (withinwidefunnel == boolean_t_true && objects.at(index).Info.Type != ObjectType::Pedestrain) {
                    if (offset > zero_float32_t) {
                        if (rfliv_ == zero_objectid_t) {
                            rfliv_ = index + one_objectid_t;
                            rflivoffset_ = offset;
                        } else {
                            if (rfliiv_ == zero_objectid_t) {
                                if ((objects.at(index).Est.DisLgt > objects.at(rfliv_ - one_objectid_t).Est.DisLgt) &&
                                    (Util::abs(objects.at(index).Est.DisLgt - objects.at(rfliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    rfliiv_ = rfliv_;
                                    rfliivoffset_ = rflivoffset_ ;
                                    rfliv_ = index + one_objectid_t;
                                    rflivoffset_ = offset;
                                } else if ((objects.at(index).Est.DisLgt < objects.at(rfliv_ - one_objectid_t).Est.DisLgt) &&
                                           (Util::abs(objects.at(index).Est.DisLgt - objects.at(rfliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    rfliiv_ = index + one_objectid_t;
                                    rfliivoffset_ = offset;
                                }
                            } else if ((objects.at(index).Est.DisLgt > objects.at(rfliiv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(rfliiv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                rfliiv_ = rfliiv_;
                                rfliivoffset_ = rfliivoffset_ ;
                                rfliiv_ = index + one_objectid_t;
                                rfliivoffset_ = offset;
                            } else if ((objects.at(index).Est.DisLgt < objects.at(rfliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(rfliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar) &&
                                       (objects.at(index).Est.DisLgt > objects.at(rfliiv_ - one_objectid_t).Est.DisLgt) &&
                                     (Util::abs(objects.at(index).Est.DisLgt - objects.at(rfliiv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                rfliiv_ = index + one_objectid_t;
                                rfliivoffset_ = offset;
                            }
                        }

                    } else {
                        if (rsliv_ == zero_objectid_t) {
                            rsliv_ = index + one_objectid_t;
                            rslivoffset_ = offset;
                        } else {
                            if (rsliiv_ == zero_objectid_t) {
                                if ((objects.at(index).Est.DisLgt > objects.at(rsliv_ - one_objectid_t).Est.DisLgt) &&
                                    (Util::abs(objects.at(index).Est.DisLgt - objects.at(rsliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    rsliiv_ = rsliv_;
                                    rsliivoffset_ = rslivoffset_ ;
                                    rsliv_ = index + one_objectid_t;
                                    rslivoffset_ = offset;
                                } else if ((objects.at(index).Est.DisLgt < objects.at(rsliv_ - one_objectid_t).Est.DisLgt) &&
                                           (Util::abs(objects.at(index).Est.DisLgt - objects.at(rsliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                    rsliiv_ = index + one_objectid_t;
                                    rsliivoffset_ = offset;
                                }
                            } else if ((objects.at(index).Est.DisLgt > objects.at(rsliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(rsliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                rsliiv_ = rsliv_;
                                rsliivoffset_ = rslivoffset_ ;
                                rsliv_ = index + one_objectid_t;
                                rslivoffset_ = offset;
                            } else if ((objects.at(index).Est.DisLgt < objects.at(rsliv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(rsliv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar) &&
                                       (objects.at(index).Est.DisLgt > objects.at(rsliiv_ - one_objectid_t).Est.DisLgt) &&
                                       (Util::abs(objects.at(index).Est.DisLgt - objects.at(rsliiv_ - one_objectid_t).Est.DisLgt) > param_conf_.maxoverlapforcameraonlycar)) {
                                rsliiv_ = index + one_objectid_t;
                                rsliivoffset_ = offset;
                            }
                        }
                    }
                    if ((rcutin_ == zero_objectid_t) ||
                        (rcutin_ > zero_objectid_t && 
                         objects.at(index).Est.DisLgt > objects.at(rcutin_ - one_objectid_t).Est.DisLgt)) {
                        basetype::float32_t offsetderiv{offsetcalculator_.getoffsetderiv()};

                        basetype::float32_t predictedoffset{offset + offsetderiv * param_conf_.cutinpredictiontime};

                        if (std::signbit(offset) != std::signbit(offsetderiv)) {
                            if (Selector::checkcutin(index, isrearobject, useprediction, predictedoffset) == boolean_t_true) {
                                rcutin_ = index + one_objectid_t;
                                rcutinoffset_ = offset;
                            }
                        }
                    }
                }
            }    
        }
        // end loop

        if ((iv_ > zero_objectid_t) &&
            (cutin_> zero_objectid_t) &&
            (objects.at(iv_ - one_objectid_t).Est.DisLgt < objects.at(cutin_ - one_objectid_t).Est.DisLgt)) {
            cutin_ = zero_objectid_t;
            cutinoffset_ = zero_float32_t;
        }

        if ((riv_ > zero_objectid_t) &&
            (rcutin_> zero_objectid_t) &&
            (objects.at(riv_ - one_objectid_t).Est.DisLgt > objects.at(rcutin_ - one_objectid_t).Est.DisLgt)) {
            rcutin_ = zero_objectid_t;
            rcutinoffset_ = zero_float32_t;
        }
        
        if (iv_  == zero_objectid_t) {
            ivquality_ = zero_int8_t;

            predictionmaxdistance_ = param_conf_.cutinmaxdistance;
        } else {
            funnel_.prepareobject(isrearobject,
                                  objects.at(iv_ - one_objectid_t).Est.DisLgt,
                                  (iv_ - one_objectid_t),
                                  road_.getlanewidth(objects.at(iv_ - one_objectid_t).Est.DisLgt));
            FunnelWidthParam funnelwidthparam2{};
            basetype::boolean_t withinfunnel2{Selector::checkinpath((iv_ - one_objectid_t),
                                                                     isrearobject,
                                                                     ivoffset_,
                                                                     funnelwidthparam2)};
            // unuesd var
            withinfunnel2 = withinfunnel2;
            offsetcalculator_.calculate(objects.at(iv_ - one_objectid_t),
                                         isrearobject,
                                         road_.getisoffsetfromlanecenterusable(),
                                         road_.getisoffsetfromlanecentervalid(),
                                         road_.getlaneoffset(),
                                         egovehinfo,
                                         c0calculator_.get_C0(),
                                         useyawplusinlanechange);

            ivquality_ = static_cast<basetype::int8_t>(2u);

            basetype::boolean_t useprediction2{Selector::checkpredicition(objects.at(Selector::iv_ - one_objectid_t))};
            if (useprediction2 == boolean_t_true) {
                basetype::float32_t offset2{offsetcalculator_.getoffset()};
                // unuesd var
                offset2 = offset2;
                basetype::float32_t offsetderiv2{offsetcalculator_.getoffsetderiv()};

                ivcutingout_ = Selector::checkcutout((iv_ - one_objectid_t),
                                                     isrearobject,
                                                     objects.at(iv_ - one_objectid_t).Est.DisLgt,
                                                     egovehinfo.SpdLgt,
                                                     ivoffset_,
                                                     offsetderiv2);
            }

            predictionmaxdistance_ = objects.at(Selector::iv_ - one_objectid_t).Est.DisLgt + param_conf_.cutinivdistanceoffset;
        }

        basetype::float32_t distanceiv2iiv{255.0f};
        if (iiv_ > zero_objectid_t) {
            distanceiv2iiv = Util::abs(objects.at(iv_ - one_objectid_t).Est.DisLgt - objects.at(iiv_ - one_objectid_t).Est.DisLgt);
        }

        if ((iiv_ > zero_objectid_t) &&
            (objects.at(iv_ - one_objectid_t).Ppty.FusionSource == ObjectFusionSource::CameraOnly) &&
            (objects.at(iiv_ - one_objectid_t).Ppty.FusionSource != ObjectFusionSource::CameraOnly)) {
            if (distanceiv2iiv < param_conf_.maxoverlapforcameraonlycar) {
                iv_ = iiv_;
                ivoffset_ = iivoffset_;
                iiv_ = zero_objectid_t;
                iivoffset_ = zero_float32_t;
            }
        }
        distanceiv2iiv = 255.0f;
        if (riiv_ > zero_objectid_t) {
            distanceiv2iiv = Util::abs(objects.at(riv_ - one_objectid_t).Est.DisLgt - objects.at(riiv_ - one_objectid_t).Est.DisLgt);
        }

        if ((riiv_ > zero_objectid_t) &&
            (objects.at(riv_ - one_objectid_t).Ppty.FusionSource == ObjectFusionSource::CameraOnly) &&
            (objects.at(riiv_ - one_objectid_t).Ppty.FusionSource != ObjectFusionSource::CameraOnly)) {
            if (distanceiv2iiv < param_conf_.maxoverlapforcameraonlycar) {
                riv_ = riiv_;
                rivoffset_ = riivoffset_;
                riiv_ = zero_objectid_t;
                riivoffset_ = zero_float32_t;
            }
        }

        if (iv_ > zero_objectid_t && pedestrain_ > zero_objectid_t) {
            if ((objects.at(iv_ - one_objectid_t).Est.DisLgt + 1.0f) < objects.at(pedestrain_ - one_objectid_t).Est.DisLgt) {
                pedestrain_ = zero_objectid_t;
            }
        }
        // log iv iiv fliv sliv fliiv sliiv cutin ivquality
        logger_.logpedestrain(pedestrain_);
        logger_.logiv(iv_);
        logger_.logiiv(iiv_);
        logger_.logfliv(fliv_);
        logger_.logsliv(sliv_);
        logger_.logfliiv(fliiv_);
        logger_.logsliiv(sliiv_);
        logger_.logcutin(cutin_);
        logger_.logriv(riv_);
        logger_.logriiv(riiv_);
        logger_.logrfliv(rfliv_);
        logger_.logrsliv(rsliv_);
        logger_.logrfliiv(rfliiv_);
        logger_.logrsliiv(rsliiv_);
        logger_.logrcutin(rcutin_);
        logger_.logivquality(ivquality_);

    }

    void Selector::reset(const basetype::objectid_t idn0) {
        inpathcounter_.at(idn0)     = zero_uint8_t;
        inwidepathcounter_.at(idn0) = zero_uint8_t;
        cutinpathcounters_.at(idn0) = zero_uint8_t;
        cutinholdcounters_.at(idn0) = zero_uint8_t;
    }

    basetype::objectid_t Selector::getiv() {
        return iv_;
    }

    basetype::objectid_t Selector::getiiv() {
        return iiv_;
    }

    basetype::objectid_t Selector::getfliv() {
        return fliv_;
    }

    basetype::objectid_t Selector::getsliv() {
        return sliv_;
    }

    basetype::objectid_t Selector::getfliiv() {
        return fliiv_;
    }

    basetype::objectid_t Selector::getsliiv() {
        return sliiv_;
    }

    basetype::objectid_t Selector::getcutin() {
        return cutin_;
    }

    basetype::float32_t Selector::getivoffset() {
        return ivoffset_;
    }

    basetype::float32_t Selector::getiivoffset() {
        return iivoffset_;
    }

    basetype::float32_t Selector::getflivoffset() {
        return flivoffset_;
    }

    basetype::float32_t Selector::getslivoffset() {
        return slivoffset_;
    }

    basetype::float32_t Selector::getfliivoffset() {
        return fliivoffset_;
    }

    basetype::float32_t Selector::getsliivoffset() {
        return sliivoffset_;
    }

    basetype::float32_t Selector::getcutinoffset() {
        return cutinoffset_;
    }

    basetype::objectid_t Selector::getriv() {
        return riv_;
    }

    basetype::objectid_t Selector::getriiv() {
        return riiv_;
    }

    basetype::objectid_t Selector::getrfliv() {
        return rfliv_;
    }

    basetype::objectid_t Selector::getrsliv() {
        return sliv_;
    }

    basetype::objectid_t Selector::getrfliiv() {
        return rfliiv_;
    }

    basetype::objectid_t Selector::getrsliiv() {
        return rsliiv_;
    }

    basetype::objectid_t Selector::getrcutin() {
        return rcutin_;
    }

    basetype::float32_t Selector::getrivoffset() {
        return rivoffset_;
    }

    basetype::float32_t Selector::getriivoffset() {
        return riivoffset_;
    }

    basetype::float32_t Selector::getrflivoffset() {
        return rflivoffset_;
    }

    basetype::float32_t Selector::getrslivoffset() {
        return rslivoffset_;
    }

    basetype::float32_t Selector::getrfliivoffset() {
        return rfliivoffset_;
    }

    basetype::float32_t Selector::getrsliivoffset() {
        return rsliivoffset_;
    }

    basetype::float32_t Selector::getrcutinoffset() {
        return rcutinoffset_;
    }

    basetype::int8_t Selector::getivquality() {
        return ivquality_;
    }

    ObjectLaneStatus Selector::getivcuttingoutsts() {
        return ivcutoutstatus_;
    }

    basetype::boolean_t Selector::getivcuttingout() {
        return ivcutingout_;
    }
}