#include "validator.h"
#include <cmath>
#include <iostream>
#include "atsutil.h"

namespace ACC_Target_Selection{

    Validator::Validator(const Param_Validator &param_conf, const Logger &logger) :
            param_conf_(param_conf),
            logger_(logger),
            pretgtaid_(zero_objectid_t),
            pretgtaspd_(zero_float32_t),
            pretgtalgtposition_(zero_float32_t),
            pretgtalatposition_(zero_float32_t) {
        for (basetype::objectid_t i = zero_objectid_t; i < c_ObjectSize; i++) {
            maturitycounter_.at(i)   = zero_uint8_t;
            coastcounter_.at(i)      = zero_uint8_t;
            jumpcounter_.at(i)       = zero_uint8_t;
            qualitycounter_.at(i)    = zero_uint8_t;
            lgtpositionmemory_.at(i) = static_cast<basetype::float32_t>(255.0f);
            latpositionmemory_.at(i) = static_cast<basetype::float32_t>(50.0f);;
            spdmemory_.at(i)         = static_cast<basetype::float32_t>(255.0f);;
        }
    }

    Validator::Validator(const Validator& r) :
            param_conf_(r.param_conf_),
            logger_(r.logger_),
            pretgtaid_(r.pretgtaid_),
            pretgtaspd_(r.pretgtaspd_),
            pretgtalgtposition_(r.pretgtalgtposition_),
            pretgtalatposition_(r.pretgtalatposition_),
            maturitycounter_(r.maturitycounter_),
            coastcounter_(r.coastcounter_),
            jumpcounter_(r.jumpcounter_),
            qualitycounter_(r.qualitycounter_),
            lgtpositionmemory_(r.lgtpositionmemory_),
            latpositionmemory_(r.latpositionmemory_),
            spdmemory_(r.spdmemory_) {
    }
    
    Validator::~Validator() {
    }

    basetype::boolean_t Validator::updatecoastcounter(const ObjectTrackStatus &trackstatus, const basetype::objectid_t &idn0) {
        basetype::boolean_t res{boolean_t_false};

        basetype::float32_t maxcoastingsamples{static_cast<basetype::uint8_t>(Util::min(Util::safetydiv(param_conf_.maxcoastingtime,
                                                                                                        c_SampleTime, eps3_float32_t),
                                                                                        254.0f))};

        if (ObjectTrackStatus::Predict == trackstatus) {
            if (coastcounter_.at(idn0) <= maxcoastingsamples) {
                coastcounter_.at(idn0) = coastcounter_.at(idn0) + one_uint8_t;
            }
        } else {
            Validator::coastcounter_.at(idn0) = zero_uint8_t;
        }

        if ((ObjectTrackStatus::Predict == trackstatus) &&
            (Validator::coastcounter_.at(idn0) <= maxcoastingsamples)) {
            res = boolean_t_true;
        }
        
        return res;
    }

    basetype::boolean_t Validator::updatejumpmaturity(const basetype::float32_t &poslgt,
                                                      const basetype::float32_t &poslat,
                                                      const basetype::float32_t &spd,
                                                      const basetype::objectid_t &idn0) {
        basetype::boolean_t res = boolean_t_false;


        if (jumpcounter_.at(idn0) >= param_conf_.minjumpthreshold) {
            res = boolean_t_true;
        }

        basetype::float32_t poslgtdelta = Util::abs(poslgt - lgtpositionmemory_.at(idn0));

        basetype::float32_t poslatdelta = Util::abs(poslat - latpositionmemory_.at(idn0));

        if ((poslgtdelta < param_conf_.maxposjumpthreshold) &&
            (poslatdelta < param_conf_.maxposjumpthreshold)) {
            if (jumpcounter_.at(idn0) <= param_conf_.minjumpthreshold) {
                jumpcounter_.at(idn0) = jumpcounter_.at(idn0) + one_uint8_t;
            }
        } else {
            jumpcounter_.at(idn0) = zero_uint8_t;
        }
        
        lgtpositionmemory_.at(idn0) = poslgt;
        latpositionmemory_.at(idn0) = poslat;
        spdmemory_.at(idn0) = spd;

        
        return res;
    }

    basetype::boolean_t Validator::updatematurity(const basetype::boolean_t &detected,
                                                  const basetype::float32_t &spd,
                                                  const basetype::float32_t &poslgt,
                                                  const ObjectFusionSource &fusionsource,
                                                  const basetype::float32_t &angle,
                                                  const basetype::objectid_t &idn0) {
        basetype::boolean_t res{boolean_t_false};

        std::vector<basetype::float32_t> stddevthresholdbps{param_conf_.stddevthresholdbps};

        std::vector<basetype::uint8_t> maturitythreshold(9, zero_uint8_t);
        basetype::uint8_t threshold;
        if ((fusionsource == ObjectFusionSource::CameraOnly) ||
            (fusionsource == ObjectFusionSource::Fused)) {
            maturitythreshold = param_conf_.minmaturitythreshold;
            threshold = Util::lookup_u8_float(maturitythreshold, stddevthresholdbps, poslgt, boolean_t_false);
        } else {
            maturitythreshold = param_conf_.minmaturitythreshold2;
            threshold = Util::lookup_u8_float(maturitythreshold, stddevthresholdbps, poslgt, boolean_t_false);
        }
        
        if (maturitycounter_.at(idn0) >= threshold) {
            res = boolean_t_true;
        }

        if (detected == boolean_t_true) {
            if (maturitycounter_.at(idn0) < param_conf_.maxmaturitythreshold) {
                maturitycounter_.at(idn0) = maturitycounter_.at(idn0) + one_uint8_t;
            }
        } else {
            Validator::maturitycounter_.at(idn0) = zero_uint8_t;
        }
        return res;
    }

    void Validator::reset(const basetype::objectid_t &idn0) {
        if (maturitycounter_.at(idn0) > zero_uint8_t) {
            maturitycounter_.at(idn0) = zero_uint8_t;
            jumpcounter_.at(idn0)     = zero_uint8_t;
            qualitycounter_.at(idn0)  = zero_uint8_t;
            coastcounter_.at(idn0)    = zero_uint8_t;
        } else {
        }
    }

    void Validator::setprevtgta(const basetype::objectid_t &tgtaidn, const basetype::objectid_t &rtgtaidn) {
        if (tgtaidn > zero_objectid_t) {
            pretgtaid_ = tgtaidn;
            pretgtaspd_ = spdmemory_.at(tgtaidn - one_objectid_t);
            pretgtalgtposition_ = lgtpositionmemory_.at(tgtaidn - one_objectid_t);
            pretgtalatposition_ = latpositionmemory_.at(tgtaidn - one_objectid_t);
        } else {
            pretgtaid_ = zero_objectid_t;
            pretgtaspd_ = zero_float32_t;
            pretgtalgtposition_ = static_cast<basetype::float32_t>(255.0f);
            pretgtalatposition_ = static_cast<basetype::float32_t>(50.0f);
        }
        if (rtgtaidn > zero_objectid_t) {
            prereartgtaid_ = rtgtaidn;
            prereartgtaspd_ = spdmemory_.at(rtgtaidn - one_objectid_t);
            prereartgtalgtposition_ = lgtpositionmemory_.at(rtgtaidn - one_objectid_t);
            prereartgtalatposition_ = latpositionmemory_.at(rtgtaidn - one_objectid_t);
        } else {
            pretgtaid_ = zero_objectid_t;
            pretgtaspd_ = zero_float32_t;
            pretgtalgtposition_ = static_cast<basetype::float32_t>(255.0f);
            pretgtalatposition_ = static_cast<basetype::float32_t>(50.0f);
        }
    }

    basetype::boolean_t Validator::checkvalid(const Object &obj, const EgoVehicleInformation &egovehinfo, const basetype::boolean_t &isrearobject) {
        basetype::boolean_t res{boolean_t_false};

        basetype::objectid_t idn{(obj.Ppty.Id) - one_objectid_t};

        basetype::boolean_t detected{boolean_t_false};
        if (obj.Ppty.TrackStatus != ObjectTrackStatus::Invalid) {
            detected = boolean_t_true;
        } else {
        }
        
        basetype::boolean_t maturityok{Validator::updatematurity(detected,
                                                                 obj.Est.SpdLgt,
                                                                 obj.Est.DisLgt,
                                                                 obj.Ppty.FusionSource,
                                                                 obj.Est.HeadingAngle,
                                                                 idn)};
        // log maturityok maturitycounters
        logger_.logmaturityok(maturityok, idn);
        logger_.logmaturitycounters(Validator::maturitycounter_.at(idn), idn);

        basetype::boolean_t jumpok{Validator::updatejumpmaturity(obj.Est.DisLgt,
                                                                 obj.Est.DisLat,
                                                                 obj.Est.Spd,
                                                                 idn)};
        // log jumpok jumcounters
        logger_.logjumpok(jumpok, idn);
        logger_.logjumpcounter(Validator::jumpcounter_.at(idn), idn);

        basetype::boolean_t coastedok{Validator::updatecoastcounter(obj.Ppty.TrackStatus,
                                                                    idn)};
        
        // log coastedok coastedcounters
        logger_.logcoastedok(coastedok, idn);
        logger_.logcoastedcounter(coastcounter_.at(idn), idn);

        basetype::boolean_t qualityok{boolean_t_false};
        if (((param_conf_.disableposnqly == boolean_t_false) &&
            (obj.Ppty.FusionSource == ObjectFusionSource::RadarOnly)) ||
            (obj.Ppty.FusionSource == ObjectFusionSource::Fused)) {
            qualityok = boolean_t_true;
        } else {
        }
        // log qualityok
        logger_.logqualityok(qualityok, idn);

        basetype::boolean_t isvalid_pre{Validator::checkvalidpre(maturityok,
                                                                 jumpok,
                                                                 qualityok,
                                                                 coastedok,
                                                                 isrearobject,
                                                                 obj.Ppty.TrackStatus,
                                                                 obj.Est.DisLgt,
                                                                 obj.Ppty.DisLgtStdDev,
                                                                 obj.Ppty.DisLatStdDev,
                                                                 obj.Ppty.AccelStdDev,
                                                                 obj.Ppty.HeadingAngleStdDev)};
        
        basetype::boolean_t isvalidtypeandstationary{Validator::checkisvalidtypeandstationary(obj.Ppty.Id,
                                                                                              obj.Info.Type,
                                                                                              obj.Info.MotionType,
                                                                                              obj.Info.MotionHistoryType,
                                                                                              egovehinfo.SpdLgt)};
        
        basetype::boolean_t useofanyobject{Validator::enableuseofanyobject(isrearobject,
                                                                           obj.Est.DisLgt,
                                                                           obj.Info.MotionType,
                                                                           obj.Info.MotionHistoryType,
                                                                           egovehinfo.SpdLgt)};
        
        basetype::boolean_t isvalid_stationary{boolean_t_false};

        if (isvalidtypeandstationary ==boolean_t_true || useofanyobject == boolean_t_true) {
            isvalid_stationary = boolean_t_true;
        } else {
        }

        basetype::boolean_t isvalidmoving{Validator::checkisvalidmoving(obj.Ppty.Id,
                                                                        obj.Info.Type,
                                                                        obj.Info.MotionType,
                                                                        obj.Info.MotionHistoryType,
                                                                        obj.Est.Spd,
                                                                        egovehinfo.SpdLgt)};
        
        basetype::boolean_t isvalidstationaryormoving{boolean_t_false};
        if (isrearobject == boolean_t_false) {
            if ((isvalidmoving == boolean_t_true || isvalid_stationary == boolean_t_true) &&
                (obj.Est.DisLgt > param_conf_.mindistanceforstationaryormoving)) {
                isvalidstationaryormoving = boolean_t_true;
            } else {
            }
        } else {
            if ((isvalidmoving == boolean_t_true || isvalid_stationary == boolean_t_true) &&
                (obj.Est.DisLgt > param_conf_.maxreardistanceforstationaryormoving)) {
                isvalidstationaryormoving = boolean_t_true;
            } else {
            }
        }
        

        // log isvalidmoving isvalidstationaryormoving
        logger_.logisvalidstationarytarget(isvalidstationaryormoving, idn);
        logger_.logisvalidmovingtarget(isvalidmoving, idn);
        
        basetype::boolean_t isvalidcameramotorcycle{Validator::checkisvalidcameramotorcycle(isrearobject,
                                                                                            obj.Info.Type,
                                                                                            obj.Est.DisLgt,
                                                                                            obj.Info.MotionType,
                                                                                            obj.Ppty.TrackStatus,
                                                                                            obj.Ppty.FusionSource,
                                                                                            jumpok,
                                                                                            maturityok)};

        basetype::boolean_t isvalidcameracar{Validator::checkisvalidcameracar(idn,
                                                                              isrearobject,
                                                                              obj.Info.Type,
                                                                              obj.Est.DisLgt,
                                                                              obj.Ppty.DisLgtStdDev,
                                                                              obj.Info.MotionType,
                                                                              obj.Ppty.TrackStatus,
                                                                              obj.Ppty.FusionSource,
                                                                              jumpok,
                                                                              maturityok,
                                                                              egovehinfo.SpdLgt)};

        basetype::float32_t maxselectionrange{Util::lookup(param_conf_.maxdynamicselectionrange,
                                                           param_conf_.maxdynamicselectionrangespeedbp,
                                                           egovehinfo.SpdLgt,
                                                           boolean_t_false)};
        if ((obj.Est.DisLgt < maxselectionrange) ||
            ((obj.Est.DisLgt < Util::max(maxselectionrange * 1.1f, maxselectionrange + 5.0f)) 
              && idn == pretgtaid_ && pretgtaid_ > zero_objectid_t)) {
            if ((isvalid_pre == boolean_t_true && isvalidstationaryormoving == boolean_t_true) ||
                (isvalidcameracar == boolean_t_true) ||
                (isvalidcameramotorcycle == boolean_t_true)) {
                res = boolean_t_true;
            } else {
                res = boolean_t_false;
            }
        } else {
            res = boolean_t_false;
        }

        if (boolean_t_true == checkisvalidpedestrain(obj, egovehinfo, isrearobject, isvalid_pre)) {
            res = boolean_t_true;
        } else {
        }
        
        return res;
    }

    basetype::boolean_t Validator::checkvalidpre(const basetype::boolean_t &maturityok,
                                                 const basetype::boolean_t &jumpok,
                                                 const basetype::boolean_t &qualityok,
                                                 const basetype::boolean_t &coastedok,
                                                 const basetype::boolean_t &isrearobject,
                                                 const ObjectTrackStatus &trackstatus,
                                                 const basetype::float32_t &objlgtdis,
                                                 const basetype::float32_t &objdisLgtstddev,
                                                 const basetype::float32_t &objdisLatstddev,
                                                 const basetype::float32_t &objaccelstddev,
                                                 const basetype::float32_t &objheadinganglestddev) {
        basetype::boolean_t res{boolean_t_false};
        if (((maturityok == boolean_t_true) && (jumpok == boolean_t_true) && (qualityok == boolean_t_true)) &&
            ((coastedok == boolean_t_true) || (trackstatus != ObjectTrackStatus::Invalid)) &&
            (((objlgtdis >= param_conf_.mintargetdistance) && (isrearobject == boolean_t_false)) ||
             ((objlgtdis >= param_conf_.maxtargetreardistance) && (isrearobject == boolean_t_true))) &&
            (objdisLgtstddev <= param_conf_.lgtdistancedevthreshold) &&
            (objdisLatstddev <= param_conf_.latdistancedevthreshold) &&
            (objaccelstddev<= param_conf_.acceldevthreshold) &&
            (objheadinganglestddev <= param_conf_.headingangledevthreshold)) {
            res = boolean_t_true;
        } else {
        }
         return res;
    }

    basetype::boolean_t Validator::checkisvalidtypeandstationary(const basetype::objectid_t &idn,
                                                                 const ObjectType &type,
                                                                 const ObjectMotionType &motiontype,
                                                                 const ObjectMotionHistoryType &motionhistorytype,
                                                                 const basetype::float32_t &egospd) {
        basetype::boolean_t res{boolean_t_false};
        if (((type == ObjectType::Car) ||
             (type == ObjectType::Motorcycle) ||
             (type == ObjectType::Truck) ||
             (type == ObjectType::UnknownVehicleClass) ||
             (type == ObjectType::Bicycle)) &&
            ((motiontype == ObjectMotionType::Stationary) &&
             (motionhistorytype != ObjectMotionHistoryType::Crossing) &&
             ((motionhistorytype != ObjectMotionHistoryType::Oncoming) ||
              (motionhistorytype == ObjectMotionHistoryType::Oncoming && (idn == pretgtaid_ ||
                                                                          type == ObjectType::Truck)))) &&
            (egospd < param_conf_.maxegospdforstationaryobject)) {
            res = boolean_t_true;
        } else {
        }
        
        return res;
    }

    basetype::boolean_t Validator::enableuseofanyobject(const basetype::boolean_t &isrearobject,
                                                        const basetype::float32_t &lgtdis,
                                                        const ObjectMotionType &motiontype,
                                                        const ObjectMotionHistoryType &motionhistorytype,
                                                        const basetype::float32_t &egospd) {
        basetype::boolean_t res = boolean_t_false;
        if (((motiontype == ObjectMotionType::Stationary) ||
             (motionhistorytype == ObjectMotionHistoryType::Stationary)) &&
            (lgtdis < param_conf_.useallobjectsmaxobjectdistance) &&
            (isrearobject == boolean_t_false) &&
            (egospd <= param_conf_.useallobjectsmaxegospd)) {
            res = boolean_t_true;
        } else {
        }
        
        return res;
    }

    basetype::boolean_t Validator::checkisvalidmoving(const basetype::objectid_t &idn,
                                                      const ObjectType &type,
                                                      const ObjectMotionType &motiontype,
                                                      const ObjectMotionHistoryType &motionhistorytype,
                                                      const basetype::float32_t &objspd,
                                                      const basetype::float32_t &egospd) {
        basetype::boolean_t res{boolean_t_false};
        
        if (((motiontype == ObjectMotionType::Receeding) ||
             (motionhistorytype == ObjectMotionHistoryType::Receeding) ||
             (motiontype == ObjectMotionType::Oncoming && idn == prereartgtaid_ && objspd < zero_float32_t && param_conf_.usefixforoncomingwithnegativespeed == boolean_t_true)) &&
            ((type != ObjectType::Unknown || egospd < param_conf_.minspdforignoringunknownobject) ||
             ((egospd - objspd) < param_conf_.maxdifferentspdunknownobject2ego) ||
             (egospd > param_conf_.minspeedofunknown))) {
            res = boolean_t_true;
        } else {
        }
        
        if (type == ObjectType::Pedestrain) {
            res = boolean_t_false;
        } else {
        }

        if ((res == boolean_t_true) &&
            (!(type == ObjectType::Unknown && egospd > param_conf_.egospdforignoringunknownobject))) {
            res = boolean_t_true;
        } else {
            res = boolean_t_false;
        }
        
        return res;
    }

    basetype::boolean_t Validator::checkisvalidcameramotorcycle(const basetype::boolean_t &isrearobject,
                                                                const ObjectType &type,
                                                                const basetype::float32_t &objlgtdis,
                                                                const ObjectMotionType &motiontype,
                                                                const ObjectTrackStatus &trackstatus,
                                                                const ObjectFusionSource &fusionsource,
                                                                const basetype::boolean_t &jumpok,
                                                                const basetype::boolean_t &maturityok) {
        basetype::boolean_t res{boolean_t_false};
        if ((type == ObjectType::Motorcycle) &&
            (((objlgtdis < param_conf_.maxdistanceforcameraonlymotor) &&
              (objlgtdis > param_conf_.mindistanceforcameraonlymotor) &&
              (isrearobject == boolean_t_false)) ||
             ((objlgtdis < param_conf_.maxreardistanceforcameraonlymotor) &&
              (objlgtdis > param_conf_.minreardistanceforcameraonlymotor) &&
              (isrearobject == boolean_t_true))) &&
            (fusionsource == ObjectFusionSource::CameraOnly) &&
            (motiontype != ObjectMotionType::Oncoming) &&
            (motiontype != ObjectMotionType::Crossing) &&
            (trackstatus != ObjectTrackStatus::Invalid) &&
            (jumpok == boolean_t_true) &&
            (maturityok == boolean_t_true)) {
            res = boolean_t_true;
        } else {
        }
        
        return res;
    }

    basetype::boolean_t Validator::checkisvalidcameracar(const basetype::objectid_t &idn0,
                                                         const basetype::boolean_t &isrearobject,
                                                         const ObjectType &type,
                                                         const basetype::float32_t &objlgtdis,
                                                         const basetype::float32_t &objdisLgtstddev,
                                                         const ObjectMotionType &motiontype,
                                                         const ObjectTrackStatus &trackstatus,
                                                         const ObjectFusionSource &fusionsource,
                                                         const basetype::boolean_t &jumpok,
                                                         const basetype::boolean_t &maturityok,
                                                         const basetype::float32_t &egospd) {
        basetype::boolean_t res{boolean_t_false};
        basetype::boolean_t ispretgta{boolean_t_false};
        if (pretgtaid_ > zero_objectid_t) {
            if (idn0 == (pretgtaid_ - one_objectid_t)) {
                ispretgta = boolean_t_true;
            } else {
            }
        }
        if ((type == ObjectType::Car || type == ObjectType::Truck || type == ObjectType::Motorcycle) &&
            ((((objlgtdis < param_conf_.maxdistanceforcameraonlycar) &&
               (objlgtdis > param_conf_.mindistanceforcameraonlycar) &&
               (isrearobject == boolean_t_false)) ||
              ((objlgtdis < param_conf_.maxreardistanceforcameraonlycar) &&
               (objlgtdis > param_conf_.minreardistanceforcameraonlycar) &&
               (isrearobject == boolean_t_true))) ||
            (egospd < param_conf_.maxegospdforcameraonlycar) ||
            (ispretgta == boolean_t_true)) &&
            (objdisLgtstddev <= param_conf_.lgtdistancedevthreshold) &&
            (fusionsource == ObjectFusionSource::CameraOnly) &&
            (motiontype != ObjectMotionType::Oncoming) &&
            (motiontype != ObjectMotionType::Crossing) &&
            (trackstatus != ObjectTrackStatus::Invalid) &&
            (jumpok == boolean_t_true) &&
            (maturityok ==boolean_t_true)) {
            res = boolean_t_true;
        } else {
        }
        
        return res;
    }

    basetype::boolean_t Validator::checkisvalidpedestrain(const Object &obj, const EgoVehicleInformation &egovehinfo, const basetype::boolean_t &isrearobject, const basetype::boolean_t &validpre) {
        basetype::boolean_t res(boolean_t_false);
        if (isrearobject == boolean_t_false) {
            if (obj.Info.Type == ObjectType::Pedestrain) {
                if (obj.Est.DisLgt < 50.0f && obj.Est.DisLgt < 3.0f * Util::max(egovehinfo.SpdLgt, 2.7f)) {
                    res = boolean_t_true;
                } else {
                }
            } else {
            }
        } else {
        }
        return res;
    }
} // namespace ACC_Target_Selection
