

/**
 * @file
 **/

#include "polynom_line.h"

#include <iostream>

#include "glog/logging.h"
#include "reference_point.h"
#include "vec2d.h"

namespace rm {

// PolynomLine::PolynomLine(
//     const std::array<double, 3>& start, const double end, const double param)
//     {}
PolynomLine::PolynomLine(std::array<double, 4> coef, double start_range,
                         double end_range)
    : coef_(coef), start_range_(start_range), end_range_(end_range) {}

void PolynomLine::DerivedFromQuarticCurve(const PolynomialCurve1d &other) {
  CHECK_EQ(other.Order(), 4U);
  param_ = other.ParamLength();
  for (size_t i = 1; i < 5; ++i) {
    coef_[i - 1] = other.Coef(i) * static_cast<double>(i);
  }
}

double PolynomLine::Evaluate(const std::uint32_t order, const double p) const {
  switch (order) {
    case 0: {
      return ((coef_[3] * p + coef_[2]) * p + coef_[1]) * p + coef_[0];
    }
    case 1: {
      return (3.0 * coef_[3] * p + 2.0 * coef_[2]) * p + coef_[1];
    }
    case 2: {
      return 6.0 * coef_[3] * p + 2.0 * coef_[2];
    }
    case 3: {
      return 6.0 * coef_[3];
    }
    default:
      return 0.0;
  }
}

std::string PolynomLine::ToString() const {
  // return absl::StrCat(absl::StrJoin(coef_, "\t"), param_, "\n");
}

void PolynomLine::ComputeCoefficients(const double x0, const double dx0,
                                      const double ddx0, const double x1,
                                      const double param) {
  DCHECK(param > 0.0);
  const double p2 = param * param;
  const double p3 = param * p2;
  coef_[0] = x0;
  coef_[1] = dx0;
  coef_[2] = 0.5 * ddx0;
  coef_[3] = (x1 - x0 - dx0 * param - coef_[2] * p2) / p3;
}

double PolynomLine::Coef(const size_t order) const {
  CHECK_GT(4U, order);
  return coef_[order];
}
double PolynomLine::GetKappa(double x) {
  const double dy = Evaluate(1, x);
  const double d2y = Evaluate(2, x);
  return d2y / std::pow((1 + dy * dy), 1.5);
}

std::vector<common::math::ReferencePoint> PolynomLine::DiscreteCurved(
    double unit_x) {
  std::vector<common::math::ReferencePoint> ans;
  double s = 0;
  uint16_t count = 0;
  for (double x = std::round(start_range_); x <= end_range_; x += unit_x) {
    count++;
    common::math::ReferencePoint point;
    double y = Evaluate(0, x);
    point.set_x(x);
    point.set_y(y);
    if (ans.size() > 1) {
      auto pre_point = ans.back();
      s += std::hypot(x - pre_point.x(), y - pre_point.y());
    }
    point.set_heading(std::atan2(Evaluate(1, x), 1));
    point.set_kappa(GetKappa(x));
    const double k1 = GetKappa(x - 0.0001);
    const double k2 = GetKappa(x + 0.0001);
    point.set_dkappa((k2 - k1) / 0.0002);
    ans.emplace_back(point);
    // max size of discrete points is 500
    if (count > 500U) return ans;
  }
  return ans;
}

std::vector<common::math::ReferencePoint> PolynomLine::DiscreteCurved(
    double end_range, double unit_x) {
  std::vector<common::math::ReferencePoint> ans;
  double s = 0;
  for (double x = start_range_; x <= end_range; x += unit_x) {
    common::math::ReferencePoint point;
    double y = Evaluate(0, x);
    point.set_x(x);
    point.set_y(y);
    if (ans.size() > 1) {
      auto pre_point = ans.back();
      s += std::hypot(x - pre_point.x(), y - pre_point.y());
    }
    point.set_heading(std::atan2(Evaluate(1, x), 1));
    point.set_kappa(GetKappa(x));
    const double k1 = GetKappa(x - 0.0001);
    const double k2 = GetKappa(x + 0.0001);
    point.set_dkappa((k2 - k1) / 0.0002);
    ans.emplace_back(point);
  }
  return ans;
}

PolynomLine PolynomLine::MoveCurved(double dy, bool is_front_marker) {
  //   double l = end_range_;
  //   double c0 = 2 * coef_[2];
  //   double c2 = Evaluate(2,l);
  //   double dr = -dy;
  //   double t = Evaluate(1,l);

  //   double c0_new = 1.0/(1.0/c0+dr);
  //   double c2_new = 1.0/(1.0/c2+dr);
  //   double t_new = t;
  //   double l_new = l +dr *t;

  //   double delta_c = (2*t_new/l_new -(c0_new+c2_new))/2;
  //   double c0_opt = c0_new +delta_c;
  //   double c2_opt = c2_new +delta_c;
  //   double c1_opt = (c2_opt-c0_opt)/l_new;
  //   std::array<double, 4> coef={coef_[0]+dy,coef_[1],c0_opt/2.0,c1_opt/6.0};

  double far_range = (!is_front_marker) ? start_range_ : end_range_;
  double dr = -dy;
  double t = Evaluate(1, far_range);
  double l_new = far_range + dr * t;
  double c0_new = math::SafeDivide(1.0, (math::SafeDivide(0.5, coef_[2]) + dr));
  double c2_new = math::SafeDivide(
      1.0, (math::SafeDivide(1.0, Evaluate(2, far_range)) + dr));
  double curv = c0_new;
  double curv_rate = math::SafeDivide((c2_new - c0_new), l_new);
  double tau = t - 0.5 * l_new * (c0_new + c2_new);
  std::array<double, 4> coef = {coef_[0] + dy, tau, 0.5 * curv,
                                curv_rate / 6.0};

  double polynom_start_range = (is_front_marker) ? start_range_ : l_new;
  double polynom_end_range = (is_front_marker) ? l_new : end_range_;
  return PolynomLine(coef, polynom_start_range, polynom_end_range);
}

PolynomLine PolynomLine::MoveCurved(double dy) {
  double dr = -dy;
  double t = Evaluate(1, end_range_);
  double l_new = end_range_ + dr * t;
  double c0_new = math::SafeDivide(1.0, (math::SafeDivide(0.5, coef_[2]) + dr));
  double c2_new = math::SafeDivide(
      1.0, (math::SafeDivide(1.0, Evaluate(2, end_range_)) + dr));
  double curv = c0_new;
  double curv_rate = math::SafeDivide((c2_new - c0_new), l_new);
  double tau = t - 0.5 * l_new * (c0_new + c2_new);
  std::array<double, 4> coef = {coef_[0] + dy, tau, 0.5 * curv,
                                curv_rate / 6.0};

  return PolynomLine(coef, start_range_, l_new);
}

double PolynomLine::DistanceToCurve(Vec2d point, double x) {
  return point.DistanceSquareTo(Vec2d(x, Evaluate(0, x)));
}

double PolynomLine::ComputeArc(double x) {
  return sqrt(1 + pow(Evaluate(1, x), 2));
}

// compute the arc length of segement from start_point to end_point on the
// polynom line
double PolynomLine::ComputeArcLengthbySimpson(double start_point,
                                              double end_point) {
  return (end_point - start_point) / 6 *
         (ComputeArc(start_point) +
          4 * ComputeArc(0.5 * (start_point + end_point)) +
          ComputeArc(end_point));
}

double PolynomLine::ComputeArcLengthbySimpson() {
  return ComputeArcLengthbySimpson(start_range_, end_range_);
}

double PolynomLine::CalcMinCartDist2AnotherLine(const PolynomLine &line) {
  double start_range = std::max(start_range_, line.start_range());
  double end_range = std::min(end_range_, line.end_range());
  double diff_c0 = coef_.at(0) - line.Coef(0);
  double diff_c1 = coef_.at(1) - line.Coef(1);
  double diff_c2 = coef_.at(2) - line.Coef(2);
  double diff_c3 = coef_.at(3) - line.Coef(3);
  double dis_start_range = diff_c0 + diff_c1 * start_range +
                           diff_c2 * start_range * start_range +
                           diff_c3 * start_range * start_range * start_range;
  double dis_end_range = diff_c0 + diff_c1 * end_range +
                         diff_c2 * end_range * end_range +
                         diff_c3 * end_range * end_range * end_range;
  if (4 * diff_c2 * diff_c2 - 12 * diff_c1 * diff_c3 >= 0.0) {
    double x_min_1 = (-2 * diff_c2 + std::sqrt(4 * diff_c2 * diff_c2 -
                                               12 * diff_c1 * diff_c3)) /
                     (6 * diff_c3);
    double x_min_2 = (-2 * diff_c2 - std::sqrt(4 * diff_c2 * diff_c2 -
                                               12 * diff_c1 * diff_c3)) /
                     (6 * diff_c3);
    double dis_x_min_1 = (x_min_1 <= end_range && x_min_1 >= start_range)
                             ? diff_c0 + diff_c1 * x_min_1 +
                                   diff_c2 * x_min_1 * x_min_1 +
                                   diff_c3 * x_min_1 * x_min_1 * x_min_1
                             : 1000.0;
    double dis_x_min_2 = (x_min_2 <= end_range && x_min_2 >= start_range)
                             ? diff_c0 + diff_c1 * x_min_2 +
                                   diff_c2 * x_min_2 * x_min_2 +
                                   diff_c3 * x_min_2 * x_min_2 * x_min_2
                             : 1000.0;
    return std::min(std::min(dis_start_range, dis_end_range),
                    std::min(dis_x_min_1, dis_x_min_2));
  } else {
    return std::min(dis_start_range, dis_end_range);
  }
}

}  // namespace rm
