#ifdef GTEST
#include "state_machine/state_machine.h"

#include "gflags.h"
#include "glog/logging.h"
#include "gtest.h"
#include "road_model_gflags.h"
namespace rm {
namespace state_machine {

enum ILCState : int {
  ILC_INIT,
  ILC_PASSIVE,
  ILC_STANDBY,
  ILC_ACTIVEFORLEFT,
  ILC_ACTIVEFORRIGHT
};
enum ILCChangeFlag : int {
  ILC_INIT_TO_PASSIVE,
  ILC_PASSIVE_TO_STANDBY,
  ILC_STANDBY_TO_PASSIVE,
  ILC_STANDBY_TO_ACTIVEFORLEFT,
  ILC_STANDBY_TO_ACTIVEFORRIGHT,
  ILC_ACTIVEFORLEFT_TO_PASSIVE,
  ILC_ACTIVEFORLEFT_TO_STANDBY,
  ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
  ILC_ACTIVEFORRIGHT_TO_PASSIVE,
  ILC_ACTIVEFORRIGHT_TO_STANDBY,
  ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT
};

StateMachine ilc_state_machine;
ILCState current_state = ILCState::ILC_INIT;

void testStateTransition(ILCChangeFlag flag, ILCState state) {
  ilc_state_machine.ChangeState(flag);
  current_state = ilc_state_machine.GetCurrentState();
  EXPECT_EQ(current_state, state);
}

TEST(TestStateMachine, test_init) {
  std::vector<std::array<int, 3>> state_machine_transition{
      {ILCState::ILC_INIT, ILCChangeFlag::ILC_INIT_TO_PASSIVE,
       ILCState::ILC_PASSIVE},
      {ILCState::ILC_PASSIVE, ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
       ILCState::ILC_STANDBY},
      {ILCState::ILC_STANDBY, ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
       ILCState::ILC_PASSIVE},
      {ILCState::ILC_STANDBY, ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
       ILCState::ILC_ACTIVEFORLEFT},
      {ILCState::ILC_STANDBY, ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
       ILCState::ILC_ACTIVEFORRIGHT},
      {ILCState::ILC_ACTIVEFORLEFT, ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
       ILCState::ILC_PASSIVE},
      {ILCState::ILC_ACTIVEFORLEFT, ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
       ILCState::ILC_STANDBY},
      {ILCState::ILC_ACTIVEFORLEFT,
       ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
       ILCState::ILC_ACTIVEFORRIGHT},
      {ILCState::ILC_ACTIVEFORRIGHT,
       ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE, ILCState::ILC_PASSIVE},
      {ILCState::ILC_ACTIVEFORRIGHT,
       ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY, ILCState::ILC_STANDBY},
      {ILCState::ILC_ACTIVEFORRIGHT,
       ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
       ILCState::ILC_ACTIVEFORLEFT}};
  ilc_state_machine.LoadStateMachine(state_machine_transition);

  ilc_state_machine.SetInitializeState(ILCState::ILC_INIT);
  EXPECT_EQ(ilc_state_machine.GetCurrentState(), ILCState::ILC_INIT);

  testStateTransition(ILCChangeFlag::ILC_INIT_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);

  ilc_state_machine.SetInitializeState(ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY,
                      ILCState::ILC_INIT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
                      ILCState::ILC_INIT);
}  // namespace state_machine

TEST(TestStateMachine, test_passive) {
  ilc_state_machine.SetInitializeState(ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
                      ILCState::ILC_STANDBY);

  ilc_state_machine.SetInitializeState(ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_INIT_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY,
                      ILCState::ILC_PASSIVE);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
                      ILCState::ILC_PASSIVE);
}

TEST(TestStateMachine, test_standby) {
  ilc_state_machine.SetInitializeState(ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);

  ilc_state_machine.SetInitializeState(ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
                      ILCState::ILC_ACTIVEFORLEFT);

  ilc_state_machine.SetInitializeState(ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_ACTIVEFORRIGHT);

  ilc_state_machine.SetInitializeState(ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_INIT_TO_PASSIVE,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY,
                      ILCState::ILC_STANDBY);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
                      ILCState::ILC_STANDBY);
}

TEST(TestStateMachine, test_active_left) {
  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);

  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
                      ILCState::ILC_STANDBY);

  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_ACTIVEFORRIGHT);

  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_INIT_TO_PASSIVE,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY,
                      ILCState::ILC_ACTIVEFORLEFT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
                      ILCState::ILC_ACTIVEFORLEFT);
}

TEST(TestStateMachine, test_active_right) {
  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE,
                      ILCState::ILC_PASSIVE);

  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY,
                      ILCState::ILC_STANDBY);

  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
                      ILCState::ILC_ACTIVEFORLEFT);

  ilc_state_machine.SetInitializeState(ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_INIT_TO_PASSIVE,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
                      ILCState::ILC_ACTIVEFORRIGHT);
  testStateTransition(ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
                      ILCState::ILC_ACTIVEFORRIGHT);
}

// int main(int argc, char **argv) {
//   testing::InitGoogleTest(&argc, argv);
//   testing::GTEST_FLAG(filter) = "TestStateMachine.*";  //过滤 只测试这个test

//   google::InitGoogleLogging(argv[0]);  // google的日志库
//   FLAGS_logtostderr = false;  // true: 设置日志消息转到标准输出
//   而不是日志文件 FLAGS_colorlogtostderr = true;  // true: log为彩色
//   //   FLAGS_max_log_size = 100; // true: 设置最大日志文件大小（以MB为单位）
//   const char *dir = FLAGS_selflog_dir.c_str();
//   if (access(dir, 0) == -1)
//     mkdir(dir, 00700);  //判断文件夹是否存在 若不存在则创建
//   FLAGS_log_dir = FLAGS_selflog_dir;  //有此句才可以输出日志文件
//   FLAGS_v = 4;  //设置clog的输出级别 ADEBUG的级别是4
//   FLAGS_v小于4时不打印ADEBUG
//   //
//   设置特定严重级别的日志的输出目录和前缀。第一个参数为日志级别，第二个参数表示输出目录及日志文件名前缀
//   google::SetLogDestination(google::GLOG_INFO,
//                             (FLAGS_selflog_dir + "INFO_").c_str());
//   google::SetLogDestination(google::GLOG_WARNING,
//                             (FLAGS_selflog_dir + "WARNING_").c_str());
//   google::SetLogDestination(google::GLOG_ERROR,
//                             (FLAGS_selflog_dir + "ERROR_").c_str());
//   google::SetLogDestination(google::GLOG_FATAL,
//                             (FLAGS_selflog_dir + "FATAL_").c_str());
//   google::SetLogFilenameExtension(".log");
//   LoadStateMachine();
//   ilc_state_machine.SetInitializeState(ILCState::ILC_PASSIVE);
//   ilc_state_machine.SetChangeFlag(ILCChangeFlag::STANDBY_PASSIVE);
//   return RUN_ALL_TESTS();
//   google::ShutdownGoogleLogging();  //关闭log服务 否则会内存溢出：
// }
}  // namespace state_machine
}  // namespace rm

#endif