#pragma once

#include "debouncer.h"
#include "delay.h"
#include "enum_list.h"
#include "flip_flop.h"
#include "frame.h"
#include "global_parameters.h"
#include "lane_info.h"
#include "lookup_table.h"
#include "object_location.h"
#include "polynom_line.h"
#include "rm_config.h"
#include "vec2d.h"
#include "vehicle_state_provider.h"

namespace rm {

enum struct ATSSWITCHSTATUS {
  NOACTION = 0,
  PASSIVE,
  ACTIVE,
};

struct DebouncerDelayPair {
  rm::Debouncer debouncer;
  rm::Delay<bool> delay;
  DebouncerDelayPair() : debouncer(0.0, 0.0), delay(0.0, 0.0, false){};
  DebouncerDelayPair(float lower_bound, float upper_bound, float rising_time,
                     float falling_time, bool init_flag = false)
      : debouncer(lower_bound, upper_bound),
        delay(rising_time, falling_time, init_flag){};
};

struct ATSSwitchDeciderInputInfoStruct {
  LaneLineCrossedType lane_crossed_type;
  float column_trq;
  float ego_velocity;
  float left_lat_offset_front_bumper;
  float right_lat_offset_front_bumper;
  float left_heading_front_bumper;
  float right_heading_front_bumper;
  float ref_lat_offset_front_bumper;
  float ref_heading_front_bumper;
  float ref_curv_cog;
  float ego_veh_curv;
  bool is_tja_ica_active;
  ObjectLocation ego_corridor_closest_object_location;
  PolynomLine left_lane_marker;
  PolynomLine right_lane_marker;
};

class ATSSwitchDeciderStateMachine;

class ATSSwitchDeciderState {
public:
  // ATSSwitchDeciderState(){};
  ATSSwitchDeciderState(bool ats_switch_flag = false,
                        ATSSWITCHSTATUS state = ATSSWITCHSTATUS::NOACTION)
      : counter_(0U), ats_switch_flag_(ats_switch_flag), state_(state){};
  virtual ~ATSSwitchDeciderState(){};
  virtual void Handle(ATSSwitchDeciderStateMachine *sm) = 0;
  bool ats_switch_flag() { return ats_switch_flag_; }
  uint8_t state() { return static_cast<uint8_t>(state_); }

protected:
  uint16_t counter_;
  bool ats_switch_flag_;
private:
  ATSSWITCHSTATUS state_;
};

class ATSSwitchNoAction : public ATSSwitchDeciderState {
public:
  ATSSwitchNoAction()
      : ATSSwitchDeciderState(false, ATSSWITCHSTATUS::NOACTION){};
  void Handle(ATSSwitchDeciderStateMachine *sm) override;

private:
  DebouncerDelayPair abs_hand_trq_dd_pair_{0.5, 1.0, 0.2, 0.0};
};

class ATSSwitchPassive : public ATSSwitchDeciderState {
public:
  ATSSwitchPassive() : ATSSwitchDeciderState(false, ATSSWITCHSTATUS::PASSIVE){};
  void Handle(ATSSwitchDeciderStateMachine *sm) override;
  bool CheckObjectHeadingCond(ATSSwitchDeciderStateMachine *sm);

private:
  // passive to active
  OneDimLookupTable curv_diff_upper_threshold_ref_curv_table_p2a_{
      {0.0, 0.0005, 0.001, 0.002, 0.005, 0.007, 0.01},
      {0.0018, 0.002, 0.002, 0.002, 0.003, 0.0035, 0.005}};
  OneDimLookupTable curv_diff_lower_threshold_ref_curv_table_p2a_{
      {0.0, 0.0005, 0.001, 0.002, 0.005, 0.007, 0.01},
      {0.001, 0.001, 0.0001, 0.001, 0.0015, 0.003, 0.004}};
  DebouncerDelayPair ref_curv_diff_dd_pair_p2a_{0.0, 0.0, 0.16, 0.0};
  DebouncerDelayPair min_lat_offset_dd_pair_p2a_{1.2, 1.3, 0.2, 0.0};
  DebouncerDelayPair ref_heading_dd_pair_p2a_{0.03, 0.038, 0.2, 0.0};
  common::FlipFlop curv_diff_cond_flipflop_p2a_;
  // passive to noaction
  OneDimLookupTable curv_diff_upper_threshold_ref_curv_table_p2n_{
      {0.0, 0.0005, 0.001, 0.002, 0.005, 0.007, 0.01},
      {0.0018, 0.002, 0.002, 0.002, 0.003, 0.0035, 0.005}};
  OneDimLookupTable curv_diff_lower_threshold_ref_curv_table_p2n_{
      {0.0, 0.0005, 0.001, 0.002, 0.005, 0.007, 0.01},
      {0.001, 0.001, 0.0001, 0.001, 0.0015, 0.003, 0.004}};
  DebouncerDelayPair ref_heading_dd_pair_p2n_{0.05, 0.06, 2.0, 0.0};
  DebouncerDelayPair ref_lat_offset_dd_pair_p2n_{0.35, 0.5, 2.0, 0.0};
  DebouncerDelayPair ref_curv_diff_dd_pair_p2n_{0.0, 0.0, 2.0, 0.0};
  rm::Delay<bool> left_intersection_delay_{0.2, 0.0, false};
  rm::Delay<bool> right_intersection_delay_{0.2, 0.0, false};
};

class ATSSwitchActive : public ATSSwitchDeciderState {
public:
  ATSSwitchActive() : ATSSwitchDeciderState(true, ATSSWITCHSTATUS::ACTIVE){};
  void Handle(ATSSwitchDeciderStateMachine *sm) override;

private:
  DebouncerDelayPair ref_lat_offset_dd_pair_{0.4, 0.5, 0.2, 0.0};
  DebouncerDelayPair ref_heading_dd_pair_{0.05, 0.06, 0.2, 0.0};
  DebouncerDelayPair ref_curv_diff_dd_pair_{0.002, 0.004, 0.2, 0.0};
  OneDimLookupTable curv_diff_upper_threshold_ref_curv_table_{
      {0.0, 0.0005, 0.001, 0.002, 0.005, 0.007, 0.01},
      {0.0018, 0.002, 0.002, 0.002, 0.003, 0.0035, 0.005}};
  OneDimLookupTable curv_diff_lower_threshold_ref_curv_table_{
      {0.0, 0.0005, 0.001, 0.002, 0.005, 0.007, 0.01},
      {0.001, 0.001, 0.0001, 0.001, 0.0015, 0.003, 0.004}};
};

class ATSSwitchDeciderStateMachine {
public:
  ATSSwitchDeciderStateMachine(rm::ATSSwitchDeciderState *state)
      : state_(state){};

  ~ATSSwitchDeciderStateMachine() { delete state_; };

  void ChangeState(rm::ATSSwitchDeciderState *state) {
    delete state_;
    state_ = state;
  };

  void UpdateInputInfo(const LaneInfo *ego_lane_info_ptr,
                       LaneLineCrossedType lane_crossed_type);

  void Request() { state_->Handle(this); };

  const ATSSwitchDeciderInputInfoStruct &input_info() const {
    return input_info_;
  }

  void Reset() {
    delete state_;
    state_ = new ATSSwitchNoAction();
  }

  bool get_ats_switch_flag() { return state_->ats_switch_flag(); }

  uint8_t state() { return state_->state(); }

private:
  rm::ATSSwitchDeciderState *state_;
  ATSSwitchDeciderInputInfoStruct input_info_;
};

class ATSSwitchDecider {
public:
  ATSSwitchDecider()
      : ats_switch_detector_state_machine_(new ATSSwitchNoAction()){};
  bool Run(Frame *frame);
  uint8_t state() { return ats_switch_detector_state_machine_.state(); }

private:
  ATSSwitchDeciderStateMachine ats_switch_detector_state_machine_;
};

} // namespace rm
