#include "ilc_state_machine.h"

#include "rm_config.h"

namespace rm {

Status ILCStateMachine::Init() {
  if (initialized_) {
    return Status::OK();
  }
  // DLOG(INFO) << "ILC state machine initializing!";
  // std::cout << "ILC state machine initializing!" << std::endl;
  ilc_lane_width_lateral_offset_table_ =
      OneDimLookupTable(GetRMConfig().ilc_base_lateral_offset_col,
                        GetRMConfig().ilc_base_lateral_offset_value);
  ilc_lane_width_lateral_offset_edge_table_ =
      OneDimLookupTable(GetRMConfig().ilc_base_lateral_offset_col,
                        GetRMConfig().ilc_road_edge_lateral_offset_value);
  ilc_lateral_offset_compensation_table_ =
      OneDimLookupTable(GetRMConfig().ilc_lateral_offset_compensation_col,
                        GetRMConfig().ilc_lateral_offset_compensation_value);
  ilc_state_machine_transition_ = {
      {ILCState::ILC_INIT, ILCChangeFlag::ILC_INIT_TO_PASSIVE,
       ILCState::ILC_PASSIVE},
      {ILCState::ILC_PASSIVE, ILCChangeFlag::ILC_PASSIVE_TO_STANDBY,
       ILCState::ILC_STANDBY},
      {ILCState::ILC_STANDBY, ILCChangeFlag::ILC_STANDBY_TO_PASSIVE,
       ILCState::ILC_PASSIVE},
      {ILCState::ILC_STANDBY, ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT,
       ILCState::ILC_ACTIVEFORLEFT},
      {ILCState::ILC_STANDBY, ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT,
       ILCState::ILC_ACTIVEFORRIGHT},
      {ILCState::ILC_ACTIVEFORLEFT, ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE,
       ILCState::ILC_PASSIVE},
      {ILCState::ILC_ACTIVEFORLEFT, ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY,
       ILCState::ILC_STANDBY},
      {ILCState::ILC_ACTIVEFORLEFT,
       ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
       ILCState::ILC_ACTIVEFORRIGHT},
      {ILCState::ILC_ACTIVEFORRIGHT,
       ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE, ILCState::ILC_PASSIVE},
      {ILCState::ILC_ACTIVEFORRIGHT,
       ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY, ILCState::ILC_STANDBY},
      {ILCState::ILC_ACTIVEFORRIGHT,
       ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT,
       ILCState::ILC_ACTIVEFORLEFT}};
  ilc_state_machine_->LoadStateMachine(ilc_state_machine_transition_);
  ilc_state_machine_->SetInitializeState(ILCState::ILC_INIT);
  RegisterStateMachineHandleFunctions();

  initialized_ = true;
  return Status::OK();
}

void ILCStateMachine::SetTaskInput(Frame *frame, LaneInfo *current_lane_info) {
  current_lane_info_ptr_ = current_lane_info;
  ego_lane_type_ = current_lane_info->mutable_lane()->lane_type();
  front_corridor_ = current_lane_info->corridor().front();
  obstacles_ptr_ = frame->road_info_provider()->mutable_obstacles();
  road_edges_ptr_ = frame->road_info_provider()->mutable_road_edges();
  ego_velocity_ = VehicleStateProvider::Instance()->lon_velocity();
  is_boundary_valid_ = (BoundaryPairType::BOTH_VALID ==
                        current_lane_info_ptr_->boundary_pair_type());
  curve_radius_ = std::abs(math::SafeDivide(
      1.0, front_corridor_.mutable_center_line()->GetKappa(0.)));
  lateral_offset_ = front_corridor_.mutable_center_line()->Coef(0);
  lane_width_ = current_lane_info->mutable_lane()->lane_width();
  ilc_lateral_offset_base_ =
      ilc_lane_width_lateral_offset_table_.Evaluate(lane_width_);
}

void ILCStateMachine::RunOnce() {
  if (GetRMConfig().ilc_table_calibration_enable) {
    ilc_lane_width_lateral_offset_table_ =
        OneDimLookupTable(GetRMConfig().ilc_base_lateral_offset_col,
                          GetRMConfig().ilc_base_lateral_offset_value);
    ilc_lane_width_lateral_offset_edge_table_ =
        OneDimLookupTable(GetRMConfig().ilc_base_lateral_offset_col,
                          GetRMConfig().ilc_road_edge_lateral_offset_value);
    ilc_lateral_offset_compensation_table_ =
        OneDimLookupTable(GetRMConfig().ilc_lateral_offset_compensation_col,
                          GetRMConfig().ilc_lateral_offset_compensation_value);
  }
  ObstaclesPreprocess();
  PassiveConditionJudge();
  left_trigger_flag_ = IsLeftDangerous();
  right_trigger_flag_ = IsRightDangerous();
  MakeDecision();
  GenerateILCCorridorAndILCTargets();
  // PrintDebugLog();
}

void ILCStateMachine::RegisterStateMachineHandleFunctions() {
  handle_state_function_map_->insert(
      std::pair<ILCState, void (ILCStateMachine::*)()>(
          ILCState::ILC_INIT, &ILCStateMachine::HandleStateInit));
  handle_state_function_map_->insert(
      std::pair<ILCState, void (ILCStateMachine::*)()>(
          ILCState::ILC_PASSIVE, &ILCStateMachine::HandleStatePassive));
  handle_state_function_map_->insert(
      std::pair<ILCState, void (ILCStateMachine::*)()>(
          ILCState::ILC_STANDBY, &ILCStateMachine::HandleStateStandby));
  handle_state_function_map_->insert(
      std::pair<ILCState, void (ILCStateMachine::*)()>(
          ILCState::ILC_ACTIVEFORLEFT,
          &ILCStateMachine::HandleStateActiveForLeft));
  handle_state_function_map_->insert(
      std::pair<ILCState, void (ILCStateMachine::*)()>(
          ILCState::ILC_ACTIVEFORRIGHT,
          &ILCStateMachine::HandleStateActiveForRight));
}

void ILCStateMachine::MakeDecision() {
  is_state_changed_ = false;

  auto it = handle_state_function_map_->find(current_state_);
  if (it != handle_state_function_map_->end()) {
    auto func = it->second;
    (this->*func)();
  } else {
    // missing handle state function
    DLOG(ERROR) << "missing handle state function: " << current_state_;
  }

  if (is_state_changed_) {
    auto prev_state = current_state_;
    if (ilc_state_machine_->ChangeState(current_changeflag_)) {
      current_state_ =
          static_cast<ILCState>(ilc_state_machine_->GetCurrentState());
      // DLOG(INFO) << "ILC state change from: " << prev_state << " -> "
      //            << current_state_ << ", Changeflag: " << current_changeflag_;
    } else {
      DLOG(ERROR) << "change state error!" << PRINT(current_state_);
    }
    ResetDebounce();
  }
}

void ILCStateMachine::ResetDebounce() {
  delay_noilc2standby_.RunDelay(false);
  delay_standby2noilc_.RunDelay(false);
  delay_standby2left_.RunDelay(false);
  delay_standby2right_.RunDelay(false);
  delay_active2noilc_.RunDelay(false);
  delay_active2active_.RunDelay(false);
  delay_active2standby_.RunDelay(false);
}

void ILCStateMachine::HandleStateInit() {
  if (initialized_) {
    current_changeflag_ = ILCChangeFlag::ILC_INIT_TO_PASSIVE;
    is_state_changed_ = true;
  }
}

void ILCStateMachine::HandleStatePassive() {
  if (delay_noilc2standby_.RunDelay(std::abs(lateral_offset_) <
                                        ilc_lateral_offset_base_ &&
                                    !is_passive_satisfied_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_PASSIVE_TO_STANDBY;
  }
}

void ILCStateMachine::HandleStateStandby() {
  if (delay_standby2noilc_.RunDelay(std::abs(lateral_offset_) >
                                        ilc_lateral_offset_base_ + 0.1 ||
                                    is_passive_satisfied_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_STANDBY_TO_PASSIVE;
  } else if (delay_standby2left_.RunDelay(left_trigger_flag_ &&
                                          !right_trigger_flag_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORLEFT;
  } else if (delay_standby2right_.RunDelay(right_trigger_flag_ &&
                                           !left_trigger_flag_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_STANDBY_TO_ACTIVEFORRIGHT;
  }
}

void ILCStateMachine::HandleStateActiveForLeft() {
  float lateral_offset_threshold = std::max(0.4, (lane_width_ - WIDTH) * 0.5);
  if (delay_active2noilc_.RunDelay(std::abs(lateral_offset_) >
                                       lateral_offset_threshold ||
                                   is_passive_satisfied_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_PASSIVE;
  } else if (delay_active2active_.RunDelay(!left_trigger_flag_ &&
                                           right_trigger_flag_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT;
  } else if (delay_active2standby_.RunDelay(
                 (!left_trigger_flag_ && !right_trigger_flag_) ||
                 (left_trigger_flag_ && right_trigger_flag_))) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_ACTIVEFORLEFT_TO_STANDBY;
  }
}

void ILCStateMachine::HandleStateActiveForRight() {
  float lateral_offset_threshold = std::max(0.4, (lane_width_ - WIDTH) * 0.5);
  if (delay_active2noilc_.RunDelay(std::abs(lateral_offset_) >
                                       lateral_offset_threshold ||
                                   is_passive_satisfied_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_PASSIVE;
  } else if (delay_active2active_.RunDelay(left_trigger_flag_ &&
                                           !right_trigger_flag_)) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT;
  } else if (delay_active2standby_.RunDelay(
                 (!left_trigger_flag_ && !right_trigger_flag_) ||
                 (left_trigger_flag_ && right_trigger_flag_))) {
    is_state_changed_ = true;
    current_changeflag_ = ILCChangeFlag::ILC_ACTIVEFORRIGHT_TO_STANDBY;
  }
}

bool ILCStateMachine::PassiveConditionJudge() {
  bool is_ilc_disable_by_lane_type = false;
  if (GetRMConfig().ilc_disable_by_lane_type) {
    is_ilc_disable_by_lane_type = (rm::LaneType::NORMAL != ego_lane_type_) &&
                                  (rm::LaneType::UNKNOWN != ego_lane_type_);
  }

  if (!is_passive_satisfied_) {
    if (!GetRMConfig().ilc_switch || !is_boundary_valid_ ||
        lane_width_ < GetRMConfig().ilc_lane_width_disable ||
        ego_velocity_ < GetRMConfig().ilc_speed_lower_limit_disable / 3.6f ||
        ego_velocity_ > GetRMConfig().ilc_speed_upper_limit_disable / 3.6f ||
        curve_radius_ < GetRMConfig().ilc_curve_radius_disable ||
        is_ilc_disable_by_lane_type) {
      is_passive_satisfied_ = true;
    }
  } else {
    if (GetRMConfig().ilc_switch && is_boundary_valid_ &&
        lane_width_ > GetRMConfig().ilc_lane_width_enable &&
        ego_velocity_ > GetRMConfig().ilc_speed_lower_limit_enable / 3.6f &&
        ego_velocity_ < GetRMConfig().ilc_speed_upper_limit_enable / 3.6f &&
        curve_radius_ > GetRMConfig().ilc_curve_radius_enable &&
        !is_ilc_disable_by_lane_type) {
      is_passive_satisfied_ = false;
    }
  }
  return is_passive_satisfied_;
}

void ILCStateMachine::GenerateILCCorridor(bool left_shift) {
  double side = left_shift ? 1.0 : -1.0;
  uint32_t ilc_target_id =
      ilc_target_list_.empty() ? 0 : ilc_target_list_.front();
  // DLOG(INFO) << PRINT(ilc_target_id);

  double ilc_lateral_offset_compensation = ilc_lateral_offset_base_;
  if ((left_shift && right_edge_danger_delay_) ||
      (!left_shift && left_edge_danger_delay_)) {
    ilc_lateral_offset_compensation =
        ilc_lane_width_lateral_offset_edge_table_.Evaluate(lane_width_);
  }

  if (GetRMConfig().ilc_lateral_offset_compensation_enable) {
    double min_lateral_distance2object = 10.0;
    if (ilc_target_list_.empty()) {
      min_lateral_distance2object =
          left_shift ? distance_refline2rightedge_ : distance_refline2leftedge_;
    } else {
      rm::ObjectLocation &obj_loc =
          left_shift ? right_potential_dang_objects_list_.front()
                     : left_potential_dang_objects_list_.front();
      min_lateral_distance2object =
          std::abs(obj_loc.reference_point2corridor_frenet.l()) -
          obj_loc.size_width * 0.5;
    }
    ilc_lateral_offset_compensation +=
        ilc_lateral_offset_compensation_table_.Evaluate(
            min_lateral_distance2object);
  }
  if ((left_shift && (is_left_object_in_range_ || is_left_edge_in_range_)) ||
      (!left_shift && (is_right_object_in_range_ || is_right_edge_in_range_))) {
    ilc_lateral_offset_compensation -=
        GetRMConfig().ilc_object_lateral_compensation;
  }
  ilc_lateral_offset_compensation = std::max(
      GetRMConfig().ilc_min_lateral_offset,
      std::min(ilc_lateral_offset_compensation,
               std::min(GetRMConfig().ilc_max_lateral_offset,
                        (lane_width_ - WIDTH) * 0.5 -
                            GetRMConfig().ilc_min_gap_to_lane_marker)));

  Corridor ilc_corridor;
  ilc_corridor.set_center_line(front_corridor_.center_line());
  ilc_corridor.set_feature_tag(rm::CorridorFeatureTag::LINEDEVIATE);
  ilc_corridor.set_corridor_id(front_corridor_.corridor_id() * 10000 + 2);
  ilc_corridor.set_ilc_target_id(ilc_target_id);
  ilc_corridor.set_virtualdrivablewidth(front_corridor_.drivable_width());
  ilc_corridor.MoveCenterLineLateral(side * ilc_lateral_offset_compensation);
  current_lane_info_ptr_->mutable_corridor()->push_back(ilc_corridor);
}

void ILCStateMachine::ObstaclesPreprocess() {
  right_potential_dang_objects_list_.clear();
  left_potential_dang_objects_list_.clear();
  for (auto it = front_corridor_.mutable_object_location()->begin();
       it != front_corridor_.mutable_object_location()->end(); it++) {
    auto obs_iter = obstacles_ptr_->find(it->object_id);
    if (obs_iter == obstacles_ptr_->end()) continue;
    obs_iter->second.Run_is_riding_marker_delay(*it, lane_width_);
    obs_iter->second.OvertakingJudge(*it, ego_velocity_);
    if (!it->in_corridor) {
      if (it->reference_point2corridor_frenet.l() <= 5.0 &&
          it->reference_point2corridor_frenet.l() > 0.0) {
        left_potential_dang_objects_list_.push_back(*it);
      } else if (it->reference_point2corridor_frenet.l() >= -5.0 &&
                 it->reference_point2corridor_frenet.l() < 0.0) {
        right_potential_dang_objects_list_.push_back(*it);
      } else {
        // do nothing
      }
    }
  }
}

bool ILCStateMachine::IsLeftDangerous() {
  is_left_object_in_range_ = false;
  left_edge_danger_delay_ = IsLeftRoadEdgeDangerous();
  if (left_edge_danger_delay_) {
    left_potential_dang_objects_list_.clear();
    ilc_target_list_.clear();
    return true;
  }
  return IsObjectDangerous(left_potential_dang_objects_list_,
                           is_left_object_in_range_);
}

bool ILCStateMachine::IsRightDangerous() {
  is_right_object_in_range_ = false;
  right_edge_danger_delay_ = IsRightRoadEdgeDangerous();
  if (right_edge_danger_delay_) {
    right_potential_dang_objects_list_.clear();
    ilc_target_list_.clear();
    return true;
  }
  return IsObjectDangerous(right_potential_dang_objects_list_,
                           is_right_object_in_range_);
}

bool ILCStateMachine::IsLeftRoadEdgeDangerous() {
  distance_refline2leftedge_ = 100.0;

  if (road_edges_ptr_->empty()) {
    // DLOG(INFO) << "No road edge found!";
    return false;
  }
  auto road_edge_iter = road_edges_ptr_->begin();
  for (; road_edge_iter != road_edges_ptr_->end(); road_edge_iter++) {
    if (RoadEdgePostion::LEFT == road_edge_iter->position()) break;
  }
  if (road_edge_iter == road_edges_ptr_->end() ||
      road_edge_iter->road_edge_segment_list().empty()) {
    // DLOG(INFO) << "Invalid road edge!";
    return false;
  }
  RoadEdgeSegment road_edge_seg;
  if (!road_edge_iter->get_current_road_edge_segment(road_edge_seg)) {
    // DLOG(INFO) << "Invalid road edge segment!";
    return false;
  }
  double distance_marker2edge = 1000.0;
  double distance_ego2edge = 1000.0;
  if (road_edge_seg.geometry().valid()) {
    distance_ego2edge = road_edge_seg.geometry().curve().Evaluate(0, 0.0);
  }
  // DLOG(INFO) << PRINT(distance_ego2edge);
  // take the first lane marker segment, since only the first corridor would
  // be shifted
  Boundary *lane_boundary_ptr =
      &(current_lane_info_ptr_->mutable_boundary()->first);
  auto lane_marker_seg =
      lane_boundary_ptr->mutable_boundary_segment_list()->begin();

  if (road_edge_seg.geometry().curve().end_range() >
          lane_marker_seg->geometry.curve().start_range() &&
      road_edge_seg.geometry().curve().start_range() <
          lane_marker_seg->geometry.curve().end_range()) {
    distance_marker2edge = std::abs(
        road_edge_seg.mutable_geometry()
            ->mutable_curve()
            ->CalcMinCartDist2AnotherLine(lane_marker_seg->geometry.curve()));
  }
  // DLOG(INFO) << PRINT(distance_marker2edge);

  distance_refline2leftedge_ =
      current_lane_info_ptr_->mutable_lane()->lane_width() * 0.5 +
      distance_marker2edge;

  if (is_left_edge_in_range_) {
    is_left_edge_in_range_ = delay_left_edge_in_range_.RunDelay(
        distance_marker2edge < GetRMConfig().ilc_road_edge_out_range);
  } else {
    is_left_edge_in_range_ = delay_left_edge_in_range_.RunDelay(
        distance_marker2edge < GetRMConfig().ilc_road_edge_in_range);
  }

  if (std::abs(distance_ego2edge) < 3.0 || distance_marker2edge < 2.0) {
    left_potential_dang_objects_list_.clear();
  }
  // std::cout << "left edge type: " << (int)road_edge_seg.type()
  //           << ", marker2edge: " << distance_marker2edge
  //           << ", refline2edge: " << distance_refline2leftedge_ << std::endl;

  if (!GetRMConfig().ilc_roadedge_switch) {
    // std::cout << "ilc_roadedge_switch: " << GetRMConfig().ilc_roadedge_switch
    //           << std::endl;
    return false;
  }

  if (left_edge_danger_) {
    if (distance_refline2leftedge_ > GetRMConfig().ilc_refline2edge_disable ||
        distance_marker2edge > GetRMConfig().ilc_marker2edge_disable ||
        road_edge_seg.type() != RoadEdgeType::ELEVATED) {
      left_edge_danger_ = false;
    }
  } else {
    if (distance_refline2leftedge_ < GetRMConfig().ilc_refline2edge_enable &&
        distance_marker2edge < GetRMConfig().ilc_marker2edge_enable &&
        road_edge_seg.type() == RoadEdgeType::ELEVATED) {
      left_edge_danger_ = true;
    }
  }
  return delay_left_edge_danger_.RunDelay(left_edge_danger_);
}

bool ILCStateMachine::IsRightRoadEdgeDangerous() {
  distance_refline2rightedge_ = 100.0;

  if (road_edges_ptr_->empty()) {
    // DLOG(INFO) << "No road edge found!";
    return false;
  }
  auto road_edge_iter = road_edges_ptr_->begin();
  for (; road_edge_iter != road_edges_ptr_->end(); road_edge_iter++) {
    if (RoadEdgePostion::RIGHT == road_edge_iter->position()) break;
  }
  if (road_edge_iter == road_edges_ptr_->end() ||
      road_edge_iter->road_edge_segment_list().empty()) {
    // DLOG(INFO) << "Invalid road edge!";
    return false;
  }
  RoadEdgeSegment road_edge_seg;
  if (!road_edge_iter->get_current_road_edge_segment(road_edge_seg)) {
    // DLOG(INFO) << "Invalid road edge segment!";
    return false;
  }
  double distance_marker2edge = 1000.0;
  double distance_ego2edge = 1000.0;
  if (road_edge_seg.geometry().valid()) {
    distance_ego2edge = road_edge_seg.geometry().curve().Evaluate(0, 0.0);
  }
  // DLOG(INFO) << PRINT(distance_ego2edge);
  // take the first lane marker segment, since only the first corridor would
  // be shifted
  Boundary *lane_boundary_ptr =
      &(current_lane_info_ptr_->mutable_boundary()->second);
  auto lane_marker_seg =
      lane_boundary_ptr->mutable_boundary_segment_list()->begin();

  if (road_edge_seg.geometry().curve().end_range() >
          lane_marker_seg->geometry.curve().start_range() &&
      road_edge_seg.geometry().curve().start_range() <
          lane_marker_seg->geometry.curve().end_range()) {
    distance_marker2edge = std::abs(
        road_edge_seg.mutable_geometry()
            ->mutable_curve()
            ->CalcMinCartDist2AnotherLine(lane_marker_seg->geometry.curve()));
  }
  // DLOG(INFO) << PRINT(distance_marker2edge);

  distance_refline2rightedge_ =
      current_lane_info_ptr_->mutable_lane()->lane_width() * 0.5 +
      distance_marker2edge;

  if (is_right_edge_in_range_) {
    is_right_edge_in_range_ = delay_right_edge_in_range_.RunDelay(
        distance_marker2edge < GetRMConfig().ilc_road_edge_out_range);
  } else {
    is_right_edge_in_range_ = delay_right_edge_in_range_.RunDelay(
        distance_marker2edge < GetRMConfig().ilc_road_edge_in_range);
  }

  if (std::abs(distance_ego2edge) < 3.0 || distance_marker2edge < 2.0) {
    right_potential_dang_objects_list_.clear();
  }
  // std::cout << "right edge type: " << (int)road_edge_seg.type()
  //           << ", marker2edge: " << distance_marker2edge
  //           << ", refline2edge: " << distance_refline2rightedge_ <<
  //           std::endl;
  if (!GetRMConfig().ilc_roadedge_switch) {
    // std::cout << "ilc_roadedge_switch: " << GetRMConfig().ilc_roadedge_switch
    //           << std::endl;
    return false;
  }

  if (right_edge_danger_) {
    if (distance_refline2rightedge_ > GetRMConfig().ilc_refline2edge_disable ||
        distance_marker2edge > GetRMConfig().ilc_marker2edge_disable ||
        road_edge_seg.type() != RoadEdgeType::ELEVATED) {
      right_edge_danger_ = false;
    }
  } else {
    if (distance_refline2rightedge_ < GetRMConfig().ilc_refline2edge_enable &&
        distance_marker2edge < GetRMConfig().ilc_marker2edge_enable &&
        road_edge_seg.type() == RoadEdgeType::ELEVATED) {
      right_edge_danger_ = true;
    }
  }
  return delay_right_edge_danger_.RunDelay(right_edge_danger_);
}

bool ILCStateMachine::IsObjectDangerous(std::list<ObjectLocation> &objects_list,
                                        bool &is_obj_in_range) {
  for (auto it = objects_list.begin(); it != objects_list.end();) {
    auto obs_iter = obstacles_ptr_->find(it->object_id);
    if (obs_iter == obstacles_ptr_->end()) {
      it = objects_list.erase(it);
      continue;
    }
    rm::Obstacle *obstacle = &(obs_iter->second);
    bool is_ilc_object =
        obstacle->classification() == rm::ObstacleType::BUS ||
        obstacle->classification() == rm::ObstacleType::TRUCK ||
        (GetRMConfig().ilc_car_ride_marker_switch &&
         obstacle->classification() == rm::ObstacleType::CAR &&
         obstacle->is_riding_on_lane_marker());

    if (is_ilc_object || !is_obj_in_range) {
      double delta_velocity =
          ego_velocity_ - it->reference_point2corridor_frenet.ds();
      double front_to_object_back = it->reference_point2corridor_frenet.s() -
                                    0.5 * obstacle->length() -
                                    rm::DISREARAXEL2FRONT;
      double back_to_object_front = it->reference_point2corridor_frenet.s() +
                                    0.5 * obstacle->length() +
                                    rm::DISREARAXEL2BACK;
      if (!is_obj_in_range) {
        double obj_in_range_overtaking_threshold = std::min(
            GetRMConfig().object_in_range_max_gap,
            std::max(GetRMConfig().object_in_range_min_gap,
                     std::abs(delta_velocity) *
                         GetRMConfig().object_in_range_overtaking_ttc));
        double obj_in_range_overtaked_threshold =
            std::min(GetRMConfig().object_in_range_max_gap,
                     std::max(GetRMConfig().object_in_range_min_gap,
                              std::abs(delta_velocity) *
                                  GetRMConfig().object_in_range_overtaked_ttc));
        if ((!obstacle->is_overtaking() &&
             front_to_object_back < obj_in_range_overtaking_threshold &&
             back_to_object_front >
                 GetRMConfig().object_in_range_min_gap * -1.0) ||
            (obstacle->is_overtaking() &&
             front_to_object_back < GetRMConfig().object_in_range_min_gap &&
             back_to_object_front > obj_in_range_overtaked_threshold * -1.0)) {
          is_obj_in_range = true;
        }
      }
      if (is_ilc_object) {
        double base_distance_front = GetRMConfig().ilc_base_gap_front;
        double base_distance_front_overtaked =
            GetRMConfig().ilc_base_gap_front_overtaked;
        double base_distance_back = GetRMConfig().ilc_base_gap_back;
        if (GetRMConfig().ilc_distance_compensation_enable) {
          double max_velocity =
              std::max(ego_velocity_, it->reference_point2corridor_frenet.ds());
          double distance_compensation =
              std::max(0., (max_velocity * 3.6 - 60) *
                               GetRMConfig().ilc_distance_compensation_factor);
          base_distance_front += distance_compensation;
          base_distance_back += distance_compensation;
          base_distance_front_overtaked += distance_compensation;
        }
        double overtaking_distance_threshold =
            std::max(base_distance_front, std::abs(delta_velocity) *
                                              GetRMConfig().ilc_overtaking_ttc);
        if (GetRMConfig().ilc_overtaking_compensation_enable) {
          overtaking_distance_threshold += std::max(
              0., (ego_velocity_ * 3.6 - 60) *
                      GetRMConfig().ilc_overtaking_compensation_factor);
        }
        overtaking_distance_threshold =
            std::min(GetRMConfig().ilc_max_overtaking_gap,
                     overtaking_distance_threshold);

        double overtaked_distance_threshold = std::min(
            GetRMConfig().ilc_max_overtaked_gap,
            std::max(base_distance_back, std::abs(delta_velocity) *
                                             GetRMConfig().ilc_overtaked_ttc));
        if ((!obstacle->is_overtaking() &&
             front_to_object_back < overtaking_distance_threshold &&
             back_to_object_front > base_distance_back * -1.0) ||
            (obstacle->is_overtaking() &&
             front_to_object_back < base_distance_front_overtaked &&
             back_to_object_front > overtaked_distance_threshold * -1.0)) {
          it++;
          continue;
        }
      }
    }
    it = objects_list.erase(it);
  }
  if (objects_list.empty()) {
    return false;
  } else {
    objects_list.sort();
    return true;
  }
}

void ILCStateMachine::StorePotentialDangerousObjects(
    std::list<ObjectLocation> &objects_list) {
  if (objects_list.empty()) {
    return;
  }
  ilc_target_list_.clear();
  for (const auto &obj : objects_list) {
    ilc_target_list_.push_back(obj.object_id);
  }
}

void ILCStateMachine::GenerateILCCorridorAndILCTargets() {
  if (ILCState::ILC_ACTIVEFORLEFT == current_state_) {
    StorePotentialDangerousObjects(left_potential_dang_objects_list_);
    GenerateILCCorridor(false);
  } else if (ILCState::ILC_ACTIVEFORRIGHT == current_state_) {
    StorePotentialDangerousObjects(right_potential_dang_objects_list_);
    GenerateILCCorridor(true);
  } else {
    ilc_target_list_.clear();
  }
}

void ILCStateMachine::PrintDebugLog() {
  std::string cur_state_str;
  switch (current_state_) {
    case ILCState::ILC_INIT:
      cur_state_str = "ILC_INIT";
      break;
    case ILCState::ILC_PASSIVE:
      cur_state_str = "ILC_PASSIVE";
      break;
    case ILCState::ILC_STANDBY:
      cur_state_str = "ILC_STANDBY";
      break;
    case ILCState::ILC_ACTIVEFORLEFT:
      cur_state_str = "ILC_ACTIVEFORLEFT";
      break;
    case ILCState::ILC_ACTIVEFORRIGHT:
      cur_state_str = "ILC_ACTIVEFORRIGHT";
      break;
    default:
      break;
  }
  std::stringstream str;
  str << "ILC monitor: " << std::endl
      << "ILC current state: " << cur_state_str
      << ", Changeflag: " << current_changeflag_ << std::endl
      << "ego velocity: "
      << "MPS: " << ego_velocity_
      << ", KPH: " << static_cast<int>(ego_velocity_ * 3.6) << std::endl
      << "left trigger flag: " << left_trigger_flag_
      << ", right trigger flag: " << right_trigger_flag_ << std::endl
      << "passive satisfied: " << is_passive_satisfied_
      << ", boundary valid: " << is_boundary_valid_ << std::endl
      << "lateral offset: " << lateral_offset_ << std::endl
      << "lane width: " << lane_width_ << ", curve radius: " << curve_radius_
      << std::endl;
  if (!ilc_target_list_.empty()) {
    str << "ILC target id: " << ilc_target_list_.front() << std::endl;
  } else if (current_state_ == ILCState::ILC_ACTIVEFORLEFT ||
             current_state_ == ILCState::ILC_ACTIVEFORRIGHT) {
    str << "ILC target id: road edge" << std::endl;
  } else {
    str << "ILC target id: None" << std::endl;
  }

  ilc_monitor_string_ = str.str();

  if (is_state_changed_) {
    // DLOG(INFO) << ilc_monitor_string_;
  }
}

void ILCStateMachine::TaskOutput(Frame *frame) {
  frame->mutable_ilc_info()->set_ilc_target_id_list(ilc_target_list_);
  frame->mutable_ilc_info()->set_ilc_monitor_str(ilc_monitor_string_);
}

}  // namespace rm