#pragma once
#include <algorithm>
#include <iostream>
#include <list>

#include "corridor.h"
#include "delay.h"
#include "frame.h"
#include "ilc_info.h"
#include "lane_info.h"
#include "lookup_table.h"
#include "state_machine/state_machine.h"
#include "status.h"
#include "vehicle_state_provider.h"

namespace rm {

enum ILCState : int {
  ILC_INIT,
  ILC_PASSIVE,
  ILC_STANDBY,
  ILC_ACTIVEFORLEFT,
  ILC_ACTIVEFORRIGHT
};

enum ILCChangeFlag : int {
  ILC_INIT_TO_PASSIVE,
  ILC_PASSIVE_TO_STANDBY,
  ILC_STANDBY_TO_PASSIVE,
  ILC_STANDBY_TO_ACTIVEFORLEFT,
  ILC_STANDBY_TO_ACTIVEFORRIGHT,
  ILC_ACTIVEFORLEFT_TO_PASSIVE,
  ILC_ACTIVEFORLEFT_TO_STANDBY,
  ILC_ACTIVEFORLEFT_TO_ACTIVEFORRIGHT,
  ILC_ACTIVEFORRIGHT_TO_PASSIVE,
  ILC_ACTIVEFORRIGHT_TO_STANDBY,
  ILC_ACTIVEFORRIGHT_TO_ACTIVEFORLEFT
};

class ILCStateMachine {
 public:
  using Status = rm::common::Status;
  using StateMachineMap = std::map<ILCState, void (ILCStateMachine::*)()>;
  using StateMachineMapShrPtr = std::shared_ptr<StateMachineMap>;

  ILCStateMachine() { Init(); }
  ~ILCStateMachine(){};

  void SetTaskInput(Frame *frame, LaneInfo *current_lane_info);
  void RunOnce();
  void TaskOutput(Frame *frame);

 private:
  Status Init();
  void MakeDecision();
  void ObstaclesPreprocess();
  void RegisterStateMachineHandleFunctions();
  void ResetDebounce();

  void HandleStateInit();
  void HandleStatePassive();
  void HandleStateStandby();
  void HandleStateActiveForLeft();
  void HandleStateActiveForRight();

  void GenerateILCCorridorAndILCTargets();
  void GenerateILCCorridor(bool left_shift);
  void StorePotentialDangerousObjects(std::list<ObjectLocation> &objects_list);

  bool IsLeftDangerous();
  bool IsRightDangerous();
  bool IsLeftRoadEdgeDangerous();
  bool IsRightRoadEdgeDangerous();
  bool IsObjectDangerous(std::list<ObjectLocation> &objects_list,
                         bool &is_obj_in_range);

  bool PassiveConditionJudge();
  void PrintDebugLog();

 private:
  state_machine::StateMachineShrPtr ilc_state_machine_ =
      std::make_shared<state_machine::StateMachine>();
  // map of handle state function
  StateMachineMapShrPtr handle_state_function_map_ =
      std::make_shared<StateMachineMap>();

  rm::LaneInfo *current_lane_info_ptr_;
  rm::Corridor front_corridor_;
  std::map<uint32_t, Obstacle> *obstacles_ptr_;
  std::vector<RoadEdge> *road_edges_ptr_;
  std::list<ObjectLocation> left_potential_dang_objects_list_;
  std::list<ObjectLocation> right_potential_dang_objects_list_;
  std::vector<uint32_t> ilc_target_list_;
  rm::Delay<bool> delay_noilc2standby_ = rm::Delay<bool>(0.2, 0.0, false);
  rm::Delay<bool> delay_standby2noilc_ = rm::Delay<bool>(0.4, 0.0, false);
  rm::Delay<bool> delay_standby2left_ = rm::Delay<bool>(0.4, 0, false);
  rm::Delay<bool> delay_standby2right_ = rm::Delay<bool>(0.4, 0, false);
  rm::Delay<bool> delay_active2noilc_ = rm::Delay<bool>(0.5, 0.0, false);
  rm::Delay<bool> delay_active2standby_ = rm::Delay<bool>(1.0, 0.0, false);
  rm::Delay<bool> delay_active2active_ = rm::Delay<bool>(0.2, 0.0, false);
  OneDimLookupTable ilc_lane_width_lateral_offset_table_;
  OneDimLookupTable ilc_lane_width_lateral_offset_edge_table_;
  OneDimLookupTable ilc_lateral_offset_compensation_table_;
  double ilc_lateral_offset_base_;

  bool initialized_{false};
  ILCState current_state_{ILCState::ILC_INIT};
  ILCChangeFlag current_changeflag_{ILCChangeFlag::ILC_INIT_TO_PASSIVE};
  std::vector<std::array<int, 3>> ilc_state_machine_transition_;
  bool is_state_changed_{false};
  float ego_velocity_{0.}, lateral_offset_{0.}, lane_width_{0.},
      curve_radius_{0.};
  bool is_boundary_valid_{false}, is_passive_satisfied_{false};
  bool left_trigger_flag_{false}, right_trigger_flag_{false};
  bool left_edge_danger_{false}, right_edge_danger_{false},
      left_edge_danger_delay_{false}, right_edge_danger_delay_{false};
  bool is_left_edge_in_range_{false}, is_right_edge_in_range_{false};
  bool is_left_object_in_range_{false}, is_right_object_in_range_{false};
  double distance_refline2leftedge_, distance_refline2rightedge_;
  rm::Delay<bool> delay_left_edge_danger_ = rm::Delay<bool>(3.0, 1.0, false);
  rm::Delay<bool> delay_right_edge_danger_ = rm::Delay<bool>(3.0, 1.0, false);
  rm::Delay<bool> delay_left_edge_in_range_ = rm::Delay<bool>(2.0, 2.0, false);
  rm::Delay<bool> delay_right_edge_in_range_ = rm::Delay<bool>(2.0, 2.0, false);
  std::string ilc_monitor_string_;
  rm::LaneType ego_lane_type_;
};

}  // namespace rm
