/*
 * @Date: 2021-11-09 13:05:16
 * @LastEditors: Please set LastEditors
 * @LastEditTime: 2022-01-19 13:34:29
 * @FilePath: /Cartesian_to_Frenet_corridor_v3/src/frenet_converter.cpp
 */

#include "frenet_converter.h"
#include "object_location.h"
#include "glog/logging.h"
namespace rm{
    using common::math::CartesianFrenetConverter;
SLPoint FrenetConverter::CalculateFrenetPoint(const Obstacle &object)
{
    std::array<double, 3> s_conditions;
    std::array<double, 3> d_conditions;

    // 1. Find coresponding reference point for cartesian_point on the ref line with all info on frenet
    ReferencePoint ref_point = FindRefPointS(object.center().x(), object.center().y());

    // 2. call cartesian_to_frenet with info in ref_point as inputs
    CartesianFrenetConverter::cartesian_to_frenet(
        ref_point.s(), ref_point.x(), ref_point.y(), ref_point.heading(),
        ref_point.kappa(), ref_point.dkappa(),
        object.center().x(), 
        object.center().y(), 
        object.speed().Length(), 
        object.acceleration().Length(),
        object.heading(), 
        object.kappa(),
        &s_conditions, &d_conditions);

        SLPoint sl_point;
        sl_point.set_s(s_conditions[0]);
        sl_point.set_ds(s_conditions[1]);
        sl_point.set_dds(s_conditions[2]);

        sl_point.set_l(d_conditions[0]);
        sl_point.set_dl(d_conditions[1]);
        sl_point.set_ddl(d_conditions[2]);
        // DLOG(INFO) << PRINT(object);
        // DLOG(INFO) << PRINT(sl_point);

    return sl_point;

    // DebugP(s_conditions[0]);
    // DebugP(s_conditions[1]);

    // DebugP(d_conditions[0]);
    // DebugP(d_conditions[1]);

    /*
    cartesian_to_frenet() 的输入参数：
    1. reference point的相关数据： (即通过FindRefPointS找到的Object映射到referenceline上的最近点)
        double rs = 10.0; // r点的s位置
        double rx = 0.0;  // r点的在cartesian坐标下的x
        double ry = 0.0;
        double rtheta = M_PI / 4.0; // r点的heading (rad)
        double rkappa = 0.1;     // r点的曲率 d(theta)/ds
        double rdkappa = 0.01;   // r点的曲率导数

    2. Object相关数据
        double x = -1.0; // Object在cartesian坐标下的x
        double y = 1.0;
        double v = 2.0;  // 绝对速度(标量)？
        double a = 0.0;  // 绝对加速度(标量)？
        double theta = M_PI / 3.0; // heading (rad)
        double kappa = 0.11; //   (注意这里，原始的方程用于计算Trajectory的转换，原始方程中
        这一变量为Trajectory的曲率，即 d(theta)/ds， 这与描述Object常用yawrate（即d(theta)/dt）不同)

    */
}

ReferencePoint FrenetConverter::FindRefPointS(double x0, double y0)
{

    // Newton Method
    // The function to be minimized is "distance_to_curve"

    // To calculate first and second derivative of the object function, we use finite difference method
    // which is a general idea when the object funcion can not be derived explicitly (as shown below)

    // As an alternative way to this case, explicit derivative could be calculated.
    /*
y
|             *
|            *
| *        *|
|  *     *__|
|    *** rx rx+0.1     dy/dt(@rx) = delta_y / delta_x
|          
|_____________________________ x

*/

        // Init guess
        double init_x = x0;
        double rx = init_x;
        double rx_update = init_x + DERI_STEP;
        Vec2d point = {x0, y0};
        double distance_last_step = ref_line_.DistanceToCurve(point, rx);
        double distance = ref_line_.DistanceToCurve(point, rx_update);

    int counter = 0;

        while (fabs(distance - distance_last_step) > EPSILON)
        {
            counter++;

            if (counter > 100)
                break;

            // #ifndef RELEASE
            // DebugP(counter);
            // #endif

        rx = rx_update;
        // Calculate the dy/dx and d2y/dx2
        //  dy/dx near point rx: with a small step 0.1 for finite difference method
        double rx_right_step = rx + DERI_STEP;
        double diff_rx = (ref_line_.DistanceToCurve(point, rx) - ref_line_.DistanceToCurve(point, rx_right_step)) / (-DERI_STEP);

        //  d2y/dx2 near point rx: with a small step 0.1 for finite difference method
        double diff_rx_right_step = (ref_line_.DistanceToCurve(point, rx_right_step) - ref_line_.DistanceToCurve(point, rx_right_step + DERI_STEP)) / (-DERI_STEP);
        double diff2_rx = (diff_rx - diff_rx_right_step) / (-DERI_STEP);

        // update the rx_update:
        rx_update = rx - diff_rx / diff2_rx;

        // Update the new distance
        distance_last_step = distance;
        distance = ref_line_.DistanceToCurve(point, rx_update);

        // DebugP(distance_last_step);
        // DebugP(distance);
    }

    // Store results to ref point
    rx_ = rx_update;
    ry_ = ref_line_.Evaluate(0,rx_update);

    // calculate rs through curve integration.
    rs_ = CurveIntegration(rx_);

    ReferencePoint ref_point;
    ref_point.set_s(rs_);
    // ref_point.set_kappa(0);
    ref_point.set_kappa( ref_line_.Evaluate(2, rx_) );
    ref_point.set_dkappa( 6.0 * ref_line_.Coef(3));
    ref_point.set_heading( ref_line_.Evaluate(1, rx_) );
    ref_point.set_x(rx_);
    ref_point.set_y(ry_);

    // debug
    // DebugPlot(x0, y0, rx_update);

    return ref_point;
}

// The polynomial curve
// double FrenetConverter::y_fun(double x) { return Coe[0] + Coe[1] * x + Coe[2] * x * x + Coe[3] * x * x * x; }

// double FrenetConverter::distance_to_curve(double x0, double y0, double x)
// {
//     double c0 = Coe[0];
//     double c1 = Coe[1];
//     double c2 = Coe[2];
//     double c3 = Coe[3];

//     double y = c0 + c1 * x + c2 * x * x + c3 * x * x * x;
//     return (x0 - x) * (x0 - x) + (y0 - y) * (y0 - y);
// }

// The fomula for curve integration derived from polynomial
// double FrenetConverter::ds_fun(double x) { return sqrt(1 + pow((Coe[1] + 2 * Coe[2] * x + 3 * Coe[3] * x * x), 2)); }

double FrenetConverter::IntegrateBySimpson(const double a, const double b)
{
    // the algorithm lead to error less than 0.02 m at a the 30m position on a 3rd order curve
    return (b - a) / 6 * (ref_line_.ComputeArc(a) + 4 * ref_line_.ComputeArc(0.5 * (a + b)) + ref_line_.ComputeArc(b));
}

double FrenetConverter::IntegrateBySimpson(const double a, const double b, const double m)
{
    // m: the numbe of segements
    double h = (b - a) / m *0.5; // half of the segement width
    double temp = h * (ref_line_.ComputeArc(a) + ref_line_.ComputeArc(b)) / 3;
    double x_mid;
    double x_k;
    int k;
    for(k = 1; k < m; k++){
        x_k = a + k * h * 2;
        temp = temp + ref_line_.ComputeArc(x_k) * 2 * h / 3;
    }

    for(k = 1; k < m + 1; k++){
        x_mid = a + h + (k - 1) * h * 2;
        temp = temp + 4 * h / 3 * ref_line_.ComputeArc(x_mid);
    }
    return temp;
}

double FrenetConverter::IntegrateByFiniteVolume(const double a, const double b)
{
    double s_accumulate = 0;
    double dx = 0.0001;
    double x = a;
    double y = ref_line_.Evaluate(0,a);
    double y_last = y;
    double dy = 0.0;

    while (x < b)
    {
        y_last = y;
        x += dx;
        y = ref_line_.Evaluate(0,x);
        dy = y - y_last;

        s_accumulate += sqrt(dx * dx + dy * dy);
    }
    return s_accumulate;
}

double FrenetConverter::CurveIntegration(double rx)
{
    // Apply basic Simpson rule
    double curve_integral;
    if (is_front_ref_line_) {
        curve_integral = IntegrateBySimpson(ref_line_.start_range(), rx) + start_offset_;
    }
    else {
        curve_integral = IntegrateBySimpson(ref_line_.end_range(), std::min(rx,ref_line_.end_range())) + start_offset_;
    }
    

    // ********alternative method and debug ************
    // double curve_integral_finitevolune = IntegrateByFiniteVolume(0.0, rx);
    // ********debug************

    return curve_integral;
}

FrenetConverter::FrenetConverter(PolynomLine ref_line)
{
    ref_line_ = ref_line;
    // Coe[0] = 3;
    // Coe[1] = 0.01;
    // Coe[2] = 0.00001;
    // Coe[3] = 0.0002;
}

FrenetConverter::FrenetConverter(const Corridor * const corridor)
{
    ref_line_ = corridor->center_line();
    start_offset_ = corridor->start_offset();
    is_front_ref_line_ = corridor->is_front_corridor();
}


}
