#ifndef SLAM_COMMON_DATA_MANAGER_H
#define SLAM_COMMON_DATA_MANAGER_H

#include <cstdint>
#include <cstdio>
#include <memory>
#include <queue>
#include <Eigen/Dense>
#include <mutex>
#include <string>
// #include <glog/logging.h>

const int GLOG_INFO = 0, GLOG_WARNING = 1, GLOG_ERROR = 2, GLOG_FATAL = 3,
  NUM_SEVERITIES = 4;

const int INFO = GLOG_INFO, WARNING = GLOG_WARNING,
  ERROR = GLOG_ERROR, FATAL = GLOG_FATAL;

#define ILOG(l, s)                \
  {                               \
    printf("%d: %s\n", l, s);     \
  }

namespace slam
{
template <typename T>
class CommonDataManager
{
public:
    void Init(const std::string _name, unsigned int maxDataSize = 100);
    bool GetData(T* result) noexcept(false);
    bool InputData(T* data) noexcept(false);

private:
    std::unique_ptr<std::queue<T>> comm_data_queue_;
    std::mutex dataMutex;
    unsigned int kMaxDataSize;
    std::string msg_name = "default";
};

template <typename T>
void CommonDataManager<T>::Init(const std::string _name, unsigned int maxDataSize)
{
    kMaxDataSize = maxDataSize;
    msg_name = _name;

    try
    {
        comm_data_queue_ = std::make_unique<std::queue<T>>();
    }
    catch (const std::bad_alloc &e)
    {
        //ILOG(ERROR, "Alloc " + msg_name + " comm data queue failed");
    }
}

template <typename T>
bool CommonDataManager<T>::GetData(T* result) noexcept(false)
{
    std::lock_guard<std::mutex> lock(dataMutex);
    if (comm_data_queue_ == nullptr)
    {
        //ILOG(ERROR, "Invalid " + msg_name + " data queue");
        return false;
    }

    if (comm_data_queue_->empty())
    {
        return false;
    }

    *result = comm_data_queue_->front();
    comm_data_queue_->pop();

    return true;
}

template <typename T>
bool CommonDataManager<T>::InputData(T* data) noexcept(false)
{
    std::lock_guard<std::mutex> lock(dataMutex);

    if (comm_data_queue_ == nullptr)
    {
        //ILOG(ERROR, "Invalid " + msg_name + " data queue");
        return false;
    }

    while (comm_data_queue_->size() > kMaxDataSize)
    {
        //ILOG(WARNING, "Data " + msg_name + " inputMsg full buffer!");
        comm_data_queue_->pop();
    }

    comm_data_queue_->push(*data);
    return true;
}
}
#endif // SLAM_COMMON_DATA_MANAGER_H
