#ifndef SLAM_INTERFACE_RTK_DATA_H_
#define SLAM_INTERFACE_RTK_DATA_H_
#pragma pack(push)
#pragma pack(1)

#include <cstdint>
#include "interface/common_type.h"

namespace slam {

// Position status
enum class PositionStatus : std::uint8_t {
  INVALID = 0,          // Invalid
  DR = 1,               // DR
  FIXED_2D = 2,         // 2D fixed
  FIXED_3D = 3,         // 3D fixed
  GNSS_DR = 4,          // GNSS + DR
  ONLY_TIME_VALID = 5   // Only time valid
};

// Position quality
enum class PositionQuality : std::uint8_t {
  INVALID = 0,          // Invalid
  SINGLE = 1,           // Single
  DGPS = 2,             // DGPS 
  RTK_FIXED = 4,        // Rtk fixed
  RTK_FLOAT = 5,        // Rtk float
  DEAD_RECKONING = 6    // Dead reckoning
};

// 
enum class GearLevelPosition : std::uint8_t {
  UNKNOWN = 0,
  POSITION_P = 1,
  POSITION_R = 2,
  POSITION_N = 3,
  POSITION_D = 4,
  POSITION_M = 5,
  POSITION_PLUS = 6,
  POSITION_MINUS = 7,
  POSITION_S = 8
};

enum class GnssInitializedType : std::uint8_t {
  ALIGNMENT = 0,          // Alignment
  ALIGNMENT_COMPLETE = 1  // Alignment complete
};

// UTC time
struct alignas(8) UtcTime {
  /// @brief Year
  std::uint8_t year;
  /// @brief Month
  std::uint8_t month;
  /// @brief Day
  std::uint8_t day;
  /// @brief Hour
  std::uint8_t hour;
  /// @brief Minute
  std::uint8_t minute;
  /// @brief Second
  std::uint8_t second;
  /// @brief Millisecond
  std::uint16_t milli_second;
  /// @brief Time accuracy, unit: ns
  std::uint16_t time_accuracy;
  /// @brief padding
  std::uint8_t padding[6];
};

// GNSS data
struct alignas(8) GnssData {
  /// @brief GNSS time
  std::uint64_t utc_timestamp;
  /// @brief GnssLat (deg)
  double latitude;
  /// @brief GnssLon (deg)
  double longitude;
  /// @brief GnssEllipsoid
  float ellipsoid;
  /// @brief GnssQuality
  PositionQuality quality;
  /// @brief GnssPosStatus
  PositionStatus position_status;
  /// @brief SatsNum
  std::uint8_t num_satellites;
  /// @brief Initialized
  GnssInitializedType initialized;
  /// @brief GnssAlt (m)
  float altitude;
  /// @brief Velocity (GnssVelNorth, GnssVelEast, GnssVelDown) (m/s)
  Vector3f velocity;
  /// @brief GnssVelHori (m/s)
  float horizon_speed;
  /// @brief GnssHeadingErr (deg)
  float heading_error;
  /// @brief GnssHeading (deg)
  float heading;
  /// @brief GnssCogErr (deg)
  float course_over_ground_err;
  /// @brief GnssCog (deg)
  float course_over_ground;
  /// @brief GnssVertVelErr (m/s)
  float vertical_velocity_error;
  /// @brief GnssVertPosErr (m)
  float vertical_position_error;
  /// @brief GnssHoriVelErr (m/s)
  float horizon_velocity_error;
  /// @brief GnssHoriPosErr (m)
  float horizon_position_error;
  /// @brief GnssGdop 
  float gdop;
  /// @brief GnssVdop
  float vdop;
  /// @brief GnssHdop
  float hdop;
  /// @brief GnssPdop
  float pdop;
  /// @brief GnssTdop
  float tdop;
  /// @brief GnssHdopEast
  float hdop_east;
  /// @brief GnssHdopNorth
  float hdop_north;
};

// IMU data
struct alignas(8) ImuData {
  /// @brief GNSS time
  std::uint64_t utc_timestamp;
  /// @brief IMU raw acceleration (m/s/s)
  Vector3f raw_acceleration;
  /// @brief IMU raw angular rate (deg/s)
  Vector3f raw_angular_rate;
};

// INS data
struct alignas(8) InsData {
  /// @brief GNSS time
  std::uint64_t utc_timestamp;
  /// @brief Latitude (deg)
  double latitude;
  /// @brief Longitude (deg)
  double longitude;
  /// @brief Ellipsoid
  float ellipsoid;
  /// @brief Position status
  PositionStatus position_status;
  /// @brief Position quality
  PositionQuality position_quality;
  std::uint8_t padding_1[2];
  /// @brief Altitude (m)
  float altitude;
  /// @brief Velocity (north, east, down) (m/s)
  Vector3f velocity;
  /// @brief GroundVel2D (m/s)
  float ground_vel_2d;
  /// @brief Acceleration (m/s/s)
  Vector3f acceleration;
  /// @brief INS_AngRateVehicle (deg/s)
  Vector3f angular_rate;
  /// @brief INS_Attitude (heading, pitch, roll) (deg)
  Vector3f attitude;
  /// @brief VertVelErr (m/s)
  float vertical_velocity_error;
  /// @brief VertPosErr (m)
  float vertical_position_error;
  /// @brief HoriVelErr (m/s)
  float horizon_velocity_error;
  /// @brief HoriPosErr (m)
  float horizon_position_error;
  /// @brief CogErr (deg)
  float course_over_ground_err;
  /// @brief Cog (deg)
  float course_over_ground;
  /// @brief Heading (deg) 
  float heading;
  /// @brief Heading error (deg)
  float heading_err;
};

// // Pbox vehicle data
// struct alignas(8) PboxVehicleData {
//   /// @brief RearLftWheelSpeed (km/h)
//   float wheel_speed_rl;
//   /// @brief RearRgtWheelSpeed
//   float wheel_speed_rr;
//   /// @brief RearLftWheelSpeedValid
//   bool is_wheel_speed_rl_valid;
//   /// @brief RearRgtWheelSpeedValid
//   bool is_wheel_speed_rr_valid; 
//   /// @brief RearLftWheelSpeedDir
//   WheelSpeedDirection wheel_speed_rl_direction;
//   /// @brief RearRgtWheelSpeedDir
//   WheelSpeedDirection wheel_speed_rr_direction;
//   /// @brief GearLvrPos
//   GearLevelPosition gear_level_position;
//   std::uint8_t padding[3];
// };

// Pbox data
struct alignas(8) RtkData {
  /// @brief GNSS data
  GnssData gnss_data;
  /// @brief IMU and INS data
  InsData ins_data;
};

} 

#pragma pack(pop)
#endif // SLAM_INTERFACE_RTK_DATA_H_
