
#include "util/slam_util.h"

namespace slam{

Quaterniond SlamUtil::EulerToQuaternion(double roll, double pitch, double yaw) {
    Quaterniond q;

    // 计算半角和它们的正余弦值
    double cy = cos(yaw * 0.5);
    double sy = sin(yaw * 0.5);
    double cp = cos(pitch * 0.5);
    double sp = sin(pitch * 0.5);
    double cr = cos(roll * 0.5);
    double sr = sin(roll * 0.5);

    // 计算四元数
    q.w = cr * cp * cy + sr * sp * sy;
    q.x = sr * cp * cy - cr * sp * sy;
    q.y = cr * sp * cy + sr * cp * sy;
    q.z = cr * cp * sy - sr * sp * cy;

    return q;
}

// 计算两个 Vector2d 之间距离的平方
double squaredDistanceTo(const slam::Vector2d& a, const slam::Vector2d& b) {
    return (a.x - b.x) * (a.x - b.x) + (a.y - b.y) * (a.y - b.y);
}

slam::Vector2d SlamUtil::globalToLocal(slam::Vector2d glb_p,slam::Vector2d center_p, double heading) {

    slam::Vector2d loc_p;
    // 平移
    double dx = glb_p.x - center_p.x;
    double dy = glb_p.y - center_p.y;

    // 旋转
    loc_p.x = dx * cos(heading) + dy * sin(heading);
    loc_p.y = -dx * sin(heading) + dy * cos(heading);

    return loc_p;
}

int SlamUtil::findNearestPointIndex(const slam::Vector2d& cur_host_pos,
                          const std::vector<slam::Vector2d>& ref_line,
                          double max_distance, double min_distance) {
    if (ref_line.empty()) return -1;

    // 使用平方距离来避免在比较时计算平方根
    double max_distance_sq = max_distance * max_distance;
    double min_distance_sq = min_distance * min_distance;

    int left = 0;
    int right = ref_line.size() - 1;
    int nearestIndex = -1;
    double nearestDistanceSq = std::numeric_limits<double>::max();

    while (left <= right) {
        int mid = left + (right - left) / 2;
        double midDistanceSq = squaredDistanceTo(ref_line[mid], cur_host_pos);

        // 更新最近点和距离
        if (midDistanceSq < nearestDistanceSq) {
            nearestDistanceSq = midDistanceSq;
            nearestIndex = mid;
        }

        // 检查是否已找到足够近的点
        if (nearestDistanceSq <= min_distance_sq) {
            return nearestIndex; // 找到足够近的点
        }

        // 比较首尾和中间点，决定下一步搜索方向
        double leftDistanceSq = squaredDistanceTo(ref_line[left], cur_host_pos);
        double rightDistanceSq = squaredDistanceTo(ref_line[right], cur_host_pos);

        if (leftDistanceSq < midDistanceSq) {
            right = mid - 1;
        } else if (rightDistanceSq < midDistanceSq) {
            left = mid + 1;
        } else {
            // 当左、中、右距离相似时，缩小搜索范围
            left++;
            right--;
        }
    }

    return nearestDistanceSq > max_distance_sq ? -1 : nearestIndex;
}


}
