#ifndef COORDINATE_UTIL_H_
#define COORDINATE_UTIL_H_

#include <cstdint>
#include "cmath"
#include "Eigen/Core"
#include "Eigen/Dense"

namespace Localization
{
namespace PointLoc
{

class CoordinateUtil {
public:
  // Latitude & longitude position
  struct PositionLL {
    /// @brief latitude
    double latitude;
    /// @brief longitude
    double longitude;
  };

  struct PositionGK {
    /// @brief gk x
    double x;
    /// @brief gk y
    double y;
    /// @brief longitude base line
    double base_lon;
  };

  // UTM position
  struct PositionUTM {
    /// @brief north
    double north;
    /// @brief east
    double east;
    /// @brief longitude zone number
    std::int8_t longitude_zone_number;
    /// @brief latitude zone
    char latitude_zone;
  };

  CoordinateUtil() = delete;
  ~CoordinateUtil() = delete;
  CoordinateUtil(CoordinateUtil const&) = delete;
  CoordinateUtil& operator=(CoordinateUtil const&) = delete;
  CoordinateUtil(CoordinateUtil&&) = delete;
  CoordinateUtil& operator=(CoordinateUtil&&) = delete;

  /// @brief trans euler ypr and rotation matrix
  static Eigen::Vector3d RotYPR2EulerYPR(const Eigen::Matrix3d &R) noexcept;
  static Eigen::Matrix3d EulerYPR2RotYPR(const Eigen::Vector3d &ypr) noexcept;
  static double GetGridConvergence(const double latitude, const double longitude);

  /// @brief From UTM to LL coordinate
  ///
  /// @param position  position
  ///
  /// @return position in LL coordinate
  static PositionLL Utm2Ll(PositionUTM const& position) noexcept;

  /// @brief From LL to UTM coordinate
  ///
  /// @param position  position
  ///
  /// @return position in UTM coordinate
  static PositionUTM Ll2Utm(PositionLL const& position) noexcept;


  /// @brief Get UTM letter designator
  ///
  /// @param latitude  latitude
  ///
  /// @return letter designator
  static char UtmLetterDesignator(double latitude) noexcept;

  static PositionGK Ll2GK(
    PositionLL const& position, double base_lon) noexcept;

  static void setBaseLon(double base_lon) noexcept;

  /// @brief Rad degree translation
  static constexpr double kRadPerDegree{M_PI / 180.0};
  static constexpr double kDegreePerRad{180.0 / M_PI};
  /// @brief caution: tencent map is diffrent from baidu map
  static constexpr std::uint64_t kGCJ02ToWGSFactorDivisor{
    static_cast<std::uint64_t>(2) << 30};
  /// @brief change kGCJ02 to wgs84
  static constexpr double kGCJ02ToWGSFactor{180.0 / kGCJ02ToWGSFactorDivisor};
  /// @brief change wgs84 to kGCJ02
  static constexpr double kWGSToGCJ02Factor{kGCJ02ToWGSFactorDivisor / 180.0};
  /// @brief GK base position
  static double base_lon_;
  

private:
  /// @brief WGS84 Parameters: major axis
  static constexpr double kWgs84A{6378137.0};
  /// @brief Minor axis
  static constexpr double kWgs84B{6356752.31424518};
  /// @brief Ellipsoid flattening
  static constexpr double kWgs84F{0.0033528107};
  /// @brief First eccentricity
  static constexpr double kWgs84E{0.0818191908};
  /****GK Start****/
  static constexpr double kWgs84C{kWgs84A / (1 - kWgs84F)};
  /// @brief 2 * f - f * f
  static constexpr double kWgs84E2{0.00669438};
  /// @brief e2 / (1 - e2)
  static constexpr double kWgs84EP2{0.006739497};
  /// @brief
  static constexpr double kDPI{0.0174532925199432957692};
  /// @brief kWgs84A * (1 - e2)
  static constexpr double kWgs84M0{6335439.32722994};
  /// @brief m2
  static constexpr double kWgs84M2{1.5 * kWgs84E2 * kWgs84M0};
  /// @brief m4
  static constexpr double kWgs84M4{1.25 * kWgs84E2 * kWgs84M2};
  /// @brief m6
  static constexpr double kWgs84M6{7 * kWgs84E2 * kWgs84M4 / 6};
  /// @brief m8
  static constexpr double kWgs84M8{9 * kWgs84E2 * kWgs84M6 / 8};
  /// @brief a0
  static constexpr double kWgs84A0{kWgs84M0 + kWgs84M2 / 2 + 3 * kWgs84M4 / 8 + 5 * kWgs84M6 / 16 + 35 * kWgs84M8 / 128};
  /// @brief a2
  static constexpr double kWgs84A2{kWgs84M2 / 2 + kWgs84M4 / 2 + 15 * kWgs84M6 / 32 + 7 * kWgs84M8 / 16};
  /// @brief a4
  static constexpr double kWgs84A4{kWgs84M4 / 8 + 3 * kWgs84M6 / 16 + 7 * kWgs84M8 / 32};
  /// @brief a6
  static constexpr double kWgs84A6{kWgs84M6 / 32 + kWgs84M8 / 16};
  /****GK End****/
  /// @brief Second eccentricity
  static constexpr double kWgs84EP{0.0820944379};
  /// @brief UTM Parameters: scale factor
  static constexpr double kUtmK0{0.9996};
  /// @brief False easting
  static constexpr double kUtmFE{500000.0};
  /// @brief False northing on north hemisphere
  static constexpr double kUtmFNN{0.0};
  /// @brief False northing on south hemisphere
  static constexpr double kUtmFNS{10000000.0};
  /// @brief E^2
  static constexpr double kUtmE2{kWgs84E * kWgs84E};
  /// @brief E^4
  static constexpr double kUtmE4{kUtmE2 * kUtmE2};
  /// @brief E^6
  static constexpr double kUtmE6{kUtmE4 * kUtmE2};
  /// @brief E'^2
  static constexpr double kUtmEP2{kUtmE2 / (1 - kUtmE2)};
};

} // namespace PointLoc
} // namespace Localization

#endif

/* vim: set ts=2 sw=2 sts=2 tw=100 et: */
