#include "coordinate_util.h"
#include <math_util.h>
namespace Localization
{
namespace PointLoc
{
double CoordinateUtil::GetGridConvergence(const double latitude, const double longitude)
{
  const double DEG2RAD = 1.0 / 180.0 * 3.141592654;
  double centerMeridian;
  const double zoneNumber = (longitude + 180.0) / 6.0; // 180.0, 6.0
  int zone = static_cast<int>(zoneNumber) + 1;
  centerMeridian = -183.0 + (static_cast<double>(zone) * 6.0); // 183.0, 6.0
  const double gridAngle = atan(tan((longitude - centerMeridian) * DEG2RAD) * sin(latitude * DEG2RAD));
  return gridAngle;
}

Eigen::Vector3d CoordinateUtil::RotYPR2EulerYPR(const Eigen::Matrix3d &R) noexcept
{
  Eigen::Vector3d n = R.col(0);
  Eigen::Vector3d o = R.col(1);
  Eigen::Vector3d a = R.col(2);
  Eigen::Vector3d ypr;
  const double y = atan2(n(1), n(0));
  const double p = atan2(-n(2), n(0) * cos(y) + n(1) * sin(y));
  const double r = atan2(a(0) * sin(y) - a(1) * cos(y), -o(0) * sin(y) + o(1) * cos(y));
  ypr(0) = y;
  ypr(1) = p;
  ypr(2) = r;

  return ypr;
}

Eigen::Matrix3d CoordinateUtil::EulerYPR2RotYPR(const Eigen::Vector3d &ypr) noexcept
{
  double y = ypr(0);
  double p = ypr(1);
  double r = ypr(2);
  Eigen::Matrix3d mRz;
  mRz << cos(y), -sin(y), 0.0, sin(y), cos(y), 0.0, 0.0, 0.0, 1.0;

  Eigen::Matrix3d mRy;
  mRy << cos(p), 0.0, sin(p), 0.0, 1.0, 0.0, -sin(p), 0.0, cos(p);

  Eigen::Matrix3d mRx;
  mRx << 1.0, 0.0, 0.0, 0.0, cos(r), -sin(r), 0.0, sin(r), cos(r);

  return mRz * mRy * mRx;
}


double CoordinateUtil::base_lon_ = 0.f;
CoordinateUtil::PositionLL CoordinateUtil::Utm2Ll(PositionUTM const& position)
    noexcept {
  //remove 500,000 meter offset for longitude
  double x{position.east - kUtmFE};
  double y{position.north};
  if (position.latitude_zone < 'N') {
    //remove 10,000,000 meter offset used for southern hemisphere
    y -= kUtmFNS;
  }
  //+3 puts origin in middle of zone
  double long_origin{static_cast<double>((
      position.longitude_zone_number - 1) * 6 - 180 + 3)};
  double m{y / kUtmK0};
  double mu{m / (kWgs84A * (1 - kUtmE2 * (0.25 + kUtmE2 *
      (3.0 / 64 + 5 * kUtmE2 / 256))))};
  double e1{(1 - sqrt(1 - kUtmE2)) / (1 + sqrt(1 - kUtmE2))};
  double phi1_rad{mu + e1 * ((1.5 - 27 * e1 * e1 / 32) * sin(2 * mu)
      + e1 * ((21.0 / 16 - 55 * e1 * e1 / 32) * sin(4 * mu)
      + 151 * e1 / 96 * sin(6 * mu)))};
  double n1{kWgs84A * MathUtil::InvSqrt(1 -
      kUtmE2 * sin(phi1_rad) * sin(phi1_rad))};
  double t1{tan(phi1_rad)*tan(phi1_rad)};
  double c1{kUtmEP2 * cos(phi1_rad) * cos(phi1_rad)};
  double r1{kWgs84A * (1 - kUtmE2) /
      pow(1 - kUtmE2 * sin(phi1_rad) * sin(phi1_rad), 1.5)};
  double d{x / (n1 * kUtmK0)};

  PositionLL pos{0.0, 0.0};
  pos.latitude = (phi1_rad - (n1 * tan(phi1_rad) / r1) * d * d * (0.5 -
      d * d * ((5 + 3 * t1 + c1 * (10 - 4 * c1 - 9 * kUtmEP2)) / 24 -
      (61 - 252 * kUtmEP2 + t1 * (90 + 45 * t1)  + c1 * (298 - 3 * c1)) *
      d * d / 720))) * kDegreePerRad;
  pos.longitude = long_origin + (d * (1 - d * d * ((1 + 2 * t1 + c1) / 6 -
      (5 + 8 * kUtmEP2 - c1 * (2 + 3 * c1) + t1 * (28 + 24 * t1)) * d * d /
      120)) / cos(phi1_rad)) * kDegreePerRad;
  return pos;
}

CoordinateUtil::PositionUTM CoordinateUtil::Ll2Utm(PositionLL const& position)
    noexcept {
  double long_temp{(position.longitude + 180) -
      static_cast<int>((position.longitude + 180) / 360) * 360 - 180};
  double lat_rad{position.latitude * kRadPerDegree};
  double long_rad{long_temp * kRadPerDegree};
  std::int8_t zone_number{static_cast<std::int8_t>((long_temp + 180) / 6 + 1)};
  if (position.latitude >= 56.0 && position.latitude < 64.0 &&
      long_temp >= 3.0 && long_temp < 12.0 ) {
    zone_number = 32;
  }
  if (position.latitude >= 72.0 && position.latitude < 84.0) {
    if (long_temp >= 0.0  && long_temp <  9.0 ) {
      zone_number = 31;
    } else if (long_temp >= 9.0  && long_temp < 21.0 ) {
      zone_number = 33;
    } else if (long_temp >= 21.0 && long_temp < 33.0) {
      zone_number = 35;
    } else if( long_temp >= 33.0 && long_temp < 42.0 ) {
      zone_number = 37;
    }
  }
  double long_origin{static_cast<double>((zone_number - 1) * 6 - 180 + 3)};
  double long_origin_rad{long_origin * kRadPerDegree};
  double n{kWgs84A / sqrt(1 - kUtmE2 * sin(lat_rad) * sin(lat_rad))};
  double t{tan(lat_rad) * tan(lat_rad)};
  double c{kUtmEP2 * cos(lat_rad) * cos(lat_rad)};
  double a{cos(lat_rad) * (long_rad - long_origin_rad)};
  double m{kWgs84A * ((1 - kUtmE2 * (0.25 + kUtmE2 * (3.0 / 64 +
      5 * kUtmE2 / 256))) * lat_rad - kUtmE2 * ((3.0 / 8 + kUtmE2 *
      (3.0 / 32 + 45 * kUtmE2 / 1024)) * sin(2 * lat_rad) -
      kUtmE2 * ((15.0 / 256 + 45 * kUtmE2 / 1024) * sin(4 * lat_rad) -
      35 * kUtmE2 / 3072 * sin(6 * lat_rad))))};

  PositionUTM pos{0.0, 0.0, 0, '\0'};
  pos.north = kUtmK0 * (m + n * tan(lat_rad) * a * a * (0.5 + a * a * ((5 - t +
      c * (9 + 4 * c)) / 24 + (61 - t * (58 - t) + 600 * c - 330 * kUtmEP2) *
      a * a / 720)));
  pos.east = kUtmK0 * n * a * (1 + a * a * ((1 - t + c) / 6 +
      (5 - t * (18 - t) + 72 * c - 58 * kUtmEP2) * a * a / 120)) + kUtmFE;
  if (position.latitude < 0) {
    pos.north += kUtmFNS;
  }
  pos.longitude_zone_number = zone_number;
  pos.latitude_zone = UtmLetterDesignator(position.latitude);
  return pos;
}

char CoordinateUtil::UtmLetterDesignator(double latitude) noexcept {
  char letter{'Z'};
  if (84.0 >= latitude && latitude >= 72.0) {
    letter = 'X';
  } else if (72.0 > latitude && latitude >= 64) {
    letter = 'W';
  } else if (64.0 > latitude && latitude >= 56.0) {
    letter = 'V';
  } else if (56.0 > latitude && latitude >= 48.0) {
    letter = 'U';
  } else if (48.0 > latitude && latitude >= 40.0) {
    letter = 'T';
  } else if (40.0 > latitude && latitude >= 32.0) {
    letter = 'S';
  } else if (32.0 > latitude && latitude >= 24.0) {
    letter = 'R';
  } else if (24.0 > latitude && latitude >= 16.0) {
    letter = 'Q';
  } else if (16.0 > latitude && latitude >= 8.0) {
    letter = 'P';
  } else if (8.0 > latitude && latitude >= 0.0) {
    letter = 'N';
  } else if (0.0 > latitude && latitude >= -8.0) {
    letter = 'M';
  } else if (-8.0 > latitude && latitude >= -16.0) {
    letter = 'L';
  } else if (-16.0 > latitude && latitude >= -24.0) {
    letter = 'K';
  } else if (-24.0 > latitude && latitude >= -32.0) {
    letter = 'J';
  } else if (-32.0 > latitude && latitude >= -40.0) {
    letter = 'H';
  } else if (-40.0 > latitude && latitude >= -48.0) {
    letter = 'G';
  } else if (-48.0 > latitude && latitude >= -56.0) {
    letter = 'F';
  } else if (-56.0 > latitude && latitude >= -64.0) {
    letter = 'E';
  } else if (-64.0 > latitude && latitude >= -72.0) {
    letter = 'D';
  } else if (-72.0 > latitude && latitude >= -80.0) {
    letter = 'C';
  }
  return letter;
}

CoordinateUtil::PositionGK CoordinateUtil::Ll2GK(
    PositionLL const& position, double base_lon) noexcept {
  CoordinateUtil::PositionGK gk{0.0, 0.0, 0.0};
  double ml{base_lon};
  double _iproj[12] = {kWgs84A, kWgs84B, kWgs84F, kWgs84C, kWgs84E2, kWgs84EP2, 
      kDPI, kWgs84A0, kWgs84A2, kWgs84A4, kWgs84A6, ml};
  double x{position.latitude};
  double y{position.longitude};
  double lat{x * _iproj[6]};
  double dL{(y - _iproj[11]) * _iproj[6]};
  double X{_iproj[7] * lat - _iproj[8] * sin(2 * lat) / 2 + 
      _iproj[9] * sin(4 * lat) / 4 - _iproj[10] * sin(6 * lat) / 6};
  double tn{tan(lat)};
  double tn2{tn * tn};
  double tn4{tn2 * tn2};
  double j2{(1 / pow(1 - _iproj[2], 2) - 1) * pow(cos(lat), 2)};
  double n{_iproj[0] / sqrt(1.0 - _iproj[4] * sin(lat) * sin(lat))};
  //
  double temp[6] = {0};
  temp[0] = n * sin(lat) * cos(lat) * dL * dL / 2;
  temp[1] = n * sin(lat) * pow(cos(lat), 3) * (5 - tn2 + 9 * j2 + 4 * j2 * j2) * pow(dL, 4) / 24;
  temp[2] = n * sin(lat) * pow(cos(lat), 5) * (61 - 58 * tn2 + tn4) * pow(dL, 6) / 720;
  temp[3] = n * cos(lat) * dL;
  temp[4] = n * pow(cos(lat), 3) * (1 - tn2 + j2) * pow(dL, 3) / 6;
  temp[5] = n * pow(cos(lat), 5) * (5 - 18 * tn2 + tn4 + 14 * j2 - 58 * tn2 * j2) * pow(dL, 5) / 120;
  // 
  gk.y = X + temp[0] + temp[1] + temp[2];
  gk.x = temp[3] + temp[4] + temp[5];
  gk.base_lon = base_lon;
  return gk;
}
} // namespace PointLoc
} // namespace Localization
