#include "point_loc_interface.h"
#include <unistd.h>
#include <random>
#include <iomanip>

namespace Localization
{
namespace PointLoc
{
PointLocInterface::PointLocInterface() {

}

PointLocInterface::~PointLocInterface() {
  StopOdom();
}

void PointLocInterface::Init() {
  std::cout<<"Init():point_loc_interface try to init"<<std::endl;
  locInited_ = false;
  computeThreadInited_ = true;
  outputLocStamp_ = 0.0;
  computeThread_ = std::thread([this] {VehICPThread();});
}

void PointLocInterface::StopOdom() {
  locInited_ = false;
  computeThreadInited_ = false;
}

void PointLocInterface::ResetOdom() {
  locInited_ = false;
  outputLocStamp_ = 0.0;
}

void PointLocInterface::SetInputFrame(MatchFrame &matchFrame) {
    std::unique_lock<std::mutex> lock(mutexFrameBuf_);
    std::shared_ptr<MatchFrame> pMatchFrame = std::make_shared<MatchFrame>(matchFrame);
    frameBuf_.emplace_back(pMatchFrame);
//    std::cout<<std::setprecision(15)<<"SetInputFrame():时间戳为"<<matchFrame.stamp<<std::endl;
    if (frameBuf_.size() > 10U) {
      frameBuf_.pop_front();
    }
}

bool PointLocInterface::TakeInputFrame(MatchFrame &matchFrame) {
  bool res = false;
  std::unique_lock<std::mutex> lock(mutexFrameBuf_);
  if (!frameBuf_.empty()) {
    matchFrame = *(frameBuf_.back());
    if (CheckInputFrame(matchFrame)) {
//      std::cout<<std::setprecision(15)<<"参与计算的的frame时间(front):"<<matchFrame.stamp<<std::endl;
//      std::cout<<std::setprecision(15)<<"frame队列末尾时间(back):"<<frameBuf_.back()->stamp<<std::endl;
      res = true;
    }
    frameBuf_.pop_front();
  } else {
    std::cerr<<"TakeInputFrame():MatchedFrame buffer is empty,continue and wait input"<<std::endl;
  }
  return res;
}
bool PointLocInterface::CheckInputFrame(MatchFrame &matchFrame) {
  bool res = true;
//  std::cout<<std::setprecision(15)<<"CheckInputFrame():时间戳为"<<matchFrame.stamp<<std::endl;
  if (matchFrame.stamp < 1e-3) {
    res = false;
    std::cerr<<"CheckInputPoints():这帧感知时间戳太小,不可信,false"<<std::endl;
  }
  static double lastTime = matchFrame.stamp;
  double curTime = matchFrame.stamp;
  if (curTime - lastTime < 1e-3) {
    res = false;
    std::cerr<<std::setprecision(15)<<"CheckInputPoints():感知时间戳重复,false,curStamp:"<<
      curTime<<",lastStamp:"<<lastTime<<std::endl;
  }
  lastTime = curTime;
  return res;
}

void PointLocInterface::OdomResProcess(double &stamp,Eigen::Vector3d &pos,Eigen::Matrix3d &rot,
                                       PointLocResult &odomRes)
{
  odomRes.stamp_ = stamp;
  odomRes.posUTM_ = pos;
  CoordinateUtil::PositionUTM utmPosition;
  utmPosition.longitude_zone_number = 51;
  utmPosition.latitude_zone = 'R';
  utmPosition.east = odomRes.posUTM_.x();
  utmPosition.north = odomRes.posUTM_.y();
  auto latLon = CoordinateUtil::Utm2Ll(utmPosition);
  odomRes.posLLA_.x() = latLon.longitude;
  odomRes.posLLA_.y() = latLon.latitude;
  odomRes.rot_ = rot;
  odomRes.quat_ = rot;
  Eigen::Vector3d YPR = RotYPR2EulerYPR(odomRes.quat_.toRotationMatrix());
  odomRes.attitude_.x() = YPR.z();
  odomRes.attitude_.y() = YPR.y();
  odomRes.attitude_.z() = YPR.x();
  SetLocOutput(odomRes);
  if (isPrintRes_) {
    std::cout<<std::setprecision(15)<<"OdomResProcess():转换完一次结果，时间戳是:"<<odomRes.stamp_<<std::endl;
    std::cout<<"OdomResProcess():转换完一次结果，UTM是:"<<odomRes.posUTM_.transpose()<<std::endl;
    std::cout<<"OdomResProcess():转换完一次结果，Lon,Lat,Alt是:"<<odomRes.posLLA_.transpose()<<std::endl;
    std::cout<<"OdomResProcess():转换完一次结果，RPY是:"<<odomRes.attitude_.transpose() * RAD_TO_DEG<<std::endl;
  }
}

bool PointLocInterface::GetLocOutput(PointLocResult &res) {
  std::unique_lock<std::mutex> lock(mutexResult_);
  res = locRes_;
  return locInited_;
}

void PointLocInterface::SetLocOutput(PointLocResult &res) {
  std::unique_lock<std::mutex> lock(mutexResult_);
  locRes_ = res;
}

void PointLocInterface::VehICPThread()
{
  std::cout<<"VehICPThread():打开VehICPThread()成功"<<std::endl;
  while (computeThreadInited_) {
    usleep(5e4);

    MatchFrame curFrame;
    if (!TakeInputFrame(curFrame)) {
      std::cerr<<"VehICPThread():没有输入数据,等待输入,continue"<<std::endl;
      PointLocResult emptyRes;
      SetLocOutput(emptyRes);
      continue;
    }

    PointLocResult locRes;
    if (locInited_ == false) {
      Eigen::Vector3d posPrior = curFrame.Egmo.utmPosition_;
      std::cout<<"VehICPThread():尝试初始化：posPrior:"<<posPrior.transpose()<<std::endl;
      if (curFrame.Egmo.solutionStatus == 1U) {
        if (computeICP_.ComputeByFrameVCS(curFrame,locRes)) {
          outputLocStamp_ = curFrame.stamp;
          locInited_ = true;
          std::cout << "VehICPThread():视觉定位初始化成功！opsOpt:" << locRes.posUTM_.transpose() << std::endl;
        } else {
          std::cout<<"VehICPThread():初始化失败！用先验计算没有收敛"<<std::endl;
        }
      } else {
        std::cout<<"VehICPThread():初始化失败！egmoCur.solutionStatus:"<<static_cast<int>(curFrame.Egmo.solutionStatus)<<std::endl;
      }
      continue;
    }

    auto frameStamp = curFrame.Egmo.stamp_;
    if (frameStamp > outputLocStamp_) {
      outputLocStamp_ = frameStamp;
    } else {
      std::cerr<<std::setprecision(15)<<"VehICPThread():input frame repeat,input stamp:"<<
               frameStamp<<"last output stamp"<<outputLocStamp_<<
               ",continue and wait new input"<<std::endl;
      continue;
    }

    static int failCount = 0;
    if (computeICP_.ComputeByFrameVCS(curFrame,locRes)) {
      failCount = 0;
      outputLocStamp_ = curFrame.stamp;
      locRes.status_ = 4U;
    } else {
      failCount++;
      std::cout<<"VehICPThread():visLoc ComputeByFrameVCS() false,failCount:"<<failCount<<std::endl;
    }

    Eigen::Vector3d finalPos = locRes.posUTM_;
    Eigen::Matrix3d finalRot = locRes.quat_.toRotationMatrix();
    OdomResProcess(outputLocStamp_,finalPos,finalRot,locRes);
    
    if (failCount > 100) {
      ResetOdom();
      std::cout<<"VehICPThread():visLoc fail, reset"<<std::endl;
    }
  }
}

} // namespace PointLoc
} // namespace Localization
